/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: utils.h,v 1.75 2010/12/27 12:40:15 zlb Exp $ */

#ifndef PHG_UTILS_H

/* Fortran names */
#define F77name(s)	s ## _		/* for names without underscore */
#define F77name_(s)	s ## __		/* for names containing underscores */

#ifdef __cplusplus
extern "C" {
#endif

/* Macro for avoiding 'unused variable ...' warnings */
#define Unused(v)	(void)(v)

/* Name of the current function */
#if defined __STDC_VERSION__
# if __STDC_VERSION__ < 199901L
#  if __GNUC__ >= 2
#   define __func__ __FUNCTION__
#  else
#   define __func__ "none"
#  endif
# endif
#endif

/* #define phgDebug(level, fmt, ...) phgInfo(level, fmt, ##__VA_ARGS__) */
#if DEBUG
# define phgDebug(s) phgInfo s
#else
# define phgDebug(s)
#endif

/* The macro CB_ARGS is used to control whether 'g' should be
 * in the argument list of the callback functions. */
#if 0
/* carry g in the argument list of callback functions */
# define CB_ARGS(e)	(GRID *g, SIMPLEX *e)
# define CB_ARGS0(g, e)	(g, e)
#else
/* don't carry g in the argument list of callback functions */
# define CB_ARGS(e)	(SIMPLEX *e)
# define CB_ARGS0(g, e)	(e)
#endif

#define IsLeaf(e) ((e)->children[0] == NULL && (e)->children[1] == NULL)
#define HasLocalNeighbour(e, v) \
    ((((SIMPLEX *)(e))->bound_type[v] & (INTERIOR | REMOTE)) == INTERIOR)
#define HasRemoteNeighbour(e, v) \
    ((((SIMPLEX *)(e))->bound_type[v] & (INTERIOR | REMOTE)) == \
	(INTERIOR | REMOTE))

#define GetNeighbour(e, v) \
    ((SIMPLEX *)(HasLocalNeighbour(e, v) ? (e)->neighbours[v] : NULL))

#define GetRNeighbour(g, e, v) \
    ((HasRemoteNeighbour(e, v) && (g)->neighbours.count > 0) \
	? (g)->neighbours.list + (size_t)((e)->neighbours[v]) : \
	  (RNEIGHBOUR *)NULL)

/**< Magic numbers */
#define MagicNumber(a, b, c, d)	(((((((INT)d) << 8) | c) << 8 | b) << 8) | a)
#define MAGIC_DOF 	MagicNumber('D', 'O', 'F', ' ')
#define MAGIC_MAT 	MagicNumber('M', 'A', 'T', ' ')
#define MAGIC_VEC 	MagicNumber('V', 'E', 'C', ' ')

#define MagicCheck(m, var) {						\
    if ((var) != NULL && (var)->magic != MAGIC ## _ ## m) {		\
	INT magic = MAGIC ## _ ## m;					\
	fprintf(stderr, "\n*** Error in %s (%s:%d): invalid or "	\
		"uninitialized %c%c%c object.\n", __func__,		\
		__FILE__, __LINE__, magic & 255, (magic >> 8) & 255,	\
			  (magic >> 16) & 255);				\
	phgAbort(1);							\
    }									\
}

/* Macro for warning an untested code segment */
#define WARN_UNTESTED	{						\
  static BOOLEAN warned = FALSE;					\
  if (!warned) {							\
    warned = TRUE;							\
    phgWarning("%s:%d: untested code segment!\n", __FILE__, __LINE__);	\
  }									\
}

#if USE_OMP
# if DEBUG
#  define CheckThread							\
    if (phgThreadId > 0)					\
	phgError(1, "The function %s (%s:%d) is not thread aware.\n",	\
			__func__, __FILE__, __LINE__);
# else	/* DEBUG */
#  define CheckThread
# endif	/* DEBUG */
#else	/* USE_OMP */
# define CheckThread
#endif	/* USE_OMP */

/**< Protocol for user functions mapping bdry types.
 *
 * The function should return a combination of PHG's bdry type, or -1 if
 * bc_type is unknown */
typedef int (*BDRY_MAP_FUNC)(int bc_type);

/* global variables */
extern const char *phgCurrentFunctionName_;

extern int phgMaxThreads, phgThreadId;
#if USE_OMP
#pragma omp threadprivate(phgThreadId)
#endif	/* USE_OMP */

extern INT phgVerbosity;
extern INT phgTraverseDepth, phgTraverseIndex, _phg_i;
extern SIMPLEX *_phg_step_element(GRID *g);
extern BOOLEAN phgInitialized;
extern BOOLEAN phgMasterSlave;
extern BDRY_MAP_FUNC phgBdryMapFunc_;
extern BOOLEAN _phg_pause;

/* common auxiliary functions */
int phgCompINT(const void *p0, const void *p1);		/* compares 2 'INT's */
int phgCompint(const void *p0, const void *p1);		/* compares 2 'int's */
INT phgBinarySearch(INT n, const void *table, const void *key, size_t size,
                                int (*comp)(const void *, const void *));

#define FunctionEntry_head						\
    double phg_t0[3];							\
    const char *phg_fname_save = phgCurrentFunctionName_;

#define FunctionEntry_tail						\
    phgCurrentFunctionName_ = __func__;					\
    phgGetTime(phg_t0);

#define FunctionEntry	FunctionEntry_head FunctionEntry_tail

/* Note: the dummy 'if (...)' here allows not including Return
 * in curly braces */
#define Return if ((phgCurrentFunctionName_ = phg_fname_save) || TRUE) return

#define PrintTime(vl) 							\
    if (vl >= 0 && vl <= phgVerbosity && IS_RANK0) {			\
	double phg_t1[3];						\
	phgGetTime(phg_t1);						\
	phgInfo(vl, "user=%f, sys=%f, wall=%f\n",			\
			phg_t1[0] - phg_t0[0], phg_t1[1] - phg_t0[1],	\
			phg_t1[2] - phg_t0[2]);				\
	phg_t0[0] = phg_t1[0]; phg_t0[1] = phg_t1[1]; phg_t0[2] = phg_t1[2]; \
    }

#if DEBUG
# define FreeAtExit_(ptr, level) {			\
    static BOOLEAN registered = FALSE;			\
    if (!registered) {					\
	phgFreeAtExit((void **)(void *)&ptr, level);	\
	registered = TRUE;				\
    }							\
}
# define FreeAtExitNoCheck_(ptr, level) \
    phgFreeAtExit((void **)(void *)&ptr, level)
#else
# define FreeAtExit_(ptr, level)
# define FreeAtExitNoCheck_(ptr, level)
#endif
#define FreeAtExit(ptr)	FreeAtExit_(ptr, 0)
#define FreeAtExitNoCheck(ptr)	FreeAtExitNoCheck_(ptr, 0)

void phgTrapSignals_(BOOLEAN init);
void phgInit(int *argc, char ***argv);
void phgSetVerbosity_(GRID *g);
int  phgSetVerbosity(int verbosity);
void phgPause(int seconds);
void phgFinalize(void);
void phgInitElements(SIMPLEX *s, int count);
SIMPLEX *phgNewElements(int count);
SIMPLEX *phgReallocElements(SIMPLEX *s, int oldcount, int newcount);
COORD *phgNewVertices(int count);
void phgNewGrid_(GRID *g);
GRID *phgNewGrid(int flags);
int phgGridId(const GRID *g);
#define phgGridIsValid(g)	(phgGridId(g) >= 0)
void *phgAlloc(size_t size);
void *phgCalloc(size_t nmemb, size_t size);

void *phgRealloc0(void *ptr, size_t size);
void *phgRealloc__(void *ptr, size_t size, size_t oldsize,
					const char *file, int line);

/* The macro phgRealloc_ is the function actually used in PHG's lib */
#if REALLOC_HACK
 /* For MVAPICH2 on DeepComp7000: use phgRealloc__ which falls back to malloc()
  *	+ memcpy() if realloc() fails (realloc() may fail when np > 64) */
# define phgRealloc_(ptr, size, oldsize) \
	 phgRealloc__(ptr, size, oldsize, __FILE__, __LINE__)
#else	/* REALLOC_HACK */
 /* Default: use phgRealloc0 which relies on realloc() ('oldsize' not used). */
# define phgRealloc_(ptr, size, oldsize) phgRealloc0(ptr, size)
#endif	/* REALLOC_HACK */

void phgFree(void *ptr);
int  phgMemcmp(const void *v0, const void *v1, size_t n);
void phgFreeAtExit(void **ptr, int level);
void phgWarning(const char *fmt, ...);
void phgAbort(int code);
void phgError(int code, const char *fmt, ...);
void phgInfo(int verbose_level, const char *fmt, ...);
int phgPrintf(const char *fmt, ...);
FILE *phgFOpen(GRID *g, const char *path, const char *mode);
int phgFPrintf(GRID *g, FILE *fp, char *fmt, ...);
int phgFClose(GRID *g, FILE *fp);
double phgGetTime(double tarray[]);

BYTE phgOppositeVertex(GRID *g, SIMPLEX *e, BYTE v, SIMPLEX *e_op);
void phgFreeElement(SIMPLEX **eptr);
void phgFreeGrid_(GRID *g);
void phgFreeGrid(GRID **gptr);
void phgTraverseElements(GRID *g, BOOLEAN(*cb)CB_ARGS(e));
void phgTraverseAllElements(GRID *g, BOOLEAN(*cb) CB_ARGS(e));
void phgCheckConformity(GRID *g);
void phgRemoveDegeneratedElements(GRID *g);
void phgUpdateNeighbours(GRID *g);
void phgUpdateEdges(GRID *g);
#if Dim == 3
void phgUpdateFaces(GRID *g);
#endif
void phgUpdateElementIndices(GRID *g, INT nleaf_old, INT nleaf_global_old);
INT  phgGetPatch(GRID *g, SIMPLEX *e, BYTE v0, BYTE v1, SIMPLEX ***p);
void phgDumpElement(GRID *g, SIMPLEX *e);
void phgDumpPatch(GRID *g, SIMPLEX *e, int v0, int v1);
void phgDumpGridInfo(GRID *g);
void phgDumpGrid(GRID *g);
void phgDumpTree(GRID *g);
void phgDumpBranch(GRID *g, SIMPLEX *e);
void phgCloseInputFile_(FILE *fp);
FILE *phgOpenInputFile_(const char *filename);
void phgImportSetBdryMapFunc(BDRY_MAP_FUNC func);
BTYPE phgImportSetDefaultBdryType(BTYPE type);
BOOLEAN phgImport(GRID *g, const char *filename, BOOLEAN distr);
#if ALLOW_CURVED_BOUNDARY
void phgInitBoundaryFunctions(GRID *g);
#endif
FLOAT phgGetBoundaryError(GRID *g, SIMPLEX *e);

GRID *phgDupGrid(GRID *g, BOOLEAN flatten);

#if USE_MPI
INT  phgCountRNeighbours(int nprocs, RNEIGHBOUR *list, INT count,
				int *cnts, int *dspls);
#endif	/* USE_MPI */

/* Macro to get the local index of a given vertex of a given edge,
 * 'edge' is the local edge no, 'vertex' is 0 or 1 */
#define GetEdgeVertex(edge, vertex) (GetEdgeVertex_[edge][vertex])
extern BYTE GetEdgeVertex_[NEdge][2];

/* Macro to get the local index of a given vertex of a given face,
 * 'face' is the local face no, 'vertex' is 0, 1, or 2 */
#define GetFaceVertex(face, vertex) (GetFaceVertex_[face][vertex])
extern BYTE GetFaceVertex_[NFace][3];

/* Macro to get the local index of the edge on the two given vertices */
#define GetEdgeNo(v0, v1) (GetEdgeNo_[v0][v1])
extern BYTE GetEdgeNo_[NVert][NVert];

/* utility function */
void phgUpdateBoundaryTypes(GRID *g);

#ifdef __cplusplus
}
#endif

#define PHG_UTILS_H
#endif
