/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: dg.c,v 1.36 2010/07/01 05:50:46 zlb Exp $ */

/* This file implements the discontinuous Galerkin elements */

#include "phg.h"

static const FLOAT *
DG_bas(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
{
    DOF hb;
    DOF_TYPE *type;

    type = (DofIsHP(dof) ? DOF_DGn[dof->hp->elem_order[e->index]] : dof->type);

    if (!USE_HB_FOR_DG || type->order < 5)
	return DOF_Pn[type->order]->BasFuncs(dof, e, no0, no1, lambda);
    hb.g = dof->g;
    hb.type = DOF_HBn[type->order];
    hb.hp = NULL;
    hb.dim = dof->dim;
    return hb.type->BasFuncs(&hb, e, no0, no1, lambda);
}

static const FLOAT *
DG_grad(DOF *dof, SIMPLEX *e, int no0, int no1, const FLOAT *lambda)
{
    DOF hb;
    DOF_TYPE *type;

    type = (DofIsHP(dof) ? DOF_DGn[dof->hp->elem_order[e->index]] : dof->type);

    if (!USE_HB_FOR_DG || type->order < 5)
	return DOF_Pn[type->order]->BasGrads(dof, e, no0, no1, lambda);

    hb.g = dof->g;
    hb.type = DOF_HBn[type->order];
    hb.hp = NULL;
    hb.dim = dof->dim;
    return hb.type->BasGrads(&hb, e, no0, no1, lambda);
}

#if 0
static void
DG_interp(DOF *dof, SIMPLEX *e, FLOAT **old_data, FLOAT **new_data)
/* interpolation */
{
    /* TODO: permute edge and face data, then call lagrange->Pn_interp */
}
#endif

HP_INFO hp_info_DG = {
    DOF_DGn,		/* types */
    "Hierarchic DG",	/* name */
    FE_L2,		/* FE space */
    1,			/* dim */
    0,			/* min_order */
    15			/* max_order */
};

static FLOAT DG0_points[] = {0.25L,0.25L,0.25L,0.25L};
DOF_TYPE DOF_DG0_ = {DofCache, "DG0", DG0_points, NULL,
		     DOF_DG0, DOF_P0, &hp_info_DG,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad, NULL, FE_L2,
		     TRUE, FALSE, -1, 1, 0, -1, 1,
		     0, 0, 0, 1};

static FLOAT DG1_points[] = {
    1.L,0.L,0.L,0.L,  0.L,1.L,0.L,0.L,  0.L,0.L,1.L,0.L,  0.L,0.L,0.L,1.L
};
DOF_TYPE DOF_DG1_ = {DofCache, "DG1", DG1_points, NULL,
		     DOF_DG0, DOF_P1, &hp_info_DG,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad, NULL, FE_L2,
		     TRUE, FALSE, -1, NVert, 1, -1, 1,
		     0, 0, 0, NVert};

static FLOAT DG2_points[] = {
    1.L,0.L,0.L,0.L,  0.L,1.L,0.L,0.L,  0.L,0.L,1.L,0.L,  0.L,0.L,0.L,1.L,

    .5L,.5L,0.L,0.L,  .5L,0.L,.5L,0.L,  .5L,0.L,0.L,.5L,
    0.L,.5L,.5L,0.L,  0.L,.5L,0.L,.5L,  0.L,0.L,.5L,.5L
};
DOF_TYPE DOF_DG2_ = {DofCache, "DG2", DG2_points, NULL,
		     DOF_DG1, DOF_P2, &hp_info_DG,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad, NULL, FE_L2,
		     TRUE, FALSE, -1, NVert + NEdge, 2, -1, 1,
		     0, 0, 0, NVert + NEdge};

static FLOAT DG3_points[] = {
    1.L,0.L,0.L,0.L,  0.L,1.L,0.L,0.L,  0.L,0.L,1.L,0.L,  0.L,0.L,0.L,1.L,

    2.L/3.L, 1.L/3.L, 0.L, 0.L,   1.L/3.L, 2.L/3.L, 0.L, 0.L,
    2.L/3.L, 0.L, 1.L/3.L, 0.L,   1.L/3.L, 0.L, 2.L/3.L, 0.L,
    2.L/3.L, 0.L, 0.L, 1.L/3.L,   1.L/3.L, 0.L, 0.L, 2.L/3.L,
    0.L, 2.L/3.L, 1.L/3.L, 0.L,   0.L, 1.L/3.L, 2.L/3.L, 0.L,
    0.L, 2.L/3.L, 0.L, 1.L/3.L,   0.L, 1.L/3.L, 0.L, 2.L/3.L,
    0.L, 0.L, 2.L/3.L, 1.L/3.L,   0.L, 0.L, 1.L/3.L, 2.L/3.L,

    0.L, 1.L/3.L, 1.L/3.L, 1.L/3.L,
    1.L/3.L, 0.L, 1.L/3.L, 1.L/3.L,
    1.L/3.L, 1.L/3.L, 0.L, 1.L/3.L,
    1.L/3.L, 1.L/3.L, 1.L/3.L, 0.L
};
DOF_TYPE DOF_DG3_ = {DofCache, "DG3", DG3_points, NULL,
		     DOF_DG2, DOF_P3, &hp_info_DG,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad, NULL, FE_L2,
		     TRUE, FALSE, -1, NVert + 2*NEdge + NFace, 3, -1, 1,
		     0, 0, 0, NVert + 2*NEdge + NFace};

static FLOAT DG4_points[] = {
    1.L,0.L,0.L,0.L,  0.L,1.L,0.L,0.L,  0.L,0.L,1.L,0.L,  0.L,0.L,0.L,1.L,

    .75L, .25L, 0.L, 0.L,   .5L, .5L, 0.L, 0.L,   .25L, .75L, 0.L, 0.L,
    .75L, 0.L, .25L, 0.L,   .5L, 0.L, .5L, 0.L,   .25L, 0.L, .75L, 0.L,
    .75L, 0.L, 0.L, .25L,   .5L, 0.L, 0.L, .5L,   .25L, 0.L, 0.L, .75L,
    0.L, .75L, .25L, 0.L,   0.L, .5L, .5L, 0.L,   0.L, .25L, .75L, 0.L,
    0.L, .75L, 0.L, .25L,   0.L, .5L, 0.L, .5L,   0.L, .25L, 0.L, .75L,
    0.L, 0.L, .75L, .25L,   0.L, 0.L, .5L, .5L,   0.L, 0.L, .25L, .75L,

    0.L, .5L, .25L, .25L,    0.L, .25L, .5L, .25L,   0.L, .25L, .25L, .5L,
    .5L, 0.L, .25L, .25L,    .25L, 0.L, .5L, .25L,   .25L, 0.L, .25L, .5L,
    .5L, .25L, 0.L, .25L,    .25L, .5L, 0.L, .25L,   .25L, .25L, 0.L, .5L,
    .5L, .25L, .25L, 0.L,    .25L, .5L, .25L, 0.L,   .25L, .25L, .5L, 0.L,

    .25L, .25L, .25L, .25L
};
DOF_TYPE DOF_DG4_ = {DofCache, "DG4", DG4_points, NULL,
		     DOF_DG3, DOF_P4, &hp_info_DG,
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
		     phgDofInitFuncPoint,
		     DG_bas, DG_grad, NULL, FE_L2,
		     TRUE, FALSE, -1, NVert + 3*(NEdge + NFace) + 1, 4, -1, 1,
		     0, 0, 0, NVert + 3*(NEdge + NFace) + 1};

DOF_TYPE *DOF_DGn[] = {
    DOF_DG0,
    DOF_DG1, DOF_DG2, DOF_DG3, DOF_DG4,
    DOF_DG5, DOF_DG6, DOF_DG7, DOF_DG8, DOF_DG9, DOF_DG10,
    DOF_DG11, DOF_DG12, DOF_DG13, DOF_DG14, DOF_DG15,
    NULL
};

#if USE_HB_FOR_DG

static void
DG_init(DOF *dof, SIMPLEX *e, GTYPE type, int index,
	DOF_USER_FUNC userfunc,
	DOF_USER_FUNC_LAMBDA userfunc_lambda,
	const FLOAT *funcvalues, FLOAT *dofvalues, FLOAT **pdofvalues)
{
    DOF_HB1->InitFunc(dof, e, type, index, userfunc, userfunc_lambda,
			funcvalues, dofvalues, pdofvalues);
}

/* DGp = Hierarchical basis of order p, p > 4 */
#define NBas(p)	NVert + (p-1) * NEdge + (p-1)*(p-2)/2 * NFace + \
		(p-1)*(p-2)*(p-3)/6

#define DofType(order, type, name, grad, base_type)			\
    DOF_TYPE type = {DofCache, name, NULL, NULL,			\
		     grad, base_type, &hp_info_DG,			\
		     phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,	\
		     DG_init, DG_bas, DG_grad, NULL, FE_L2,		\
		     FALSE, FALSE, -1, NBas(order), order, -1, 1,	\
		     0, 0, 0, NBas(order)};
DofType(5,  DOF_DG5_,  "DG5",  DOF_DG4,  DOF_HB5);
DofType(6,  DOF_DG6_,  "DG6",  DOF_DG5,  DOF_HB6);
DofType(7,  DOF_DG7_,  "DG7",  DOF_DG6,  DOF_HB7);
DofType(8,  DOF_DG8_,  "DG8",  DOF_DG7,  DOF_HB8);
DofType(9,  DOF_DG9_,  "DG9",  DOF_DG8,  DOF_HB9);
DofType(10, DOF_DG10_, "DG10", DOF_DG9,  DOF_HB10);
DofType(11, DOF_DG11_, "DG11", DOF_DG10, DOF_HB11);
DofType(12, DOF_DG12_, "DG12", DOF_DG11, DOF_HB12);
DofType(13, DOF_DG13_, "DG13", DOF_DG12, DOF_HB13);
DofType(14, DOF_DG14_, "DG14", DOF_DG13, DOF_HB14);
DofType(15, DOF_DG15_, "DG15", DOF_DG14, DOF_HB15);

#else	/* USE_HB_FOR_DG */

/* Note: insert the file "/tmp/DOF_DGn.inc" generated by the script
 *	 "../utils/lagrange_gen" below */

/*======= The following structs are generated by 'utils/lagrange_gen' ======*/

static FLOAT DG5_points[] = {
    5.L/5.L,  0.L/5.L,  0.L/5.L,  0.L/5.L,
    0.L/5.L,  5.L/5.L,  0.L/5.L,  0.L/5.L,
    0.L/5.L,  0.L/5.L,  5.L/5.L,  0.L/5.L,
    0.L/5.L,  0.L/5.L,  0.L/5.L,  5.L/5.L,
    4.L/5.L,  1.L/5.L,  0.L/5.L,  0.L/5.L,
    3.L/5.L,  2.L/5.L,  0.L/5.L,  0.L/5.L,
    2.L/5.L,  3.L/5.L,  0.L/5.L,  0.L/5.L,
    1.L/5.L,  4.L/5.L,  0.L/5.L,  0.L/5.L,
    4.L/5.L,  0.L/5.L,  1.L/5.L,  0.L/5.L,
    3.L/5.L,  0.L/5.L,  2.L/5.L,  0.L/5.L,
    2.L/5.L,  0.L/5.L,  3.L/5.L,  0.L/5.L,
    1.L/5.L,  0.L/5.L,  4.L/5.L,  0.L/5.L,
    4.L/5.L,  0.L/5.L,  0.L/5.L,  1.L/5.L,
    3.L/5.L,  0.L/5.L,  0.L/5.L,  2.L/5.L,
    2.L/5.L,  0.L/5.L,  0.L/5.L,  3.L/5.L,
    1.L/5.L,  0.L/5.L,  0.L/5.L,  4.L/5.L,
    0.L/5.L,  4.L/5.L,  1.L/5.L,  0.L/5.L,
    0.L/5.L,  3.L/5.L,  2.L/5.L,  0.L/5.L,
    0.L/5.L,  2.L/5.L,  3.L/5.L,  0.L/5.L,
    0.L/5.L,  1.L/5.L,  4.L/5.L,  0.L/5.L,
    0.L/5.L,  4.L/5.L,  0.L/5.L,  1.L/5.L,
    0.L/5.L,  3.L/5.L,  0.L/5.L,  2.L/5.L,
    0.L/5.L,  2.L/5.L,  0.L/5.L,  3.L/5.L,
    0.L/5.L,  1.L/5.L,  0.L/5.L,  4.L/5.L,
    0.L/5.L,  0.L/5.L,  4.L/5.L,  1.L/5.L,
    0.L/5.L,  0.L/5.L,  3.L/5.L,  2.L/5.L,
    0.L/5.L,  0.L/5.L,  2.L/5.L,  3.L/5.L,
    0.L/5.L,  0.L/5.L,  1.L/5.L,  4.L/5.L,
    0.L/5.L,  3.L/5.L,  1.L/5.L,  1.L/5.L,
    0.L/5.L,  2.L/5.L,  2.L/5.L,  1.L/5.L,
    0.L/5.L,  2.L/5.L,  1.L/5.L,  2.L/5.L,
    0.L/5.L,  1.L/5.L,  3.L/5.L,  1.L/5.L,
    0.L/5.L,  1.L/5.L,  2.L/5.L,  2.L/5.L,
    0.L/5.L,  1.L/5.L,  1.L/5.L,  3.L/5.L,
    3.L/5.L,  0.L/5.L,  1.L/5.L,  1.L/5.L,
    2.L/5.L,  0.L/5.L,  2.L/5.L,  1.L/5.L,
    2.L/5.L,  0.L/5.L,  1.L/5.L,  2.L/5.L,
    1.L/5.L,  0.L/5.L,  3.L/5.L,  1.L/5.L,
    1.L/5.L,  0.L/5.L,  2.L/5.L,  2.L/5.L,
    1.L/5.L,  0.L/5.L,  1.L/5.L,  3.L/5.L,
    3.L/5.L,  1.L/5.L,  0.L/5.L,  1.L/5.L,
    2.L/5.L,  2.L/5.L,  0.L/5.L,  1.L/5.L,
    2.L/5.L,  1.L/5.L,  0.L/5.L,  2.L/5.L,
    1.L/5.L,  3.L/5.L,  0.L/5.L,  1.L/5.L,
    1.L/5.L,  2.L/5.L,  0.L/5.L,  2.L/5.L,
    1.L/5.L,  1.L/5.L,  0.L/5.L,  3.L/5.L,
    3.L/5.L,  1.L/5.L,  1.L/5.L,  0.L/5.L,
    2.L/5.L,  2.L/5.L,  1.L/5.L,  0.L/5.L,
    2.L/5.L,  1.L/5.L,  2.L/5.L,  0.L/5.L,
    1.L/5.L,  3.L/5.L,  1.L/5.L,  0.L/5.L,
    1.L/5.L,  2.L/5.L,  2.L/5.L,  0.L/5.L,
    1.L/5.L,  1.L/5.L,  3.L/5.L,  0.L/5.L,
    2.L/5.L,  1.L/5.L,  1.L/5.L,  1.L/5.L,
    1.L/5.L,  2.L/5.L,  1.L/5.L,  1.L/5.L,
    1.L/5.L,  1.L/5.L,  2.L/5.L,  1.L/5.L,
    1.L/5.L,  1.L/5.L,  1.L/5.L,  2.L/5.L
};
DOF_TYPE DOF_DG5_ = {
    DofCache, "DG5", DG5_points, NULL,
    DOF_DG4, DOF_P5, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 56, 5, -1, 1,
    0, 0, 0, 56
};

static FLOAT DG6_points[] = {
    6.L/6.L,  0.L/6.L,  0.L/6.L,  0.L/6.L,
    0.L/6.L,  6.L/6.L,  0.L/6.L,  0.L/6.L,
    0.L/6.L,  0.L/6.L,  6.L/6.L,  0.L/6.L,
    0.L/6.L,  0.L/6.L,  0.L/6.L,  6.L/6.L,
    5.L/6.L,  1.L/6.L,  0.L/6.L,  0.L/6.L,
    4.L/6.L,  2.L/6.L,  0.L/6.L,  0.L/6.L,
    3.L/6.L,  3.L/6.L,  0.L/6.L,  0.L/6.L,
    2.L/6.L,  4.L/6.L,  0.L/6.L,  0.L/6.L,
    1.L/6.L,  5.L/6.L,  0.L/6.L,  0.L/6.L,
    5.L/6.L,  0.L/6.L,  1.L/6.L,  0.L/6.L,
    4.L/6.L,  0.L/6.L,  2.L/6.L,  0.L/6.L,
    3.L/6.L,  0.L/6.L,  3.L/6.L,  0.L/6.L,
    2.L/6.L,  0.L/6.L,  4.L/6.L,  0.L/6.L,
    1.L/6.L,  0.L/6.L,  5.L/6.L,  0.L/6.L,
    5.L/6.L,  0.L/6.L,  0.L/6.L,  1.L/6.L,
    4.L/6.L,  0.L/6.L,  0.L/6.L,  2.L/6.L,
    3.L/6.L,  0.L/6.L,  0.L/6.L,  3.L/6.L,
    2.L/6.L,  0.L/6.L,  0.L/6.L,  4.L/6.L,
    1.L/6.L,  0.L/6.L,  0.L/6.L,  5.L/6.L,
    0.L/6.L,  5.L/6.L,  1.L/6.L,  0.L/6.L,
    0.L/6.L,  4.L/6.L,  2.L/6.L,  0.L/6.L,
    0.L/6.L,  3.L/6.L,  3.L/6.L,  0.L/6.L,
    0.L/6.L,  2.L/6.L,  4.L/6.L,  0.L/6.L,
    0.L/6.L,  1.L/6.L,  5.L/6.L,  0.L/6.L,
    0.L/6.L,  5.L/6.L,  0.L/6.L,  1.L/6.L,
    0.L/6.L,  4.L/6.L,  0.L/6.L,  2.L/6.L,
    0.L/6.L,  3.L/6.L,  0.L/6.L,  3.L/6.L,
    0.L/6.L,  2.L/6.L,  0.L/6.L,  4.L/6.L,
    0.L/6.L,  1.L/6.L,  0.L/6.L,  5.L/6.L,
    0.L/6.L,  0.L/6.L,  5.L/6.L,  1.L/6.L,
    0.L/6.L,  0.L/6.L,  4.L/6.L,  2.L/6.L,
    0.L/6.L,  0.L/6.L,  3.L/6.L,  3.L/6.L,
    0.L/6.L,  0.L/6.L,  2.L/6.L,  4.L/6.L,
    0.L/6.L,  0.L/6.L,  1.L/6.L,  5.L/6.L,
    0.L/6.L,  4.L/6.L,  1.L/6.L,  1.L/6.L,
    0.L/6.L,  3.L/6.L,  2.L/6.L,  1.L/6.L,
    0.L/6.L,  3.L/6.L,  1.L/6.L,  2.L/6.L,
    0.L/6.L,  2.L/6.L,  3.L/6.L,  1.L/6.L,
    0.L/6.L,  2.L/6.L,  2.L/6.L,  2.L/6.L,
    0.L/6.L,  2.L/6.L,  1.L/6.L,  3.L/6.L,
    0.L/6.L,  1.L/6.L,  4.L/6.L,  1.L/6.L,
    0.L/6.L,  1.L/6.L,  3.L/6.L,  2.L/6.L,
    0.L/6.L,  1.L/6.L,  2.L/6.L,  3.L/6.L,
    0.L/6.L,  1.L/6.L,  1.L/6.L,  4.L/6.L,
    4.L/6.L,  0.L/6.L,  1.L/6.L,  1.L/6.L,
    3.L/6.L,  0.L/6.L,  2.L/6.L,  1.L/6.L,
    3.L/6.L,  0.L/6.L,  1.L/6.L,  2.L/6.L,
    2.L/6.L,  0.L/6.L,  3.L/6.L,  1.L/6.L,
    2.L/6.L,  0.L/6.L,  2.L/6.L,  2.L/6.L,
    2.L/6.L,  0.L/6.L,  1.L/6.L,  3.L/6.L,
    1.L/6.L,  0.L/6.L,  4.L/6.L,  1.L/6.L,
    1.L/6.L,  0.L/6.L,  3.L/6.L,  2.L/6.L,
    1.L/6.L,  0.L/6.L,  2.L/6.L,  3.L/6.L,
    1.L/6.L,  0.L/6.L,  1.L/6.L,  4.L/6.L,
    4.L/6.L,  1.L/6.L,  0.L/6.L,  1.L/6.L,
    3.L/6.L,  2.L/6.L,  0.L/6.L,  1.L/6.L,
    3.L/6.L,  1.L/6.L,  0.L/6.L,  2.L/6.L,
    2.L/6.L,  3.L/6.L,  0.L/6.L,  1.L/6.L,
    2.L/6.L,  2.L/6.L,  0.L/6.L,  2.L/6.L,
    2.L/6.L,  1.L/6.L,  0.L/6.L,  3.L/6.L,
    1.L/6.L,  4.L/6.L,  0.L/6.L,  1.L/6.L,
    1.L/6.L,  3.L/6.L,  0.L/6.L,  2.L/6.L,
    1.L/6.L,  2.L/6.L,  0.L/6.L,  3.L/6.L,
    1.L/6.L,  1.L/6.L,  0.L/6.L,  4.L/6.L,
    4.L/6.L,  1.L/6.L,  1.L/6.L,  0.L/6.L,
    3.L/6.L,  2.L/6.L,  1.L/6.L,  0.L/6.L,
    3.L/6.L,  1.L/6.L,  2.L/6.L,  0.L/6.L,
    2.L/6.L,  3.L/6.L,  1.L/6.L,  0.L/6.L,
    2.L/6.L,  2.L/6.L,  2.L/6.L,  0.L/6.L,
    2.L/6.L,  1.L/6.L,  3.L/6.L,  0.L/6.L,
    1.L/6.L,  4.L/6.L,  1.L/6.L,  0.L/6.L,
    1.L/6.L,  3.L/6.L,  2.L/6.L,  0.L/6.L,
    1.L/6.L,  2.L/6.L,  3.L/6.L,  0.L/6.L,
    1.L/6.L,  1.L/6.L,  4.L/6.L,  0.L/6.L,
    3.L/6.L,  1.L/6.L,  1.L/6.L,  1.L/6.L,
    2.L/6.L,  2.L/6.L,  1.L/6.L,  1.L/6.L,
    2.L/6.L,  1.L/6.L,  2.L/6.L,  1.L/6.L,
    2.L/6.L,  1.L/6.L,  1.L/6.L,  2.L/6.L,
    1.L/6.L,  3.L/6.L,  1.L/6.L,  1.L/6.L,
    1.L/6.L,  2.L/6.L,  2.L/6.L,  1.L/6.L,
    1.L/6.L,  2.L/6.L,  1.L/6.L,  2.L/6.L,
    1.L/6.L,  1.L/6.L,  3.L/6.L,  1.L/6.L,
    1.L/6.L,  1.L/6.L,  2.L/6.L,  2.L/6.L,
    1.L/6.L,  1.L/6.L,  1.L/6.L,  3.L/6.L
};
DOF_TYPE DOF_DG6_ = {
    DofCache, "DG6", DG6_points, NULL,
    DOF_DG5, DOF_P6, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 84, 6, -1, 1,
    0, 0, 0, 84
};

static FLOAT DG7_points[] = {
    7.L/7.L,  0.L/7.L,  0.L/7.L,  0.L/7.L,
    0.L/7.L,  7.L/7.L,  0.L/7.L,  0.L/7.L,
    0.L/7.L,  0.L/7.L,  7.L/7.L,  0.L/7.L,
    0.L/7.L,  0.L/7.L,  0.L/7.L,  7.L/7.L,
    6.L/7.L,  1.L/7.L,  0.L/7.L,  0.L/7.L,
    5.L/7.L,  2.L/7.L,  0.L/7.L,  0.L/7.L,
    4.L/7.L,  3.L/7.L,  0.L/7.L,  0.L/7.L,
    3.L/7.L,  4.L/7.L,  0.L/7.L,  0.L/7.L,
    2.L/7.L,  5.L/7.L,  0.L/7.L,  0.L/7.L,
    1.L/7.L,  6.L/7.L,  0.L/7.L,  0.L/7.L,
    6.L/7.L,  0.L/7.L,  1.L/7.L,  0.L/7.L,
    5.L/7.L,  0.L/7.L,  2.L/7.L,  0.L/7.L,
    4.L/7.L,  0.L/7.L,  3.L/7.L,  0.L/7.L,
    3.L/7.L,  0.L/7.L,  4.L/7.L,  0.L/7.L,
    2.L/7.L,  0.L/7.L,  5.L/7.L,  0.L/7.L,
    1.L/7.L,  0.L/7.L,  6.L/7.L,  0.L/7.L,
    6.L/7.L,  0.L/7.L,  0.L/7.L,  1.L/7.L,
    5.L/7.L,  0.L/7.L,  0.L/7.L,  2.L/7.L,
    4.L/7.L,  0.L/7.L,  0.L/7.L,  3.L/7.L,
    3.L/7.L,  0.L/7.L,  0.L/7.L,  4.L/7.L,
    2.L/7.L,  0.L/7.L,  0.L/7.L,  5.L/7.L,
    1.L/7.L,  0.L/7.L,  0.L/7.L,  6.L/7.L,
    0.L/7.L,  6.L/7.L,  1.L/7.L,  0.L/7.L,
    0.L/7.L,  5.L/7.L,  2.L/7.L,  0.L/7.L,
    0.L/7.L,  4.L/7.L,  3.L/7.L,  0.L/7.L,
    0.L/7.L,  3.L/7.L,  4.L/7.L,  0.L/7.L,
    0.L/7.L,  2.L/7.L,  5.L/7.L,  0.L/7.L,
    0.L/7.L,  1.L/7.L,  6.L/7.L,  0.L/7.L,
    0.L/7.L,  6.L/7.L,  0.L/7.L,  1.L/7.L,
    0.L/7.L,  5.L/7.L,  0.L/7.L,  2.L/7.L,
    0.L/7.L,  4.L/7.L,  0.L/7.L,  3.L/7.L,
    0.L/7.L,  3.L/7.L,  0.L/7.L,  4.L/7.L,
    0.L/7.L,  2.L/7.L,  0.L/7.L,  5.L/7.L,
    0.L/7.L,  1.L/7.L,  0.L/7.L,  6.L/7.L,
    0.L/7.L,  0.L/7.L,  6.L/7.L,  1.L/7.L,
    0.L/7.L,  0.L/7.L,  5.L/7.L,  2.L/7.L,
    0.L/7.L,  0.L/7.L,  4.L/7.L,  3.L/7.L,
    0.L/7.L,  0.L/7.L,  3.L/7.L,  4.L/7.L,
    0.L/7.L,  0.L/7.L,  2.L/7.L,  5.L/7.L,
    0.L/7.L,  0.L/7.L,  1.L/7.L,  6.L/7.L,
    0.L/7.L,  5.L/7.L,  1.L/7.L,  1.L/7.L,
    0.L/7.L,  4.L/7.L,  2.L/7.L,  1.L/7.L,
    0.L/7.L,  4.L/7.L,  1.L/7.L,  2.L/7.L,
    0.L/7.L,  3.L/7.L,  3.L/7.L,  1.L/7.L,
    0.L/7.L,  3.L/7.L,  2.L/7.L,  2.L/7.L,
    0.L/7.L,  3.L/7.L,  1.L/7.L,  3.L/7.L,
    0.L/7.L,  2.L/7.L,  4.L/7.L,  1.L/7.L,
    0.L/7.L,  2.L/7.L,  3.L/7.L,  2.L/7.L,
    0.L/7.L,  2.L/7.L,  2.L/7.L,  3.L/7.L,
    0.L/7.L,  2.L/7.L,  1.L/7.L,  4.L/7.L,
    0.L/7.L,  1.L/7.L,  5.L/7.L,  1.L/7.L,
    0.L/7.L,  1.L/7.L,  4.L/7.L,  2.L/7.L,
    0.L/7.L,  1.L/7.L,  3.L/7.L,  3.L/7.L,
    0.L/7.L,  1.L/7.L,  2.L/7.L,  4.L/7.L,
    0.L/7.L,  1.L/7.L,  1.L/7.L,  5.L/7.L,
    5.L/7.L,  0.L/7.L,  1.L/7.L,  1.L/7.L,
    4.L/7.L,  0.L/7.L,  2.L/7.L,  1.L/7.L,
    4.L/7.L,  0.L/7.L,  1.L/7.L,  2.L/7.L,
    3.L/7.L,  0.L/7.L,  3.L/7.L,  1.L/7.L,
    3.L/7.L,  0.L/7.L,  2.L/7.L,  2.L/7.L,
    3.L/7.L,  0.L/7.L,  1.L/7.L,  3.L/7.L,
    2.L/7.L,  0.L/7.L,  4.L/7.L,  1.L/7.L,
    2.L/7.L,  0.L/7.L,  3.L/7.L,  2.L/7.L,
    2.L/7.L,  0.L/7.L,  2.L/7.L,  3.L/7.L,
    2.L/7.L,  0.L/7.L,  1.L/7.L,  4.L/7.L,
    1.L/7.L,  0.L/7.L,  5.L/7.L,  1.L/7.L,
    1.L/7.L,  0.L/7.L,  4.L/7.L,  2.L/7.L,
    1.L/7.L,  0.L/7.L,  3.L/7.L,  3.L/7.L,
    1.L/7.L,  0.L/7.L,  2.L/7.L,  4.L/7.L,
    1.L/7.L,  0.L/7.L,  1.L/7.L,  5.L/7.L,
    5.L/7.L,  1.L/7.L,  0.L/7.L,  1.L/7.L,
    4.L/7.L,  2.L/7.L,  0.L/7.L,  1.L/7.L,
    4.L/7.L,  1.L/7.L,  0.L/7.L,  2.L/7.L,
    3.L/7.L,  3.L/7.L,  0.L/7.L,  1.L/7.L,
    3.L/7.L,  2.L/7.L,  0.L/7.L,  2.L/7.L,
    3.L/7.L,  1.L/7.L,  0.L/7.L,  3.L/7.L,
    2.L/7.L,  4.L/7.L,  0.L/7.L,  1.L/7.L,
    2.L/7.L,  3.L/7.L,  0.L/7.L,  2.L/7.L,
    2.L/7.L,  2.L/7.L,  0.L/7.L,  3.L/7.L,
    2.L/7.L,  1.L/7.L,  0.L/7.L,  4.L/7.L,
    1.L/7.L,  5.L/7.L,  0.L/7.L,  1.L/7.L,
    1.L/7.L,  4.L/7.L,  0.L/7.L,  2.L/7.L,
    1.L/7.L,  3.L/7.L,  0.L/7.L,  3.L/7.L,
    1.L/7.L,  2.L/7.L,  0.L/7.L,  4.L/7.L,
    1.L/7.L,  1.L/7.L,  0.L/7.L,  5.L/7.L,
    5.L/7.L,  1.L/7.L,  1.L/7.L,  0.L/7.L,
    4.L/7.L,  2.L/7.L,  1.L/7.L,  0.L/7.L,
    4.L/7.L,  1.L/7.L,  2.L/7.L,  0.L/7.L,
    3.L/7.L,  3.L/7.L,  1.L/7.L,  0.L/7.L,
    3.L/7.L,  2.L/7.L,  2.L/7.L,  0.L/7.L,
    3.L/7.L,  1.L/7.L,  3.L/7.L,  0.L/7.L,
    2.L/7.L,  4.L/7.L,  1.L/7.L,  0.L/7.L,
    2.L/7.L,  3.L/7.L,  2.L/7.L,  0.L/7.L,
    2.L/7.L,  2.L/7.L,  3.L/7.L,  0.L/7.L,
    2.L/7.L,  1.L/7.L,  4.L/7.L,  0.L/7.L,
    1.L/7.L,  5.L/7.L,  1.L/7.L,  0.L/7.L,
    1.L/7.L,  4.L/7.L,  2.L/7.L,  0.L/7.L,
    1.L/7.L,  3.L/7.L,  3.L/7.L,  0.L/7.L,
    1.L/7.L,  2.L/7.L,  4.L/7.L,  0.L/7.L,
    1.L/7.L,  1.L/7.L,  5.L/7.L,  0.L/7.L,
    4.L/7.L,  1.L/7.L,  1.L/7.L,  1.L/7.L,
    3.L/7.L,  2.L/7.L,  1.L/7.L,  1.L/7.L,
    3.L/7.L,  1.L/7.L,  2.L/7.L,  1.L/7.L,
    3.L/7.L,  1.L/7.L,  1.L/7.L,  2.L/7.L,
    2.L/7.L,  3.L/7.L,  1.L/7.L,  1.L/7.L,
    2.L/7.L,  2.L/7.L,  2.L/7.L,  1.L/7.L,
    2.L/7.L,  2.L/7.L,  1.L/7.L,  2.L/7.L,
    2.L/7.L,  1.L/7.L,  3.L/7.L,  1.L/7.L,
    2.L/7.L,  1.L/7.L,  2.L/7.L,  2.L/7.L,
    2.L/7.L,  1.L/7.L,  1.L/7.L,  3.L/7.L,
    1.L/7.L,  4.L/7.L,  1.L/7.L,  1.L/7.L,
    1.L/7.L,  3.L/7.L,  2.L/7.L,  1.L/7.L,
    1.L/7.L,  3.L/7.L,  1.L/7.L,  2.L/7.L,
    1.L/7.L,  2.L/7.L,  3.L/7.L,  1.L/7.L,
    1.L/7.L,  2.L/7.L,  2.L/7.L,  2.L/7.L,
    1.L/7.L,  2.L/7.L,  1.L/7.L,  3.L/7.L,
    1.L/7.L,  1.L/7.L,  4.L/7.L,  1.L/7.L,
    1.L/7.L,  1.L/7.L,  3.L/7.L,  2.L/7.L,
    1.L/7.L,  1.L/7.L,  2.L/7.L,  3.L/7.L,
    1.L/7.L,  1.L/7.L,  1.L/7.L,  4.L/7.L
};
DOF_TYPE DOF_DG7_ = {
    DofCache, "DG7", DG7_points, NULL,
    DOF_DG6, DOF_P7, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 120, 7, -1, 1,
    0, 0, 0, 120
};

static FLOAT DG8_points[] = {
    8.L/8.L,  0.L/8.L,  0.L/8.L,  0.L/8.L,
    0.L/8.L,  8.L/8.L,  0.L/8.L,  0.L/8.L,
    0.L/8.L,  0.L/8.L,  8.L/8.L,  0.L/8.L,
    0.L/8.L,  0.L/8.L,  0.L/8.L,  8.L/8.L,
    7.L/8.L,  1.L/8.L,  0.L/8.L,  0.L/8.L,
    6.L/8.L,  2.L/8.L,  0.L/8.L,  0.L/8.L,
    5.L/8.L,  3.L/8.L,  0.L/8.L,  0.L/8.L,
    4.L/8.L,  4.L/8.L,  0.L/8.L,  0.L/8.L,
    3.L/8.L,  5.L/8.L,  0.L/8.L,  0.L/8.L,
    2.L/8.L,  6.L/8.L,  0.L/8.L,  0.L/8.L,
    1.L/8.L,  7.L/8.L,  0.L/8.L,  0.L/8.L,
    7.L/8.L,  0.L/8.L,  1.L/8.L,  0.L/8.L,
    6.L/8.L,  0.L/8.L,  2.L/8.L,  0.L/8.L,
    5.L/8.L,  0.L/8.L,  3.L/8.L,  0.L/8.L,
    4.L/8.L,  0.L/8.L,  4.L/8.L,  0.L/8.L,
    3.L/8.L,  0.L/8.L,  5.L/8.L,  0.L/8.L,
    2.L/8.L,  0.L/8.L,  6.L/8.L,  0.L/8.L,
    1.L/8.L,  0.L/8.L,  7.L/8.L,  0.L/8.L,
    7.L/8.L,  0.L/8.L,  0.L/8.L,  1.L/8.L,
    6.L/8.L,  0.L/8.L,  0.L/8.L,  2.L/8.L,
    5.L/8.L,  0.L/8.L,  0.L/8.L,  3.L/8.L,
    4.L/8.L,  0.L/8.L,  0.L/8.L,  4.L/8.L,
    3.L/8.L,  0.L/8.L,  0.L/8.L,  5.L/8.L,
    2.L/8.L,  0.L/8.L,  0.L/8.L,  6.L/8.L,
    1.L/8.L,  0.L/8.L,  0.L/8.L,  7.L/8.L,
    0.L/8.L,  7.L/8.L,  1.L/8.L,  0.L/8.L,
    0.L/8.L,  6.L/8.L,  2.L/8.L,  0.L/8.L,
    0.L/8.L,  5.L/8.L,  3.L/8.L,  0.L/8.L,
    0.L/8.L,  4.L/8.L,  4.L/8.L,  0.L/8.L,
    0.L/8.L,  3.L/8.L,  5.L/8.L,  0.L/8.L,
    0.L/8.L,  2.L/8.L,  6.L/8.L,  0.L/8.L,
    0.L/8.L,  1.L/8.L,  7.L/8.L,  0.L/8.L,
    0.L/8.L,  7.L/8.L,  0.L/8.L,  1.L/8.L,
    0.L/8.L,  6.L/8.L,  0.L/8.L,  2.L/8.L,
    0.L/8.L,  5.L/8.L,  0.L/8.L,  3.L/8.L,
    0.L/8.L,  4.L/8.L,  0.L/8.L,  4.L/8.L,
    0.L/8.L,  3.L/8.L,  0.L/8.L,  5.L/8.L,
    0.L/8.L,  2.L/8.L,  0.L/8.L,  6.L/8.L,
    0.L/8.L,  1.L/8.L,  0.L/8.L,  7.L/8.L,
    0.L/8.L,  0.L/8.L,  7.L/8.L,  1.L/8.L,
    0.L/8.L,  0.L/8.L,  6.L/8.L,  2.L/8.L,
    0.L/8.L,  0.L/8.L,  5.L/8.L,  3.L/8.L,
    0.L/8.L,  0.L/8.L,  4.L/8.L,  4.L/8.L,
    0.L/8.L,  0.L/8.L,  3.L/8.L,  5.L/8.L,
    0.L/8.L,  0.L/8.L,  2.L/8.L,  6.L/8.L,
    0.L/8.L,  0.L/8.L,  1.L/8.L,  7.L/8.L,
    0.L/8.L,  6.L/8.L,  1.L/8.L,  1.L/8.L,
    0.L/8.L,  5.L/8.L,  2.L/8.L,  1.L/8.L,
    0.L/8.L,  5.L/8.L,  1.L/8.L,  2.L/8.L,
    0.L/8.L,  4.L/8.L,  3.L/8.L,  1.L/8.L,
    0.L/8.L,  4.L/8.L,  2.L/8.L,  2.L/8.L,
    0.L/8.L,  4.L/8.L,  1.L/8.L,  3.L/8.L,
    0.L/8.L,  3.L/8.L,  4.L/8.L,  1.L/8.L,
    0.L/8.L,  3.L/8.L,  3.L/8.L,  2.L/8.L,
    0.L/8.L,  3.L/8.L,  2.L/8.L,  3.L/8.L,
    0.L/8.L,  3.L/8.L,  1.L/8.L,  4.L/8.L,
    0.L/8.L,  2.L/8.L,  5.L/8.L,  1.L/8.L,
    0.L/8.L,  2.L/8.L,  4.L/8.L,  2.L/8.L,
    0.L/8.L,  2.L/8.L,  3.L/8.L,  3.L/8.L,
    0.L/8.L,  2.L/8.L,  2.L/8.L,  4.L/8.L,
    0.L/8.L,  2.L/8.L,  1.L/8.L,  5.L/8.L,
    0.L/8.L,  1.L/8.L,  6.L/8.L,  1.L/8.L,
    0.L/8.L,  1.L/8.L,  5.L/8.L,  2.L/8.L,
    0.L/8.L,  1.L/8.L,  4.L/8.L,  3.L/8.L,
    0.L/8.L,  1.L/8.L,  3.L/8.L,  4.L/8.L,
    0.L/8.L,  1.L/8.L,  2.L/8.L,  5.L/8.L,
    0.L/8.L,  1.L/8.L,  1.L/8.L,  6.L/8.L,
    6.L/8.L,  0.L/8.L,  1.L/8.L,  1.L/8.L,
    5.L/8.L,  0.L/8.L,  2.L/8.L,  1.L/8.L,
    5.L/8.L,  0.L/8.L,  1.L/8.L,  2.L/8.L,
    4.L/8.L,  0.L/8.L,  3.L/8.L,  1.L/8.L,
    4.L/8.L,  0.L/8.L,  2.L/8.L,  2.L/8.L,
    4.L/8.L,  0.L/8.L,  1.L/8.L,  3.L/8.L,
    3.L/8.L,  0.L/8.L,  4.L/8.L,  1.L/8.L,
    3.L/8.L,  0.L/8.L,  3.L/8.L,  2.L/8.L,
    3.L/8.L,  0.L/8.L,  2.L/8.L,  3.L/8.L,
    3.L/8.L,  0.L/8.L,  1.L/8.L,  4.L/8.L,
    2.L/8.L,  0.L/8.L,  5.L/8.L,  1.L/8.L,
    2.L/8.L,  0.L/8.L,  4.L/8.L,  2.L/8.L,
    2.L/8.L,  0.L/8.L,  3.L/8.L,  3.L/8.L,
    2.L/8.L,  0.L/8.L,  2.L/8.L,  4.L/8.L,
    2.L/8.L,  0.L/8.L,  1.L/8.L,  5.L/8.L,
    1.L/8.L,  0.L/8.L,  6.L/8.L,  1.L/8.L,
    1.L/8.L,  0.L/8.L,  5.L/8.L,  2.L/8.L,
    1.L/8.L,  0.L/8.L,  4.L/8.L,  3.L/8.L,
    1.L/8.L,  0.L/8.L,  3.L/8.L,  4.L/8.L,
    1.L/8.L,  0.L/8.L,  2.L/8.L,  5.L/8.L,
    1.L/8.L,  0.L/8.L,  1.L/8.L,  6.L/8.L,
    6.L/8.L,  1.L/8.L,  0.L/8.L,  1.L/8.L,
    5.L/8.L,  2.L/8.L,  0.L/8.L,  1.L/8.L,
    5.L/8.L,  1.L/8.L,  0.L/8.L,  2.L/8.L,
    4.L/8.L,  3.L/8.L,  0.L/8.L,  1.L/8.L,
    4.L/8.L,  2.L/8.L,  0.L/8.L,  2.L/8.L,
    4.L/8.L,  1.L/8.L,  0.L/8.L,  3.L/8.L,
    3.L/8.L,  4.L/8.L,  0.L/8.L,  1.L/8.L,
    3.L/8.L,  3.L/8.L,  0.L/8.L,  2.L/8.L,
    3.L/8.L,  2.L/8.L,  0.L/8.L,  3.L/8.L,
    3.L/8.L,  1.L/8.L,  0.L/8.L,  4.L/8.L,
    2.L/8.L,  5.L/8.L,  0.L/8.L,  1.L/8.L,
    2.L/8.L,  4.L/8.L,  0.L/8.L,  2.L/8.L,
    2.L/8.L,  3.L/8.L,  0.L/8.L,  3.L/8.L,
    2.L/8.L,  2.L/8.L,  0.L/8.L,  4.L/8.L,
    2.L/8.L,  1.L/8.L,  0.L/8.L,  5.L/8.L,
    1.L/8.L,  6.L/8.L,  0.L/8.L,  1.L/8.L,
    1.L/8.L,  5.L/8.L,  0.L/8.L,  2.L/8.L,
    1.L/8.L,  4.L/8.L,  0.L/8.L,  3.L/8.L,
    1.L/8.L,  3.L/8.L,  0.L/8.L,  4.L/8.L,
    1.L/8.L,  2.L/8.L,  0.L/8.L,  5.L/8.L,
    1.L/8.L,  1.L/8.L,  0.L/8.L,  6.L/8.L,
    6.L/8.L,  1.L/8.L,  1.L/8.L,  0.L/8.L,
    5.L/8.L,  2.L/8.L,  1.L/8.L,  0.L/8.L,
    5.L/8.L,  1.L/8.L,  2.L/8.L,  0.L/8.L,
    4.L/8.L,  3.L/8.L,  1.L/8.L,  0.L/8.L,
    4.L/8.L,  2.L/8.L,  2.L/8.L,  0.L/8.L,
    4.L/8.L,  1.L/8.L,  3.L/8.L,  0.L/8.L,
    3.L/8.L,  4.L/8.L,  1.L/8.L,  0.L/8.L,
    3.L/8.L,  3.L/8.L,  2.L/8.L,  0.L/8.L,
    3.L/8.L,  2.L/8.L,  3.L/8.L,  0.L/8.L,
    3.L/8.L,  1.L/8.L,  4.L/8.L,  0.L/8.L,
    2.L/8.L,  5.L/8.L,  1.L/8.L,  0.L/8.L,
    2.L/8.L,  4.L/8.L,  2.L/8.L,  0.L/8.L,
    2.L/8.L,  3.L/8.L,  3.L/8.L,  0.L/8.L,
    2.L/8.L,  2.L/8.L,  4.L/8.L,  0.L/8.L,
    2.L/8.L,  1.L/8.L,  5.L/8.L,  0.L/8.L,
    1.L/8.L,  6.L/8.L,  1.L/8.L,  0.L/8.L,
    1.L/8.L,  5.L/8.L,  2.L/8.L,  0.L/8.L,
    1.L/8.L,  4.L/8.L,  3.L/8.L,  0.L/8.L,
    1.L/8.L,  3.L/8.L,  4.L/8.L,  0.L/8.L,
    1.L/8.L,  2.L/8.L,  5.L/8.L,  0.L/8.L,
    1.L/8.L,  1.L/8.L,  6.L/8.L,  0.L/8.L,
    5.L/8.L,  1.L/8.L,  1.L/8.L,  1.L/8.L,
    4.L/8.L,  2.L/8.L,  1.L/8.L,  1.L/8.L,
    4.L/8.L,  1.L/8.L,  2.L/8.L,  1.L/8.L,
    4.L/8.L,  1.L/8.L,  1.L/8.L,  2.L/8.L,
    3.L/8.L,  3.L/8.L,  1.L/8.L,  1.L/8.L,
    3.L/8.L,  2.L/8.L,  2.L/8.L,  1.L/8.L,
    3.L/8.L,  2.L/8.L,  1.L/8.L,  2.L/8.L,
    3.L/8.L,  1.L/8.L,  3.L/8.L,  1.L/8.L,
    3.L/8.L,  1.L/8.L,  2.L/8.L,  2.L/8.L,
    3.L/8.L,  1.L/8.L,  1.L/8.L,  3.L/8.L,
    2.L/8.L,  4.L/8.L,  1.L/8.L,  1.L/8.L,
    2.L/8.L,  3.L/8.L,  2.L/8.L,  1.L/8.L,
    2.L/8.L,  3.L/8.L,  1.L/8.L,  2.L/8.L,
    2.L/8.L,  2.L/8.L,  3.L/8.L,  1.L/8.L,
    2.L/8.L,  2.L/8.L,  2.L/8.L,  2.L/8.L,
    2.L/8.L,  2.L/8.L,  1.L/8.L,  3.L/8.L,
    2.L/8.L,  1.L/8.L,  4.L/8.L,  1.L/8.L,
    2.L/8.L,  1.L/8.L,  3.L/8.L,  2.L/8.L,
    2.L/8.L,  1.L/8.L,  2.L/8.L,  3.L/8.L,
    2.L/8.L,  1.L/8.L,  1.L/8.L,  4.L/8.L,
    1.L/8.L,  5.L/8.L,  1.L/8.L,  1.L/8.L,
    1.L/8.L,  4.L/8.L,  2.L/8.L,  1.L/8.L,
    1.L/8.L,  4.L/8.L,  1.L/8.L,  2.L/8.L,
    1.L/8.L,  3.L/8.L,  3.L/8.L,  1.L/8.L,
    1.L/8.L,  3.L/8.L,  2.L/8.L,  2.L/8.L,
    1.L/8.L,  3.L/8.L,  1.L/8.L,  3.L/8.L,
    1.L/8.L,  2.L/8.L,  4.L/8.L,  1.L/8.L,
    1.L/8.L,  2.L/8.L,  3.L/8.L,  2.L/8.L,
    1.L/8.L,  2.L/8.L,  2.L/8.L,  3.L/8.L,
    1.L/8.L,  2.L/8.L,  1.L/8.L,  4.L/8.L,
    1.L/8.L,  1.L/8.L,  5.L/8.L,  1.L/8.L,
    1.L/8.L,  1.L/8.L,  4.L/8.L,  2.L/8.L,
    1.L/8.L,  1.L/8.L,  3.L/8.L,  3.L/8.L,
    1.L/8.L,  1.L/8.L,  2.L/8.L,  4.L/8.L,
    1.L/8.L,  1.L/8.L,  1.L/8.L,  5.L/8.L
};
DOF_TYPE DOF_DG8_ = {
    DofCache, "DG8", DG8_points, NULL,
    DOF_DG7, DOF_P8, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 165, 8, -1, 1,
    0, 0, 0, 165
};

static FLOAT DG9_points[] = {
    9.L/9.L,  0.L/9.L,  0.L/9.L,  0.L/9.L,
    0.L/9.L,  9.L/9.L,  0.L/9.L,  0.L/9.L,
    0.L/9.L,  0.L/9.L,  9.L/9.L,  0.L/9.L,
    0.L/9.L,  0.L/9.L,  0.L/9.L,  9.L/9.L,
    8.L/9.L,  1.L/9.L,  0.L/9.L,  0.L/9.L,
    7.L/9.L,  2.L/9.L,  0.L/9.L,  0.L/9.L,
    6.L/9.L,  3.L/9.L,  0.L/9.L,  0.L/9.L,
    5.L/9.L,  4.L/9.L,  0.L/9.L,  0.L/9.L,
    4.L/9.L,  5.L/9.L,  0.L/9.L,  0.L/9.L,
    3.L/9.L,  6.L/9.L,  0.L/9.L,  0.L/9.L,
    2.L/9.L,  7.L/9.L,  0.L/9.L,  0.L/9.L,
    1.L/9.L,  8.L/9.L,  0.L/9.L,  0.L/9.L,
    8.L/9.L,  0.L/9.L,  1.L/9.L,  0.L/9.L,
    7.L/9.L,  0.L/9.L,  2.L/9.L,  0.L/9.L,
    6.L/9.L,  0.L/9.L,  3.L/9.L,  0.L/9.L,
    5.L/9.L,  0.L/9.L,  4.L/9.L,  0.L/9.L,
    4.L/9.L,  0.L/9.L,  5.L/9.L,  0.L/9.L,
    3.L/9.L,  0.L/9.L,  6.L/9.L,  0.L/9.L,
    2.L/9.L,  0.L/9.L,  7.L/9.L,  0.L/9.L,
    1.L/9.L,  0.L/9.L,  8.L/9.L,  0.L/9.L,
    8.L/9.L,  0.L/9.L,  0.L/9.L,  1.L/9.L,
    7.L/9.L,  0.L/9.L,  0.L/9.L,  2.L/9.L,
    6.L/9.L,  0.L/9.L,  0.L/9.L,  3.L/9.L,
    5.L/9.L,  0.L/9.L,  0.L/9.L,  4.L/9.L,
    4.L/9.L,  0.L/9.L,  0.L/9.L,  5.L/9.L,
    3.L/9.L,  0.L/9.L,  0.L/9.L,  6.L/9.L,
    2.L/9.L,  0.L/9.L,  0.L/9.L,  7.L/9.L,
    1.L/9.L,  0.L/9.L,  0.L/9.L,  8.L/9.L,
    0.L/9.L,  8.L/9.L,  1.L/9.L,  0.L/9.L,
    0.L/9.L,  7.L/9.L,  2.L/9.L,  0.L/9.L,
    0.L/9.L,  6.L/9.L,  3.L/9.L,  0.L/9.L,
    0.L/9.L,  5.L/9.L,  4.L/9.L,  0.L/9.L,
    0.L/9.L,  4.L/9.L,  5.L/9.L,  0.L/9.L,
    0.L/9.L,  3.L/9.L,  6.L/9.L,  0.L/9.L,
    0.L/9.L,  2.L/9.L,  7.L/9.L,  0.L/9.L,
    0.L/9.L,  1.L/9.L,  8.L/9.L,  0.L/9.L,
    0.L/9.L,  8.L/9.L,  0.L/9.L,  1.L/9.L,
    0.L/9.L,  7.L/9.L,  0.L/9.L,  2.L/9.L,
    0.L/9.L,  6.L/9.L,  0.L/9.L,  3.L/9.L,
    0.L/9.L,  5.L/9.L,  0.L/9.L,  4.L/9.L,
    0.L/9.L,  4.L/9.L,  0.L/9.L,  5.L/9.L,
    0.L/9.L,  3.L/9.L,  0.L/9.L,  6.L/9.L,
    0.L/9.L,  2.L/9.L,  0.L/9.L,  7.L/9.L,
    0.L/9.L,  1.L/9.L,  0.L/9.L,  8.L/9.L,
    0.L/9.L,  0.L/9.L,  8.L/9.L,  1.L/9.L,
    0.L/9.L,  0.L/9.L,  7.L/9.L,  2.L/9.L,
    0.L/9.L,  0.L/9.L,  6.L/9.L,  3.L/9.L,
    0.L/9.L,  0.L/9.L,  5.L/9.L,  4.L/9.L,
    0.L/9.L,  0.L/9.L,  4.L/9.L,  5.L/9.L,
    0.L/9.L,  0.L/9.L,  3.L/9.L,  6.L/9.L,
    0.L/9.L,  0.L/9.L,  2.L/9.L,  7.L/9.L,
    0.L/9.L,  0.L/9.L,  1.L/9.L,  8.L/9.L,
    0.L/9.L,  7.L/9.L,  1.L/9.L,  1.L/9.L,
    0.L/9.L,  6.L/9.L,  2.L/9.L,  1.L/9.L,
    0.L/9.L,  6.L/9.L,  1.L/9.L,  2.L/9.L,
    0.L/9.L,  5.L/9.L,  3.L/9.L,  1.L/9.L,
    0.L/9.L,  5.L/9.L,  2.L/9.L,  2.L/9.L,
    0.L/9.L,  5.L/9.L,  1.L/9.L,  3.L/9.L,
    0.L/9.L,  4.L/9.L,  4.L/9.L,  1.L/9.L,
    0.L/9.L,  4.L/9.L,  3.L/9.L,  2.L/9.L,
    0.L/9.L,  4.L/9.L,  2.L/9.L,  3.L/9.L,
    0.L/9.L,  4.L/9.L,  1.L/9.L,  4.L/9.L,
    0.L/9.L,  3.L/9.L,  5.L/9.L,  1.L/9.L,
    0.L/9.L,  3.L/9.L,  4.L/9.L,  2.L/9.L,
    0.L/9.L,  3.L/9.L,  3.L/9.L,  3.L/9.L,
    0.L/9.L,  3.L/9.L,  2.L/9.L,  4.L/9.L,
    0.L/9.L,  3.L/9.L,  1.L/9.L,  5.L/9.L,
    0.L/9.L,  2.L/9.L,  6.L/9.L,  1.L/9.L,
    0.L/9.L,  2.L/9.L,  5.L/9.L,  2.L/9.L,
    0.L/9.L,  2.L/9.L,  4.L/9.L,  3.L/9.L,
    0.L/9.L,  2.L/9.L,  3.L/9.L,  4.L/9.L,
    0.L/9.L,  2.L/9.L,  2.L/9.L,  5.L/9.L,
    0.L/9.L,  2.L/9.L,  1.L/9.L,  6.L/9.L,
    0.L/9.L,  1.L/9.L,  7.L/9.L,  1.L/9.L,
    0.L/9.L,  1.L/9.L,  6.L/9.L,  2.L/9.L,
    0.L/9.L,  1.L/9.L,  5.L/9.L,  3.L/9.L,
    0.L/9.L,  1.L/9.L,  4.L/9.L,  4.L/9.L,
    0.L/9.L,  1.L/9.L,  3.L/9.L,  5.L/9.L,
    0.L/9.L,  1.L/9.L,  2.L/9.L,  6.L/9.L,
    0.L/9.L,  1.L/9.L,  1.L/9.L,  7.L/9.L,
    7.L/9.L,  0.L/9.L,  1.L/9.L,  1.L/9.L,
    6.L/9.L,  0.L/9.L,  2.L/9.L,  1.L/9.L,
    6.L/9.L,  0.L/9.L,  1.L/9.L,  2.L/9.L,
    5.L/9.L,  0.L/9.L,  3.L/9.L,  1.L/9.L,
    5.L/9.L,  0.L/9.L,  2.L/9.L,  2.L/9.L,
    5.L/9.L,  0.L/9.L,  1.L/9.L,  3.L/9.L,
    4.L/9.L,  0.L/9.L,  4.L/9.L,  1.L/9.L,
    4.L/9.L,  0.L/9.L,  3.L/9.L,  2.L/9.L,
    4.L/9.L,  0.L/9.L,  2.L/9.L,  3.L/9.L,
    4.L/9.L,  0.L/9.L,  1.L/9.L,  4.L/9.L,
    3.L/9.L,  0.L/9.L,  5.L/9.L,  1.L/9.L,
    3.L/9.L,  0.L/9.L,  4.L/9.L,  2.L/9.L,
    3.L/9.L,  0.L/9.L,  3.L/9.L,  3.L/9.L,
    3.L/9.L,  0.L/9.L,  2.L/9.L,  4.L/9.L,
    3.L/9.L,  0.L/9.L,  1.L/9.L,  5.L/9.L,
    2.L/9.L,  0.L/9.L,  6.L/9.L,  1.L/9.L,
    2.L/9.L,  0.L/9.L,  5.L/9.L,  2.L/9.L,
    2.L/9.L,  0.L/9.L,  4.L/9.L,  3.L/9.L,
    2.L/9.L,  0.L/9.L,  3.L/9.L,  4.L/9.L,
    2.L/9.L,  0.L/9.L,  2.L/9.L,  5.L/9.L,
    2.L/9.L,  0.L/9.L,  1.L/9.L,  6.L/9.L,
    1.L/9.L,  0.L/9.L,  7.L/9.L,  1.L/9.L,
    1.L/9.L,  0.L/9.L,  6.L/9.L,  2.L/9.L,
    1.L/9.L,  0.L/9.L,  5.L/9.L,  3.L/9.L,
    1.L/9.L,  0.L/9.L,  4.L/9.L,  4.L/9.L,
    1.L/9.L,  0.L/9.L,  3.L/9.L,  5.L/9.L,
    1.L/9.L,  0.L/9.L,  2.L/9.L,  6.L/9.L,
    1.L/9.L,  0.L/9.L,  1.L/9.L,  7.L/9.L,
    7.L/9.L,  1.L/9.L,  0.L/9.L,  1.L/9.L,
    6.L/9.L,  2.L/9.L,  0.L/9.L,  1.L/9.L,
    6.L/9.L,  1.L/9.L,  0.L/9.L,  2.L/9.L,
    5.L/9.L,  3.L/9.L,  0.L/9.L,  1.L/9.L,
    5.L/9.L,  2.L/9.L,  0.L/9.L,  2.L/9.L,
    5.L/9.L,  1.L/9.L,  0.L/9.L,  3.L/9.L,
    4.L/9.L,  4.L/9.L,  0.L/9.L,  1.L/9.L,
    4.L/9.L,  3.L/9.L,  0.L/9.L,  2.L/9.L,
    4.L/9.L,  2.L/9.L,  0.L/9.L,  3.L/9.L,
    4.L/9.L,  1.L/9.L,  0.L/9.L,  4.L/9.L,
    3.L/9.L,  5.L/9.L,  0.L/9.L,  1.L/9.L,
    3.L/9.L,  4.L/9.L,  0.L/9.L,  2.L/9.L,
    3.L/9.L,  3.L/9.L,  0.L/9.L,  3.L/9.L,
    3.L/9.L,  2.L/9.L,  0.L/9.L,  4.L/9.L,
    3.L/9.L,  1.L/9.L,  0.L/9.L,  5.L/9.L,
    2.L/9.L,  6.L/9.L,  0.L/9.L,  1.L/9.L,
    2.L/9.L,  5.L/9.L,  0.L/9.L,  2.L/9.L,
    2.L/9.L,  4.L/9.L,  0.L/9.L,  3.L/9.L,
    2.L/9.L,  3.L/9.L,  0.L/9.L,  4.L/9.L,
    2.L/9.L,  2.L/9.L,  0.L/9.L,  5.L/9.L,
    2.L/9.L,  1.L/9.L,  0.L/9.L,  6.L/9.L,
    1.L/9.L,  7.L/9.L,  0.L/9.L,  1.L/9.L,
    1.L/9.L,  6.L/9.L,  0.L/9.L,  2.L/9.L,
    1.L/9.L,  5.L/9.L,  0.L/9.L,  3.L/9.L,
    1.L/9.L,  4.L/9.L,  0.L/9.L,  4.L/9.L,
    1.L/9.L,  3.L/9.L,  0.L/9.L,  5.L/9.L,
    1.L/9.L,  2.L/9.L,  0.L/9.L,  6.L/9.L,
    1.L/9.L,  1.L/9.L,  0.L/9.L,  7.L/9.L,
    7.L/9.L,  1.L/9.L,  1.L/9.L,  0.L/9.L,
    6.L/9.L,  2.L/9.L,  1.L/9.L,  0.L/9.L,
    6.L/9.L,  1.L/9.L,  2.L/9.L,  0.L/9.L,
    5.L/9.L,  3.L/9.L,  1.L/9.L,  0.L/9.L,
    5.L/9.L,  2.L/9.L,  2.L/9.L,  0.L/9.L,
    5.L/9.L,  1.L/9.L,  3.L/9.L,  0.L/9.L,
    4.L/9.L,  4.L/9.L,  1.L/9.L,  0.L/9.L,
    4.L/9.L,  3.L/9.L,  2.L/9.L,  0.L/9.L,
    4.L/9.L,  2.L/9.L,  3.L/9.L,  0.L/9.L,
    4.L/9.L,  1.L/9.L,  4.L/9.L,  0.L/9.L,
    3.L/9.L,  5.L/9.L,  1.L/9.L,  0.L/9.L,
    3.L/9.L,  4.L/9.L,  2.L/9.L,  0.L/9.L,
    3.L/9.L,  3.L/9.L,  3.L/9.L,  0.L/9.L,
    3.L/9.L,  2.L/9.L,  4.L/9.L,  0.L/9.L,
    3.L/9.L,  1.L/9.L,  5.L/9.L,  0.L/9.L,
    2.L/9.L,  6.L/9.L,  1.L/9.L,  0.L/9.L,
    2.L/9.L,  5.L/9.L,  2.L/9.L,  0.L/9.L,
    2.L/9.L,  4.L/9.L,  3.L/9.L,  0.L/9.L,
    2.L/9.L,  3.L/9.L,  4.L/9.L,  0.L/9.L,
    2.L/9.L,  2.L/9.L,  5.L/9.L,  0.L/9.L,
    2.L/9.L,  1.L/9.L,  6.L/9.L,  0.L/9.L,
    1.L/9.L,  7.L/9.L,  1.L/9.L,  0.L/9.L,
    1.L/9.L,  6.L/9.L,  2.L/9.L,  0.L/9.L,
    1.L/9.L,  5.L/9.L,  3.L/9.L,  0.L/9.L,
    1.L/9.L,  4.L/9.L,  4.L/9.L,  0.L/9.L,
    1.L/9.L,  3.L/9.L,  5.L/9.L,  0.L/9.L,
    1.L/9.L,  2.L/9.L,  6.L/9.L,  0.L/9.L,
    1.L/9.L,  1.L/9.L,  7.L/9.L,  0.L/9.L,
    6.L/9.L,  1.L/9.L,  1.L/9.L,  1.L/9.L,
    5.L/9.L,  2.L/9.L,  1.L/9.L,  1.L/9.L,
    5.L/9.L,  1.L/9.L,  2.L/9.L,  1.L/9.L,
    5.L/9.L,  1.L/9.L,  1.L/9.L,  2.L/9.L,
    4.L/9.L,  3.L/9.L,  1.L/9.L,  1.L/9.L,
    4.L/9.L,  2.L/9.L,  2.L/9.L,  1.L/9.L,
    4.L/9.L,  2.L/9.L,  1.L/9.L,  2.L/9.L,
    4.L/9.L,  1.L/9.L,  3.L/9.L,  1.L/9.L,
    4.L/9.L,  1.L/9.L,  2.L/9.L,  2.L/9.L,
    4.L/9.L,  1.L/9.L,  1.L/9.L,  3.L/9.L,
    3.L/9.L,  4.L/9.L,  1.L/9.L,  1.L/9.L,
    3.L/9.L,  3.L/9.L,  2.L/9.L,  1.L/9.L,
    3.L/9.L,  3.L/9.L,  1.L/9.L,  2.L/9.L,
    3.L/9.L,  2.L/9.L,  3.L/9.L,  1.L/9.L,
    3.L/9.L,  2.L/9.L,  2.L/9.L,  2.L/9.L,
    3.L/9.L,  2.L/9.L,  1.L/9.L,  3.L/9.L,
    3.L/9.L,  1.L/9.L,  4.L/9.L,  1.L/9.L,
    3.L/9.L,  1.L/9.L,  3.L/9.L,  2.L/9.L,
    3.L/9.L,  1.L/9.L,  2.L/9.L,  3.L/9.L,
    3.L/9.L,  1.L/9.L,  1.L/9.L,  4.L/9.L,
    2.L/9.L,  5.L/9.L,  1.L/9.L,  1.L/9.L,
    2.L/9.L,  4.L/9.L,  2.L/9.L,  1.L/9.L,
    2.L/9.L,  4.L/9.L,  1.L/9.L,  2.L/9.L,
    2.L/9.L,  3.L/9.L,  3.L/9.L,  1.L/9.L,
    2.L/9.L,  3.L/9.L,  2.L/9.L,  2.L/9.L,
    2.L/9.L,  3.L/9.L,  1.L/9.L,  3.L/9.L,
    2.L/9.L,  2.L/9.L,  4.L/9.L,  1.L/9.L,
    2.L/9.L,  2.L/9.L,  3.L/9.L,  2.L/9.L,
    2.L/9.L,  2.L/9.L,  2.L/9.L,  3.L/9.L,
    2.L/9.L,  2.L/9.L,  1.L/9.L,  4.L/9.L,
    2.L/9.L,  1.L/9.L,  5.L/9.L,  1.L/9.L,
    2.L/9.L,  1.L/9.L,  4.L/9.L,  2.L/9.L,
    2.L/9.L,  1.L/9.L,  3.L/9.L,  3.L/9.L,
    2.L/9.L,  1.L/9.L,  2.L/9.L,  4.L/9.L,
    2.L/9.L,  1.L/9.L,  1.L/9.L,  5.L/9.L,
    1.L/9.L,  6.L/9.L,  1.L/9.L,  1.L/9.L,
    1.L/9.L,  5.L/9.L,  2.L/9.L,  1.L/9.L,
    1.L/9.L,  5.L/9.L,  1.L/9.L,  2.L/9.L,
    1.L/9.L,  4.L/9.L,  3.L/9.L,  1.L/9.L,
    1.L/9.L,  4.L/9.L,  2.L/9.L,  2.L/9.L,
    1.L/9.L,  4.L/9.L,  1.L/9.L,  3.L/9.L,
    1.L/9.L,  3.L/9.L,  4.L/9.L,  1.L/9.L,
    1.L/9.L,  3.L/9.L,  3.L/9.L,  2.L/9.L,
    1.L/9.L,  3.L/9.L,  2.L/9.L,  3.L/9.L,
    1.L/9.L,  3.L/9.L,  1.L/9.L,  4.L/9.L,
    1.L/9.L,  2.L/9.L,  5.L/9.L,  1.L/9.L,
    1.L/9.L,  2.L/9.L,  4.L/9.L,  2.L/9.L,
    1.L/9.L,  2.L/9.L,  3.L/9.L,  3.L/9.L,
    1.L/9.L,  2.L/9.L,  2.L/9.L,  4.L/9.L,
    1.L/9.L,  2.L/9.L,  1.L/9.L,  5.L/9.L,
    1.L/9.L,  1.L/9.L,  6.L/9.L,  1.L/9.L,
    1.L/9.L,  1.L/9.L,  5.L/9.L,  2.L/9.L,
    1.L/9.L,  1.L/9.L,  4.L/9.L,  3.L/9.L,
    1.L/9.L,  1.L/9.L,  3.L/9.L,  4.L/9.L,
    1.L/9.L,  1.L/9.L,  2.L/9.L,  5.L/9.L,
    1.L/9.L,  1.L/9.L,  1.L/9.L,  6.L/9.L
};
DOF_TYPE DOF_DG9_ = {
    DofCache, "DG9", DG9_points, NULL,
    DOF_DG8, DOF_P9, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 220, 9, -1, 1,
    0, 0, 0, 220
};

static FLOAT DG10_points[] = {
   10.L/10.L,  0.L/10.L,  0.L/10.L,  0.L/10.L,
    0.L/10.L, 10.L/10.L,  0.L/10.L,  0.L/10.L,
    0.L/10.L,  0.L/10.L, 10.L/10.L,  0.L/10.L,
    0.L/10.L,  0.L/10.L,  0.L/10.L, 10.L/10.L,
    9.L/10.L,  1.L/10.L,  0.L/10.L,  0.L/10.L,
    8.L/10.L,  2.L/10.L,  0.L/10.L,  0.L/10.L,
    7.L/10.L,  3.L/10.L,  0.L/10.L,  0.L/10.L,
    6.L/10.L,  4.L/10.L,  0.L/10.L,  0.L/10.L,
    5.L/10.L,  5.L/10.L,  0.L/10.L,  0.L/10.L,
    4.L/10.L,  6.L/10.L,  0.L/10.L,  0.L/10.L,
    3.L/10.L,  7.L/10.L,  0.L/10.L,  0.L/10.L,
    2.L/10.L,  8.L/10.L,  0.L/10.L,  0.L/10.L,
    1.L/10.L,  9.L/10.L,  0.L/10.L,  0.L/10.L,
    9.L/10.L,  0.L/10.L,  1.L/10.L,  0.L/10.L,
    8.L/10.L,  0.L/10.L,  2.L/10.L,  0.L/10.L,
    7.L/10.L,  0.L/10.L,  3.L/10.L,  0.L/10.L,
    6.L/10.L,  0.L/10.L,  4.L/10.L,  0.L/10.L,
    5.L/10.L,  0.L/10.L,  5.L/10.L,  0.L/10.L,
    4.L/10.L,  0.L/10.L,  6.L/10.L,  0.L/10.L,
    3.L/10.L,  0.L/10.L,  7.L/10.L,  0.L/10.L,
    2.L/10.L,  0.L/10.L,  8.L/10.L,  0.L/10.L,
    1.L/10.L,  0.L/10.L,  9.L/10.L,  0.L/10.L,
    9.L/10.L,  0.L/10.L,  0.L/10.L,  1.L/10.L,
    8.L/10.L,  0.L/10.L,  0.L/10.L,  2.L/10.L,
    7.L/10.L,  0.L/10.L,  0.L/10.L,  3.L/10.L,
    6.L/10.L,  0.L/10.L,  0.L/10.L,  4.L/10.L,
    5.L/10.L,  0.L/10.L,  0.L/10.L,  5.L/10.L,
    4.L/10.L,  0.L/10.L,  0.L/10.L,  6.L/10.L,
    3.L/10.L,  0.L/10.L,  0.L/10.L,  7.L/10.L,
    2.L/10.L,  0.L/10.L,  0.L/10.L,  8.L/10.L,
    1.L/10.L,  0.L/10.L,  0.L/10.L,  9.L/10.L,
    0.L/10.L,  9.L/10.L,  1.L/10.L,  0.L/10.L,
    0.L/10.L,  8.L/10.L,  2.L/10.L,  0.L/10.L,
    0.L/10.L,  7.L/10.L,  3.L/10.L,  0.L/10.L,
    0.L/10.L,  6.L/10.L,  4.L/10.L,  0.L/10.L,
    0.L/10.L,  5.L/10.L,  5.L/10.L,  0.L/10.L,
    0.L/10.L,  4.L/10.L,  6.L/10.L,  0.L/10.L,
    0.L/10.L,  3.L/10.L,  7.L/10.L,  0.L/10.L,
    0.L/10.L,  2.L/10.L,  8.L/10.L,  0.L/10.L,
    0.L/10.L,  1.L/10.L,  9.L/10.L,  0.L/10.L,
    0.L/10.L,  9.L/10.L,  0.L/10.L,  1.L/10.L,
    0.L/10.L,  8.L/10.L,  0.L/10.L,  2.L/10.L,
    0.L/10.L,  7.L/10.L,  0.L/10.L,  3.L/10.L,
    0.L/10.L,  6.L/10.L,  0.L/10.L,  4.L/10.L,
    0.L/10.L,  5.L/10.L,  0.L/10.L,  5.L/10.L,
    0.L/10.L,  4.L/10.L,  0.L/10.L,  6.L/10.L,
    0.L/10.L,  3.L/10.L,  0.L/10.L,  7.L/10.L,
    0.L/10.L,  2.L/10.L,  0.L/10.L,  8.L/10.L,
    0.L/10.L,  1.L/10.L,  0.L/10.L,  9.L/10.L,
    0.L/10.L,  0.L/10.L,  9.L/10.L,  1.L/10.L,
    0.L/10.L,  0.L/10.L,  8.L/10.L,  2.L/10.L,
    0.L/10.L,  0.L/10.L,  7.L/10.L,  3.L/10.L,
    0.L/10.L,  0.L/10.L,  6.L/10.L,  4.L/10.L,
    0.L/10.L,  0.L/10.L,  5.L/10.L,  5.L/10.L,
    0.L/10.L,  0.L/10.L,  4.L/10.L,  6.L/10.L,
    0.L/10.L,  0.L/10.L,  3.L/10.L,  7.L/10.L,
    0.L/10.L,  0.L/10.L,  2.L/10.L,  8.L/10.L,
    0.L/10.L,  0.L/10.L,  1.L/10.L,  9.L/10.L,
    0.L/10.L,  8.L/10.L,  1.L/10.L,  1.L/10.L,
    0.L/10.L,  7.L/10.L,  2.L/10.L,  1.L/10.L,
    0.L/10.L,  7.L/10.L,  1.L/10.L,  2.L/10.L,
    0.L/10.L,  6.L/10.L,  3.L/10.L,  1.L/10.L,
    0.L/10.L,  6.L/10.L,  2.L/10.L,  2.L/10.L,
    0.L/10.L,  6.L/10.L,  1.L/10.L,  3.L/10.L,
    0.L/10.L,  5.L/10.L,  4.L/10.L,  1.L/10.L,
    0.L/10.L,  5.L/10.L,  3.L/10.L,  2.L/10.L,
    0.L/10.L,  5.L/10.L,  2.L/10.L,  3.L/10.L,
    0.L/10.L,  5.L/10.L,  1.L/10.L,  4.L/10.L,
    0.L/10.L,  4.L/10.L,  5.L/10.L,  1.L/10.L,
    0.L/10.L,  4.L/10.L,  4.L/10.L,  2.L/10.L,
    0.L/10.L,  4.L/10.L,  3.L/10.L,  3.L/10.L,
    0.L/10.L,  4.L/10.L,  2.L/10.L,  4.L/10.L,
    0.L/10.L,  4.L/10.L,  1.L/10.L,  5.L/10.L,
    0.L/10.L,  3.L/10.L,  6.L/10.L,  1.L/10.L,
    0.L/10.L,  3.L/10.L,  5.L/10.L,  2.L/10.L,
    0.L/10.L,  3.L/10.L,  4.L/10.L,  3.L/10.L,
    0.L/10.L,  3.L/10.L,  3.L/10.L,  4.L/10.L,
    0.L/10.L,  3.L/10.L,  2.L/10.L,  5.L/10.L,
    0.L/10.L,  3.L/10.L,  1.L/10.L,  6.L/10.L,
    0.L/10.L,  2.L/10.L,  7.L/10.L,  1.L/10.L,
    0.L/10.L,  2.L/10.L,  6.L/10.L,  2.L/10.L,
    0.L/10.L,  2.L/10.L,  5.L/10.L,  3.L/10.L,
    0.L/10.L,  2.L/10.L,  4.L/10.L,  4.L/10.L,
    0.L/10.L,  2.L/10.L,  3.L/10.L,  5.L/10.L,
    0.L/10.L,  2.L/10.L,  2.L/10.L,  6.L/10.L,
    0.L/10.L,  2.L/10.L,  1.L/10.L,  7.L/10.L,
    0.L/10.L,  1.L/10.L,  8.L/10.L,  1.L/10.L,
    0.L/10.L,  1.L/10.L,  7.L/10.L,  2.L/10.L,
    0.L/10.L,  1.L/10.L,  6.L/10.L,  3.L/10.L,
    0.L/10.L,  1.L/10.L,  5.L/10.L,  4.L/10.L,
    0.L/10.L,  1.L/10.L,  4.L/10.L,  5.L/10.L,
    0.L/10.L,  1.L/10.L,  3.L/10.L,  6.L/10.L,
    0.L/10.L,  1.L/10.L,  2.L/10.L,  7.L/10.L,
    0.L/10.L,  1.L/10.L,  1.L/10.L,  8.L/10.L,
    8.L/10.L,  0.L/10.L,  1.L/10.L,  1.L/10.L,
    7.L/10.L,  0.L/10.L,  2.L/10.L,  1.L/10.L,
    7.L/10.L,  0.L/10.L,  1.L/10.L,  2.L/10.L,
    6.L/10.L,  0.L/10.L,  3.L/10.L,  1.L/10.L,
    6.L/10.L,  0.L/10.L,  2.L/10.L,  2.L/10.L,
    6.L/10.L,  0.L/10.L,  1.L/10.L,  3.L/10.L,
    5.L/10.L,  0.L/10.L,  4.L/10.L,  1.L/10.L,
    5.L/10.L,  0.L/10.L,  3.L/10.L,  2.L/10.L,
    5.L/10.L,  0.L/10.L,  2.L/10.L,  3.L/10.L,
    5.L/10.L,  0.L/10.L,  1.L/10.L,  4.L/10.L,
    4.L/10.L,  0.L/10.L,  5.L/10.L,  1.L/10.L,
    4.L/10.L,  0.L/10.L,  4.L/10.L,  2.L/10.L,
    4.L/10.L,  0.L/10.L,  3.L/10.L,  3.L/10.L,
    4.L/10.L,  0.L/10.L,  2.L/10.L,  4.L/10.L,
    4.L/10.L,  0.L/10.L,  1.L/10.L,  5.L/10.L,
    3.L/10.L,  0.L/10.L,  6.L/10.L,  1.L/10.L,
    3.L/10.L,  0.L/10.L,  5.L/10.L,  2.L/10.L,
    3.L/10.L,  0.L/10.L,  4.L/10.L,  3.L/10.L,
    3.L/10.L,  0.L/10.L,  3.L/10.L,  4.L/10.L,
    3.L/10.L,  0.L/10.L,  2.L/10.L,  5.L/10.L,
    3.L/10.L,  0.L/10.L,  1.L/10.L,  6.L/10.L,
    2.L/10.L,  0.L/10.L,  7.L/10.L,  1.L/10.L,
    2.L/10.L,  0.L/10.L,  6.L/10.L,  2.L/10.L,
    2.L/10.L,  0.L/10.L,  5.L/10.L,  3.L/10.L,
    2.L/10.L,  0.L/10.L,  4.L/10.L,  4.L/10.L,
    2.L/10.L,  0.L/10.L,  3.L/10.L,  5.L/10.L,
    2.L/10.L,  0.L/10.L,  2.L/10.L,  6.L/10.L,
    2.L/10.L,  0.L/10.L,  1.L/10.L,  7.L/10.L,
    1.L/10.L,  0.L/10.L,  8.L/10.L,  1.L/10.L,
    1.L/10.L,  0.L/10.L,  7.L/10.L,  2.L/10.L,
    1.L/10.L,  0.L/10.L,  6.L/10.L,  3.L/10.L,
    1.L/10.L,  0.L/10.L,  5.L/10.L,  4.L/10.L,
    1.L/10.L,  0.L/10.L,  4.L/10.L,  5.L/10.L,
    1.L/10.L,  0.L/10.L,  3.L/10.L,  6.L/10.L,
    1.L/10.L,  0.L/10.L,  2.L/10.L,  7.L/10.L,
    1.L/10.L,  0.L/10.L,  1.L/10.L,  8.L/10.L,
    8.L/10.L,  1.L/10.L,  0.L/10.L,  1.L/10.L,
    7.L/10.L,  2.L/10.L,  0.L/10.L,  1.L/10.L,
    7.L/10.L,  1.L/10.L,  0.L/10.L,  2.L/10.L,
    6.L/10.L,  3.L/10.L,  0.L/10.L,  1.L/10.L,
    6.L/10.L,  2.L/10.L,  0.L/10.L,  2.L/10.L,
    6.L/10.L,  1.L/10.L,  0.L/10.L,  3.L/10.L,
    5.L/10.L,  4.L/10.L,  0.L/10.L,  1.L/10.L,
    5.L/10.L,  3.L/10.L,  0.L/10.L,  2.L/10.L,
    5.L/10.L,  2.L/10.L,  0.L/10.L,  3.L/10.L,
    5.L/10.L,  1.L/10.L,  0.L/10.L,  4.L/10.L,
    4.L/10.L,  5.L/10.L,  0.L/10.L,  1.L/10.L,
    4.L/10.L,  4.L/10.L,  0.L/10.L,  2.L/10.L,
    4.L/10.L,  3.L/10.L,  0.L/10.L,  3.L/10.L,
    4.L/10.L,  2.L/10.L,  0.L/10.L,  4.L/10.L,
    4.L/10.L,  1.L/10.L,  0.L/10.L,  5.L/10.L,
    3.L/10.L,  6.L/10.L,  0.L/10.L,  1.L/10.L,
    3.L/10.L,  5.L/10.L,  0.L/10.L,  2.L/10.L,
    3.L/10.L,  4.L/10.L,  0.L/10.L,  3.L/10.L,
    3.L/10.L,  3.L/10.L,  0.L/10.L,  4.L/10.L,
    3.L/10.L,  2.L/10.L,  0.L/10.L,  5.L/10.L,
    3.L/10.L,  1.L/10.L,  0.L/10.L,  6.L/10.L,
    2.L/10.L,  7.L/10.L,  0.L/10.L,  1.L/10.L,
    2.L/10.L,  6.L/10.L,  0.L/10.L,  2.L/10.L,
    2.L/10.L,  5.L/10.L,  0.L/10.L,  3.L/10.L,
    2.L/10.L,  4.L/10.L,  0.L/10.L,  4.L/10.L,
    2.L/10.L,  3.L/10.L,  0.L/10.L,  5.L/10.L,
    2.L/10.L,  2.L/10.L,  0.L/10.L,  6.L/10.L,
    2.L/10.L,  1.L/10.L,  0.L/10.L,  7.L/10.L,
    1.L/10.L,  8.L/10.L,  0.L/10.L,  1.L/10.L,
    1.L/10.L,  7.L/10.L,  0.L/10.L,  2.L/10.L,
    1.L/10.L,  6.L/10.L,  0.L/10.L,  3.L/10.L,
    1.L/10.L,  5.L/10.L,  0.L/10.L,  4.L/10.L,
    1.L/10.L,  4.L/10.L,  0.L/10.L,  5.L/10.L,
    1.L/10.L,  3.L/10.L,  0.L/10.L,  6.L/10.L,
    1.L/10.L,  2.L/10.L,  0.L/10.L,  7.L/10.L,
    1.L/10.L,  1.L/10.L,  0.L/10.L,  8.L/10.L,
    8.L/10.L,  1.L/10.L,  1.L/10.L,  0.L/10.L,
    7.L/10.L,  2.L/10.L,  1.L/10.L,  0.L/10.L,
    7.L/10.L,  1.L/10.L,  2.L/10.L,  0.L/10.L,
    6.L/10.L,  3.L/10.L,  1.L/10.L,  0.L/10.L,
    6.L/10.L,  2.L/10.L,  2.L/10.L,  0.L/10.L,
    6.L/10.L,  1.L/10.L,  3.L/10.L,  0.L/10.L,
    5.L/10.L,  4.L/10.L,  1.L/10.L,  0.L/10.L,
    5.L/10.L,  3.L/10.L,  2.L/10.L,  0.L/10.L,
    5.L/10.L,  2.L/10.L,  3.L/10.L,  0.L/10.L,
    5.L/10.L,  1.L/10.L,  4.L/10.L,  0.L/10.L,
    4.L/10.L,  5.L/10.L,  1.L/10.L,  0.L/10.L,
    4.L/10.L,  4.L/10.L,  2.L/10.L,  0.L/10.L,
    4.L/10.L,  3.L/10.L,  3.L/10.L,  0.L/10.L,
    4.L/10.L,  2.L/10.L,  4.L/10.L,  0.L/10.L,
    4.L/10.L,  1.L/10.L,  5.L/10.L,  0.L/10.L,
    3.L/10.L,  6.L/10.L,  1.L/10.L,  0.L/10.L,
    3.L/10.L,  5.L/10.L,  2.L/10.L,  0.L/10.L,
    3.L/10.L,  4.L/10.L,  3.L/10.L,  0.L/10.L,
    3.L/10.L,  3.L/10.L,  4.L/10.L,  0.L/10.L,
    3.L/10.L,  2.L/10.L,  5.L/10.L,  0.L/10.L,
    3.L/10.L,  1.L/10.L,  6.L/10.L,  0.L/10.L,
    2.L/10.L,  7.L/10.L,  1.L/10.L,  0.L/10.L,
    2.L/10.L,  6.L/10.L,  2.L/10.L,  0.L/10.L,
    2.L/10.L,  5.L/10.L,  3.L/10.L,  0.L/10.L,
    2.L/10.L,  4.L/10.L,  4.L/10.L,  0.L/10.L,
    2.L/10.L,  3.L/10.L,  5.L/10.L,  0.L/10.L,
    2.L/10.L,  2.L/10.L,  6.L/10.L,  0.L/10.L,
    2.L/10.L,  1.L/10.L,  7.L/10.L,  0.L/10.L,
    1.L/10.L,  8.L/10.L,  1.L/10.L,  0.L/10.L,
    1.L/10.L,  7.L/10.L,  2.L/10.L,  0.L/10.L,
    1.L/10.L,  6.L/10.L,  3.L/10.L,  0.L/10.L,
    1.L/10.L,  5.L/10.L,  4.L/10.L,  0.L/10.L,
    1.L/10.L,  4.L/10.L,  5.L/10.L,  0.L/10.L,
    1.L/10.L,  3.L/10.L,  6.L/10.L,  0.L/10.L,
    1.L/10.L,  2.L/10.L,  7.L/10.L,  0.L/10.L,
    1.L/10.L,  1.L/10.L,  8.L/10.L,  0.L/10.L,
    7.L/10.L,  1.L/10.L,  1.L/10.L,  1.L/10.L,
    6.L/10.L,  2.L/10.L,  1.L/10.L,  1.L/10.L,
    6.L/10.L,  1.L/10.L,  2.L/10.L,  1.L/10.L,
    6.L/10.L,  1.L/10.L,  1.L/10.L,  2.L/10.L,
    5.L/10.L,  3.L/10.L,  1.L/10.L,  1.L/10.L,
    5.L/10.L,  2.L/10.L,  2.L/10.L,  1.L/10.L,
    5.L/10.L,  2.L/10.L,  1.L/10.L,  2.L/10.L,
    5.L/10.L,  1.L/10.L,  3.L/10.L,  1.L/10.L,
    5.L/10.L,  1.L/10.L,  2.L/10.L,  2.L/10.L,
    5.L/10.L,  1.L/10.L,  1.L/10.L,  3.L/10.L,
    4.L/10.L,  4.L/10.L,  1.L/10.L,  1.L/10.L,
    4.L/10.L,  3.L/10.L,  2.L/10.L,  1.L/10.L,
    4.L/10.L,  3.L/10.L,  1.L/10.L,  2.L/10.L,
    4.L/10.L,  2.L/10.L,  3.L/10.L,  1.L/10.L,
    4.L/10.L,  2.L/10.L,  2.L/10.L,  2.L/10.L,
    4.L/10.L,  2.L/10.L,  1.L/10.L,  3.L/10.L,
    4.L/10.L,  1.L/10.L,  4.L/10.L,  1.L/10.L,
    4.L/10.L,  1.L/10.L,  3.L/10.L,  2.L/10.L,
    4.L/10.L,  1.L/10.L,  2.L/10.L,  3.L/10.L,
    4.L/10.L,  1.L/10.L,  1.L/10.L,  4.L/10.L,
    3.L/10.L,  5.L/10.L,  1.L/10.L,  1.L/10.L,
    3.L/10.L,  4.L/10.L,  2.L/10.L,  1.L/10.L,
    3.L/10.L,  4.L/10.L,  1.L/10.L,  2.L/10.L,
    3.L/10.L,  3.L/10.L,  3.L/10.L,  1.L/10.L,
    3.L/10.L,  3.L/10.L,  2.L/10.L,  2.L/10.L,
    3.L/10.L,  3.L/10.L,  1.L/10.L,  3.L/10.L,
    3.L/10.L,  2.L/10.L,  4.L/10.L,  1.L/10.L,
    3.L/10.L,  2.L/10.L,  3.L/10.L,  2.L/10.L,
    3.L/10.L,  2.L/10.L,  2.L/10.L,  3.L/10.L,
    3.L/10.L,  2.L/10.L,  1.L/10.L,  4.L/10.L,
    3.L/10.L,  1.L/10.L,  5.L/10.L,  1.L/10.L,
    3.L/10.L,  1.L/10.L,  4.L/10.L,  2.L/10.L,
    3.L/10.L,  1.L/10.L,  3.L/10.L,  3.L/10.L,
    3.L/10.L,  1.L/10.L,  2.L/10.L,  4.L/10.L,
    3.L/10.L,  1.L/10.L,  1.L/10.L,  5.L/10.L,
    2.L/10.L,  6.L/10.L,  1.L/10.L,  1.L/10.L,
    2.L/10.L,  5.L/10.L,  2.L/10.L,  1.L/10.L,
    2.L/10.L,  5.L/10.L,  1.L/10.L,  2.L/10.L,
    2.L/10.L,  4.L/10.L,  3.L/10.L,  1.L/10.L,
    2.L/10.L,  4.L/10.L,  2.L/10.L,  2.L/10.L,
    2.L/10.L,  4.L/10.L,  1.L/10.L,  3.L/10.L,
    2.L/10.L,  3.L/10.L,  4.L/10.L,  1.L/10.L,
    2.L/10.L,  3.L/10.L,  3.L/10.L,  2.L/10.L,
    2.L/10.L,  3.L/10.L,  2.L/10.L,  3.L/10.L,
    2.L/10.L,  3.L/10.L,  1.L/10.L,  4.L/10.L,
    2.L/10.L,  2.L/10.L,  5.L/10.L,  1.L/10.L,
    2.L/10.L,  2.L/10.L,  4.L/10.L,  2.L/10.L,
    2.L/10.L,  2.L/10.L,  3.L/10.L,  3.L/10.L,
    2.L/10.L,  2.L/10.L,  2.L/10.L,  4.L/10.L,
    2.L/10.L,  2.L/10.L,  1.L/10.L,  5.L/10.L,
    2.L/10.L,  1.L/10.L,  6.L/10.L,  1.L/10.L,
    2.L/10.L,  1.L/10.L,  5.L/10.L,  2.L/10.L,
    2.L/10.L,  1.L/10.L,  4.L/10.L,  3.L/10.L,
    2.L/10.L,  1.L/10.L,  3.L/10.L,  4.L/10.L,
    2.L/10.L,  1.L/10.L,  2.L/10.L,  5.L/10.L,
    2.L/10.L,  1.L/10.L,  1.L/10.L,  6.L/10.L,
    1.L/10.L,  7.L/10.L,  1.L/10.L,  1.L/10.L,
    1.L/10.L,  6.L/10.L,  2.L/10.L,  1.L/10.L,
    1.L/10.L,  6.L/10.L,  1.L/10.L,  2.L/10.L,
    1.L/10.L,  5.L/10.L,  3.L/10.L,  1.L/10.L,
    1.L/10.L,  5.L/10.L,  2.L/10.L,  2.L/10.L,
    1.L/10.L,  5.L/10.L,  1.L/10.L,  3.L/10.L,
    1.L/10.L,  4.L/10.L,  4.L/10.L,  1.L/10.L,
    1.L/10.L,  4.L/10.L,  3.L/10.L,  2.L/10.L,
    1.L/10.L,  4.L/10.L,  2.L/10.L,  3.L/10.L,
    1.L/10.L,  4.L/10.L,  1.L/10.L,  4.L/10.L,
    1.L/10.L,  3.L/10.L,  5.L/10.L,  1.L/10.L,
    1.L/10.L,  3.L/10.L,  4.L/10.L,  2.L/10.L,
    1.L/10.L,  3.L/10.L,  3.L/10.L,  3.L/10.L,
    1.L/10.L,  3.L/10.L,  2.L/10.L,  4.L/10.L,
    1.L/10.L,  3.L/10.L,  1.L/10.L,  5.L/10.L,
    1.L/10.L,  2.L/10.L,  6.L/10.L,  1.L/10.L,
    1.L/10.L,  2.L/10.L,  5.L/10.L,  2.L/10.L,
    1.L/10.L,  2.L/10.L,  4.L/10.L,  3.L/10.L,
    1.L/10.L,  2.L/10.L,  3.L/10.L,  4.L/10.L,
    1.L/10.L,  2.L/10.L,  2.L/10.L,  5.L/10.L,
    1.L/10.L,  2.L/10.L,  1.L/10.L,  6.L/10.L,
    1.L/10.L,  1.L/10.L,  7.L/10.L,  1.L/10.L,
    1.L/10.L,  1.L/10.L,  6.L/10.L,  2.L/10.L,
    1.L/10.L,  1.L/10.L,  5.L/10.L,  3.L/10.L,
    1.L/10.L,  1.L/10.L,  4.L/10.L,  4.L/10.L,
    1.L/10.L,  1.L/10.L,  3.L/10.L,  5.L/10.L,
    1.L/10.L,  1.L/10.L,  2.L/10.L,  6.L/10.L,
    1.L/10.L,  1.L/10.L,  1.L/10.L,  7.L/10.L
};
DOF_TYPE DOF_DG10_ = {
    DofCache, "DG10", DG10_points, NULL,
    DOF_DG9, DOF_P10, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 286, 10, -1, 1,
    0, 0, 0, 286
};

static FLOAT DG11_points[] = {
   11.L/11.L,  0.L/11.L,  0.L/11.L,  0.L/11.L,
    0.L/11.L, 11.L/11.L,  0.L/11.L,  0.L/11.L,
    0.L/11.L,  0.L/11.L, 11.L/11.L,  0.L/11.L,
    0.L/11.L,  0.L/11.L,  0.L/11.L, 11.L/11.L,
   10.L/11.L,  1.L/11.L,  0.L/11.L,  0.L/11.L,
    9.L/11.L,  2.L/11.L,  0.L/11.L,  0.L/11.L,
    8.L/11.L,  3.L/11.L,  0.L/11.L,  0.L/11.L,
    7.L/11.L,  4.L/11.L,  0.L/11.L,  0.L/11.L,
    6.L/11.L,  5.L/11.L,  0.L/11.L,  0.L/11.L,
    5.L/11.L,  6.L/11.L,  0.L/11.L,  0.L/11.L,
    4.L/11.L,  7.L/11.L,  0.L/11.L,  0.L/11.L,
    3.L/11.L,  8.L/11.L,  0.L/11.L,  0.L/11.L,
    2.L/11.L,  9.L/11.L,  0.L/11.L,  0.L/11.L,
    1.L/11.L, 10.L/11.L,  0.L/11.L,  0.L/11.L,
   10.L/11.L,  0.L/11.L,  1.L/11.L,  0.L/11.L,
    9.L/11.L,  0.L/11.L,  2.L/11.L,  0.L/11.L,
    8.L/11.L,  0.L/11.L,  3.L/11.L,  0.L/11.L,
    7.L/11.L,  0.L/11.L,  4.L/11.L,  0.L/11.L,
    6.L/11.L,  0.L/11.L,  5.L/11.L,  0.L/11.L,
    5.L/11.L,  0.L/11.L,  6.L/11.L,  0.L/11.L,
    4.L/11.L,  0.L/11.L,  7.L/11.L,  0.L/11.L,
    3.L/11.L,  0.L/11.L,  8.L/11.L,  0.L/11.L,
    2.L/11.L,  0.L/11.L,  9.L/11.L,  0.L/11.L,
    1.L/11.L,  0.L/11.L, 10.L/11.L,  0.L/11.L,
   10.L/11.L,  0.L/11.L,  0.L/11.L,  1.L/11.L,
    9.L/11.L,  0.L/11.L,  0.L/11.L,  2.L/11.L,
    8.L/11.L,  0.L/11.L,  0.L/11.L,  3.L/11.L,
    7.L/11.L,  0.L/11.L,  0.L/11.L,  4.L/11.L,
    6.L/11.L,  0.L/11.L,  0.L/11.L,  5.L/11.L,
    5.L/11.L,  0.L/11.L,  0.L/11.L,  6.L/11.L,
    4.L/11.L,  0.L/11.L,  0.L/11.L,  7.L/11.L,
    3.L/11.L,  0.L/11.L,  0.L/11.L,  8.L/11.L,
    2.L/11.L,  0.L/11.L,  0.L/11.L,  9.L/11.L,
    1.L/11.L,  0.L/11.L,  0.L/11.L, 10.L/11.L,
    0.L/11.L, 10.L/11.L,  1.L/11.L,  0.L/11.L,
    0.L/11.L,  9.L/11.L,  2.L/11.L,  0.L/11.L,
    0.L/11.L,  8.L/11.L,  3.L/11.L,  0.L/11.L,
    0.L/11.L,  7.L/11.L,  4.L/11.L,  0.L/11.L,
    0.L/11.L,  6.L/11.L,  5.L/11.L,  0.L/11.L,
    0.L/11.L,  5.L/11.L,  6.L/11.L,  0.L/11.L,
    0.L/11.L,  4.L/11.L,  7.L/11.L,  0.L/11.L,
    0.L/11.L,  3.L/11.L,  8.L/11.L,  0.L/11.L,
    0.L/11.L,  2.L/11.L,  9.L/11.L,  0.L/11.L,
    0.L/11.L,  1.L/11.L, 10.L/11.L,  0.L/11.L,
    0.L/11.L, 10.L/11.L,  0.L/11.L,  1.L/11.L,
    0.L/11.L,  9.L/11.L,  0.L/11.L,  2.L/11.L,
    0.L/11.L,  8.L/11.L,  0.L/11.L,  3.L/11.L,
    0.L/11.L,  7.L/11.L,  0.L/11.L,  4.L/11.L,
    0.L/11.L,  6.L/11.L,  0.L/11.L,  5.L/11.L,
    0.L/11.L,  5.L/11.L,  0.L/11.L,  6.L/11.L,
    0.L/11.L,  4.L/11.L,  0.L/11.L,  7.L/11.L,
    0.L/11.L,  3.L/11.L,  0.L/11.L,  8.L/11.L,
    0.L/11.L,  2.L/11.L,  0.L/11.L,  9.L/11.L,
    0.L/11.L,  1.L/11.L,  0.L/11.L, 10.L/11.L,
    0.L/11.L,  0.L/11.L, 10.L/11.L,  1.L/11.L,
    0.L/11.L,  0.L/11.L,  9.L/11.L,  2.L/11.L,
    0.L/11.L,  0.L/11.L,  8.L/11.L,  3.L/11.L,
    0.L/11.L,  0.L/11.L,  7.L/11.L,  4.L/11.L,
    0.L/11.L,  0.L/11.L,  6.L/11.L,  5.L/11.L,
    0.L/11.L,  0.L/11.L,  5.L/11.L,  6.L/11.L,
    0.L/11.L,  0.L/11.L,  4.L/11.L,  7.L/11.L,
    0.L/11.L,  0.L/11.L,  3.L/11.L,  8.L/11.L,
    0.L/11.L,  0.L/11.L,  2.L/11.L,  9.L/11.L,
    0.L/11.L,  0.L/11.L,  1.L/11.L, 10.L/11.L,
    0.L/11.L,  9.L/11.L,  1.L/11.L,  1.L/11.L,
    0.L/11.L,  8.L/11.L,  2.L/11.L,  1.L/11.L,
    0.L/11.L,  8.L/11.L,  1.L/11.L,  2.L/11.L,
    0.L/11.L,  7.L/11.L,  3.L/11.L,  1.L/11.L,
    0.L/11.L,  7.L/11.L,  2.L/11.L,  2.L/11.L,
    0.L/11.L,  7.L/11.L,  1.L/11.L,  3.L/11.L,
    0.L/11.L,  6.L/11.L,  4.L/11.L,  1.L/11.L,
    0.L/11.L,  6.L/11.L,  3.L/11.L,  2.L/11.L,
    0.L/11.L,  6.L/11.L,  2.L/11.L,  3.L/11.L,
    0.L/11.L,  6.L/11.L,  1.L/11.L,  4.L/11.L,
    0.L/11.L,  5.L/11.L,  5.L/11.L,  1.L/11.L,
    0.L/11.L,  5.L/11.L,  4.L/11.L,  2.L/11.L,
    0.L/11.L,  5.L/11.L,  3.L/11.L,  3.L/11.L,
    0.L/11.L,  5.L/11.L,  2.L/11.L,  4.L/11.L,
    0.L/11.L,  5.L/11.L,  1.L/11.L,  5.L/11.L,
    0.L/11.L,  4.L/11.L,  6.L/11.L,  1.L/11.L,
    0.L/11.L,  4.L/11.L,  5.L/11.L,  2.L/11.L,
    0.L/11.L,  4.L/11.L,  4.L/11.L,  3.L/11.L,
    0.L/11.L,  4.L/11.L,  3.L/11.L,  4.L/11.L,
    0.L/11.L,  4.L/11.L,  2.L/11.L,  5.L/11.L,
    0.L/11.L,  4.L/11.L,  1.L/11.L,  6.L/11.L,
    0.L/11.L,  3.L/11.L,  7.L/11.L,  1.L/11.L,
    0.L/11.L,  3.L/11.L,  6.L/11.L,  2.L/11.L,
    0.L/11.L,  3.L/11.L,  5.L/11.L,  3.L/11.L,
    0.L/11.L,  3.L/11.L,  4.L/11.L,  4.L/11.L,
    0.L/11.L,  3.L/11.L,  3.L/11.L,  5.L/11.L,
    0.L/11.L,  3.L/11.L,  2.L/11.L,  6.L/11.L,
    0.L/11.L,  3.L/11.L,  1.L/11.L,  7.L/11.L,
    0.L/11.L,  2.L/11.L,  8.L/11.L,  1.L/11.L,
    0.L/11.L,  2.L/11.L,  7.L/11.L,  2.L/11.L,
    0.L/11.L,  2.L/11.L,  6.L/11.L,  3.L/11.L,
    0.L/11.L,  2.L/11.L,  5.L/11.L,  4.L/11.L,
    0.L/11.L,  2.L/11.L,  4.L/11.L,  5.L/11.L,
    0.L/11.L,  2.L/11.L,  3.L/11.L,  6.L/11.L,
    0.L/11.L,  2.L/11.L,  2.L/11.L,  7.L/11.L,
    0.L/11.L,  2.L/11.L,  1.L/11.L,  8.L/11.L,
    0.L/11.L,  1.L/11.L,  9.L/11.L,  1.L/11.L,
    0.L/11.L,  1.L/11.L,  8.L/11.L,  2.L/11.L,
    0.L/11.L,  1.L/11.L,  7.L/11.L,  3.L/11.L,
    0.L/11.L,  1.L/11.L,  6.L/11.L,  4.L/11.L,
    0.L/11.L,  1.L/11.L,  5.L/11.L,  5.L/11.L,
    0.L/11.L,  1.L/11.L,  4.L/11.L,  6.L/11.L,
    0.L/11.L,  1.L/11.L,  3.L/11.L,  7.L/11.L,
    0.L/11.L,  1.L/11.L,  2.L/11.L,  8.L/11.L,
    0.L/11.L,  1.L/11.L,  1.L/11.L,  9.L/11.L,
    9.L/11.L,  0.L/11.L,  1.L/11.L,  1.L/11.L,
    8.L/11.L,  0.L/11.L,  2.L/11.L,  1.L/11.L,
    8.L/11.L,  0.L/11.L,  1.L/11.L,  2.L/11.L,
    7.L/11.L,  0.L/11.L,  3.L/11.L,  1.L/11.L,
    7.L/11.L,  0.L/11.L,  2.L/11.L,  2.L/11.L,
    7.L/11.L,  0.L/11.L,  1.L/11.L,  3.L/11.L,
    6.L/11.L,  0.L/11.L,  4.L/11.L,  1.L/11.L,
    6.L/11.L,  0.L/11.L,  3.L/11.L,  2.L/11.L,
    6.L/11.L,  0.L/11.L,  2.L/11.L,  3.L/11.L,
    6.L/11.L,  0.L/11.L,  1.L/11.L,  4.L/11.L,
    5.L/11.L,  0.L/11.L,  5.L/11.L,  1.L/11.L,
    5.L/11.L,  0.L/11.L,  4.L/11.L,  2.L/11.L,
    5.L/11.L,  0.L/11.L,  3.L/11.L,  3.L/11.L,
    5.L/11.L,  0.L/11.L,  2.L/11.L,  4.L/11.L,
    5.L/11.L,  0.L/11.L,  1.L/11.L,  5.L/11.L,
    4.L/11.L,  0.L/11.L,  6.L/11.L,  1.L/11.L,
    4.L/11.L,  0.L/11.L,  5.L/11.L,  2.L/11.L,
    4.L/11.L,  0.L/11.L,  4.L/11.L,  3.L/11.L,
    4.L/11.L,  0.L/11.L,  3.L/11.L,  4.L/11.L,
    4.L/11.L,  0.L/11.L,  2.L/11.L,  5.L/11.L,
    4.L/11.L,  0.L/11.L,  1.L/11.L,  6.L/11.L,
    3.L/11.L,  0.L/11.L,  7.L/11.L,  1.L/11.L,
    3.L/11.L,  0.L/11.L,  6.L/11.L,  2.L/11.L,
    3.L/11.L,  0.L/11.L,  5.L/11.L,  3.L/11.L,
    3.L/11.L,  0.L/11.L,  4.L/11.L,  4.L/11.L,
    3.L/11.L,  0.L/11.L,  3.L/11.L,  5.L/11.L,
    3.L/11.L,  0.L/11.L,  2.L/11.L,  6.L/11.L,
    3.L/11.L,  0.L/11.L,  1.L/11.L,  7.L/11.L,
    2.L/11.L,  0.L/11.L,  8.L/11.L,  1.L/11.L,
    2.L/11.L,  0.L/11.L,  7.L/11.L,  2.L/11.L,
    2.L/11.L,  0.L/11.L,  6.L/11.L,  3.L/11.L,
    2.L/11.L,  0.L/11.L,  5.L/11.L,  4.L/11.L,
    2.L/11.L,  0.L/11.L,  4.L/11.L,  5.L/11.L,
    2.L/11.L,  0.L/11.L,  3.L/11.L,  6.L/11.L,
    2.L/11.L,  0.L/11.L,  2.L/11.L,  7.L/11.L,
    2.L/11.L,  0.L/11.L,  1.L/11.L,  8.L/11.L,
    1.L/11.L,  0.L/11.L,  9.L/11.L,  1.L/11.L,
    1.L/11.L,  0.L/11.L,  8.L/11.L,  2.L/11.L,
    1.L/11.L,  0.L/11.L,  7.L/11.L,  3.L/11.L,
    1.L/11.L,  0.L/11.L,  6.L/11.L,  4.L/11.L,
    1.L/11.L,  0.L/11.L,  5.L/11.L,  5.L/11.L,
    1.L/11.L,  0.L/11.L,  4.L/11.L,  6.L/11.L,
    1.L/11.L,  0.L/11.L,  3.L/11.L,  7.L/11.L,
    1.L/11.L,  0.L/11.L,  2.L/11.L,  8.L/11.L,
    1.L/11.L,  0.L/11.L,  1.L/11.L,  9.L/11.L,
    9.L/11.L,  1.L/11.L,  0.L/11.L,  1.L/11.L,
    8.L/11.L,  2.L/11.L,  0.L/11.L,  1.L/11.L,
    8.L/11.L,  1.L/11.L,  0.L/11.L,  2.L/11.L,
    7.L/11.L,  3.L/11.L,  0.L/11.L,  1.L/11.L,
    7.L/11.L,  2.L/11.L,  0.L/11.L,  2.L/11.L,
    7.L/11.L,  1.L/11.L,  0.L/11.L,  3.L/11.L,
    6.L/11.L,  4.L/11.L,  0.L/11.L,  1.L/11.L,
    6.L/11.L,  3.L/11.L,  0.L/11.L,  2.L/11.L,
    6.L/11.L,  2.L/11.L,  0.L/11.L,  3.L/11.L,
    6.L/11.L,  1.L/11.L,  0.L/11.L,  4.L/11.L,
    5.L/11.L,  5.L/11.L,  0.L/11.L,  1.L/11.L,
    5.L/11.L,  4.L/11.L,  0.L/11.L,  2.L/11.L,
    5.L/11.L,  3.L/11.L,  0.L/11.L,  3.L/11.L,
    5.L/11.L,  2.L/11.L,  0.L/11.L,  4.L/11.L,
    5.L/11.L,  1.L/11.L,  0.L/11.L,  5.L/11.L,
    4.L/11.L,  6.L/11.L,  0.L/11.L,  1.L/11.L,
    4.L/11.L,  5.L/11.L,  0.L/11.L,  2.L/11.L,
    4.L/11.L,  4.L/11.L,  0.L/11.L,  3.L/11.L,
    4.L/11.L,  3.L/11.L,  0.L/11.L,  4.L/11.L,
    4.L/11.L,  2.L/11.L,  0.L/11.L,  5.L/11.L,
    4.L/11.L,  1.L/11.L,  0.L/11.L,  6.L/11.L,
    3.L/11.L,  7.L/11.L,  0.L/11.L,  1.L/11.L,
    3.L/11.L,  6.L/11.L,  0.L/11.L,  2.L/11.L,
    3.L/11.L,  5.L/11.L,  0.L/11.L,  3.L/11.L,
    3.L/11.L,  4.L/11.L,  0.L/11.L,  4.L/11.L,
    3.L/11.L,  3.L/11.L,  0.L/11.L,  5.L/11.L,
    3.L/11.L,  2.L/11.L,  0.L/11.L,  6.L/11.L,
    3.L/11.L,  1.L/11.L,  0.L/11.L,  7.L/11.L,
    2.L/11.L,  8.L/11.L,  0.L/11.L,  1.L/11.L,
    2.L/11.L,  7.L/11.L,  0.L/11.L,  2.L/11.L,
    2.L/11.L,  6.L/11.L,  0.L/11.L,  3.L/11.L,
    2.L/11.L,  5.L/11.L,  0.L/11.L,  4.L/11.L,
    2.L/11.L,  4.L/11.L,  0.L/11.L,  5.L/11.L,
    2.L/11.L,  3.L/11.L,  0.L/11.L,  6.L/11.L,
    2.L/11.L,  2.L/11.L,  0.L/11.L,  7.L/11.L,
    2.L/11.L,  1.L/11.L,  0.L/11.L,  8.L/11.L,
    1.L/11.L,  9.L/11.L,  0.L/11.L,  1.L/11.L,
    1.L/11.L,  8.L/11.L,  0.L/11.L,  2.L/11.L,
    1.L/11.L,  7.L/11.L,  0.L/11.L,  3.L/11.L,
    1.L/11.L,  6.L/11.L,  0.L/11.L,  4.L/11.L,
    1.L/11.L,  5.L/11.L,  0.L/11.L,  5.L/11.L,
    1.L/11.L,  4.L/11.L,  0.L/11.L,  6.L/11.L,
    1.L/11.L,  3.L/11.L,  0.L/11.L,  7.L/11.L,
    1.L/11.L,  2.L/11.L,  0.L/11.L,  8.L/11.L,
    1.L/11.L,  1.L/11.L,  0.L/11.L,  9.L/11.L,
    9.L/11.L,  1.L/11.L,  1.L/11.L,  0.L/11.L,
    8.L/11.L,  2.L/11.L,  1.L/11.L,  0.L/11.L,
    8.L/11.L,  1.L/11.L,  2.L/11.L,  0.L/11.L,
    7.L/11.L,  3.L/11.L,  1.L/11.L,  0.L/11.L,
    7.L/11.L,  2.L/11.L,  2.L/11.L,  0.L/11.L,
    7.L/11.L,  1.L/11.L,  3.L/11.L,  0.L/11.L,
    6.L/11.L,  4.L/11.L,  1.L/11.L,  0.L/11.L,
    6.L/11.L,  3.L/11.L,  2.L/11.L,  0.L/11.L,
    6.L/11.L,  2.L/11.L,  3.L/11.L,  0.L/11.L,
    6.L/11.L,  1.L/11.L,  4.L/11.L,  0.L/11.L,
    5.L/11.L,  5.L/11.L,  1.L/11.L,  0.L/11.L,
    5.L/11.L,  4.L/11.L,  2.L/11.L,  0.L/11.L,
    5.L/11.L,  3.L/11.L,  3.L/11.L,  0.L/11.L,
    5.L/11.L,  2.L/11.L,  4.L/11.L,  0.L/11.L,
    5.L/11.L,  1.L/11.L,  5.L/11.L,  0.L/11.L,
    4.L/11.L,  6.L/11.L,  1.L/11.L,  0.L/11.L,
    4.L/11.L,  5.L/11.L,  2.L/11.L,  0.L/11.L,
    4.L/11.L,  4.L/11.L,  3.L/11.L,  0.L/11.L,
    4.L/11.L,  3.L/11.L,  4.L/11.L,  0.L/11.L,
    4.L/11.L,  2.L/11.L,  5.L/11.L,  0.L/11.L,
    4.L/11.L,  1.L/11.L,  6.L/11.L,  0.L/11.L,
    3.L/11.L,  7.L/11.L,  1.L/11.L,  0.L/11.L,
    3.L/11.L,  6.L/11.L,  2.L/11.L,  0.L/11.L,
    3.L/11.L,  5.L/11.L,  3.L/11.L,  0.L/11.L,
    3.L/11.L,  4.L/11.L,  4.L/11.L,  0.L/11.L,
    3.L/11.L,  3.L/11.L,  5.L/11.L,  0.L/11.L,
    3.L/11.L,  2.L/11.L,  6.L/11.L,  0.L/11.L,
    3.L/11.L,  1.L/11.L,  7.L/11.L,  0.L/11.L,
    2.L/11.L,  8.L/11.L,  1.L/11.L,  0.L/11.L,
    2.L/11.L,  7.L/11.L,  2.L/11.L,  0.L/11.L,
    2.L/11.L,  6.L/11.L,  3.L/11.L,  0.L/11.L,
    2.L/11.L,  5.L/11.L,  4.L/11.L,  0.L/11.L,
    2.L/11.L,  4.L/11.L,  5.L/11.L,  0.L/11.L,
    2.L/11.L,  3.L/11.L,  6.L/11.L,  0.L/11.L,
    2.L/11.L,  2.L/11.L,  7.L/11.L,  0.L/11.L,
    2.L/11.L,  1.L/11.L,  8.L/11.L,  0.L/11.L,
    1.L/11.L,  9.L/11.L,  1.L/11.L,  0.L/11.L,
    1.L/11.L,  8.L/11.L,  2.L/11.L,  0.L/11.L,
    1.L/11.L,  7.L/11.L,  3.L/11.L,  0.L/11.L,
    1.L/11.L,  6.L/11.L,  4.L/11.L,  0.L/11.L,
    1.L/11.L,  5.L/11.L,  5.L/11.L,  0.L/11.L,
    1.L/11.L,  4.L/11.L,  6.L/11.L,  0.L/11.L,
    1.L/11.L,  3.L/11.L,  7.L/11.L,  0.L/11.L,
    1.L/11.L,  2.L/11.L,  8.L/11.L,  0.L/11.L,
    1.L/11.L,  1.L/11.L,  9.L/11.L,  0.L/11.L,
    8.L/11.L,  1.L/11.L,  1.L/11.L,  1.L/11.L,
    7.L/11.L,  2.L/11.L,  1.L/11.L,  1.L/11.L,
    7.L/11.L,  1.L/11.L,  2.L/11.L,  1.L/11.L,
    7.L/11.L,  1.L/11.L,  1.L/11.L,  2.L/11.L,
    6.L/11.L,  3.L/11.L,  1.L/11.L,  1.L/11.L,
    6.L/11.L,  2.L/11.L,  2.L/11.L,  1.L/11.L,
    6.L/11.L,  2.L/11.L,  1.L/11.L,  2.L/11.L,
    6.L/11.L,  1.L/11.L,  3.L/11.L,  1.L/11.L,
    6.L/11.L,  1.L/11.L,  2.L/11.L,  2.L/11.L,
    6.L/11.L,  1.L/11.L,  1.L/11.L,  3.L/11.L,
    5.L/11.L,  4.L/11.L,  1.L/11.L,  1.L/11.L,
    5.L/11.L,  3.L/11.L,  2.L/11.L,  1.L/11.L,
    5.L/11.L,  3.L/11.L,  1.L/11.L,  2.L/11.L,
    5.L/11.L,  2.L/11.L,  3.L/11.L,  1.L/11.L,
    5.L/11.L,  2.L/11.L,  2.L/11.L,  2.L/11.L,
    5.L/11.L,  2.L/11.L,  1.L/11.L,  3.L/11.L,
    5.L/11.L,  1.L/11.L,  4.L/11.L,  1.L/11.L,
    5.L/11.L,  1.L/11.L,  3.L/11.L,  2.L/11.L,
    5.L/11.L,  1.L/11.L,  2.L/11.L,  3.L/11.L,
    5.L/11.L,  1.L/11.L,  1.L/11.L,  4.L/11.L,
    4.L/11.L,  5.L/11.L,  1.L/11.L,  1.L/11.L,
    4.L/11.L,  4.L/11.L,  2.L/11.L,  1.L/11.L,
    4.L/11.L,  4.L/11.L,  1.L/11.L,  2.L/11.L,
    4.L/11.L,  3.L/11.L,  3.L/11.L,  1.L/11.L,
    4.L/11.L,  3.L/11.L,  2.L/11.L,  2.L/11.L,
    4.L/11.L,  3.L/11.L,  1.L/11.L,  3.L/11.L,
    4.L/11.L,  2.L/11.L,  4.L/11.L,  1.L/11.L,
    4.L/11.L,  2.L/11.L,  3.L/11.L,  2.L/11.L,
    4.L/11.L,  2.L/11.L,  2.L/11.L,  3.L/11.L,
    4.L/11.L,  2.L/11.L,  1.L/11.L,  4.L/11.L,
    4.L/11.L,  1.L/11.L,  5.L/11.L,  1.L/11.L,
    4.L/11.L,  1.L/11.L,  4.L/11.L,  2.L/11.L,
    4.L/11.L,  1.L/11.L,  3.L/11.L,  3.L/11.L,
    4.L/11.L,  1.L/11.L,  2.L/11.L,  4.L/11.L,
    4.L/11.L,  1.L/11.L,  1.L/11.L,  5.L/11.L,
    3.L/11.L,  6.L/11.L,  1.L/11.L,  1.L/11.L,
    3.L/11.L,  5.L/11.L,  2.L/11.L,  1.L/11.L,
    3.L/11.L,  5.L/11.L,  1.L/11.L,  2.L/11.L,
    3.L/11.L,  4.L/11.L,  3.L/11.L,  1.L/11.L,
    3.L/11.L,  4.L/11.L,  2.L/11.L,  2.L/11.L,
    3.L/11.L,  4.L/11.L,  1.L/11.L,  3.L/11.L,
    3.L/11.L,  3.L/11.L,  4.L/11.L,  1.L/11.L,
    3.L/11.L,  3.L/11.L,  3.L/11.L,  2.L/11.L,
    3.L/11.L,  3.L/11.L,  2.L/11.L,  3.L/11.L,
    3.L/11.L,  3.L/11.L,  1.L/11.L,  4.L/11.L,
    3.L/11.L,  2.L/11.L,  5.L/11.L,  1.L/11.L,
    3.L/11.L,  2.L/11.L,  4.L/11.L,  2.L/11.L,
    3.L/11.L,  2.L/11.L,  3.L/11.L,  3.L/11.L,
    3.L/11.L,  2.L/11.L,  2.L/11.L,  4.L/11.L,
    3.L/11.L,  2.L/11.L,  1.L/11.L,  5.L/11.L,
    3.L/11.L,  1.L/11.L,  6.L/11.L,  1.L/11.L,
    3.L/11.L,  1.L/11.L,  5.L/11.L,  2.L/11.L,
    3.L/11.L,  1.L/11.L,  4.L/11.L,  3.L/11.L,
    3.L/11.L,  1.L/11.L,  3.L/11.L,  4.L/11.L,
    3.L/11.L,  1.L/11.L,  2.L/11.L,  5.L/11.L,
    3.L/11.L,  1.L/11.L,  1.L/11.L,  6.L/11.L,
    2.L/11.L,  7.L/11.L,  1.L/11.L,  1.L/11.L,
    2.L/11.L,  6.L/11.L,  2.L/11.L,  1.L/11.L,
    2.L/11.L,  6.L/11.L,  1.L/11.L,  2.L/11.L,
    2.L/11.L,  5.L/11.L,  3.L/11.L,  1.L/11.L,
    2.L/11.L,  5.L/11.L,  2.L/11.L,  2.L/11.L,
    2.L/11.L,  5.L/11.L,  1.L/11.L,  3.L/11.L,
    2.L/11.L,  4.L/11.L,  4.L/11.L,  1.L/11.L,
    2.L/11.L,  4.L/11.L,  3.L/11.L,  2.L/11.L,
    2.L/11.L,  4.L/11.L,  2.L/11.L,  3.L/11.L,
    2.L/11.L,  4.L/11.L,  1.L/11.L,  4.L/11.L,
    2.L/11.L,  3.L/11.L,  5.L/11.L,  1.L/11.L,
    2.L/11.L,  3.L/11.L,  4.L/11.L,  2.L/11.L,
    2.L/11.L,  3.L/11.L,  3.L/11.L,  3.L/11.L,
    2.L/11.L,  3.L/11.L,  2.L/11.L,  4.L/11.L,
    2.L/11.L,  3.L/11.L,  1.L/11.L,  5.L/11.L,
    2.L/11.L,  2.L/11.L,  6.L/11.L,  1.L/11.L,
    2.L/11.L,  2.L/11.L,  5.L/11.L,  2.L/11.L,
    2.L/11.L,  2.L/11.L,  4.L/11.L,  3.L/11.L,
    2.L/11.L,  2.L/11.L,  3.L/11.L,  4.L/11.L,
    2.L/11.L,  2.L/11.L,  2.L/11.L,  5.L/11.L,
    2.L/11.L,  2.L/11.L,  1.L/11.L,  6.L/11.L,
    2.L/11.L,  1.L/11.L,  7.L/11.L,  1.L/11.L,
    2.L/11.L,  1.L/11.L,  6.L/11.L,  2.L/11.L,
    2.L/11.L,  1.L/11.L,  5.L/11.L,  3.L/11.L,
    2.L/11.L,  1.L/11.L,  4.L/11.L,  4.L/11.L,
    2.L/11.L,  1.L/11.L,  3.L/11.L,  5.L/11.L,
    2.L/11.L,  1.L/11.L,  2.L/11.L,  6.L/11.L,
    2.L/11.L,  1.L/11.L,  1.L/11.L,  7.L/11.L,
    1.L/11.L,  8.L/11.L,  1.L/11.L,  1.L/11.L,
    1.L/11.L,  7.L/11.L,  2.L/11.L,  1.L/11.L,
    1.L/11.L,  7.L/11.L,  1.L/11.L,  2.L/11.L,
    1.L/11.L,  6.L/11.L,  3.L/11.L,  1.L/11.L,
    1.L/11.L,  6.L/11.L,  2.L/11.L,  2.L/11.L,
    1.L/11.L,  6.L/11.L,  1.L/11.L,  3.L/11.L,
    1.L/11.L,  5.L/11.L,  4.L/11.L,  1.L/11.L,
    1.L/11.L,  5.L/11.L,  3.L/11.L,  2.L/11.L,
    1.L/11.L,  5.L/11.L,  2.L/11.L,  3.L/11.L,
    1.L/11.L,  5.L/11.L,  1.L/11.L,  4.L/11.L,
    1.L/11.L,  4.L/11.L,  5.L/11.L,  1.L/11.L,
    1.L/11.L,  4.L/11.L,  4.L/11.L,  2.L/11.L,
    1.L/11.L,  4.L/11.L,  3.L/11.L,  3.L/11.L,
    1.L/11.L,  4.L/11.L,  2.L/11.L,  4.L/11.L,
    1.L/11.L,  4.L/11.L,  1.L/11.L,  5.L/11.L,
    1.L/11.L,  3.L/11.L,  6.L/11.L,  1.L/11.L,
    1.L/11.L,  3.L/11.L,  5.L/11.L,  2.L/11.L,
    1.L/11.L,  3.L/11.L,  4.L/11.L,  3.L/11.L,
    1.L/11.L,  3.L/11.L,  3.L/11.L,  4.L/11.L,
    1.L/11.L,  3.L/11.L,  2.L/11.L,  5.L/11.L,
    1.L/11.L,  3.L/11.L,  1.L/11.L,  6.L/11.L,
    1.L/11.L,  2.L/11.L,  7.L/11.L,  1.L/11.L,
    1.L/11.L,  2.L/11.L,  6.L/11.L,  2.L/11.L,
    1.L/11.L,  2.L/11.L,  5.L/11.L,  3.L/11.L,
    1.L/11.L,  2.L/11.L,  4.L/11.L,  4.L/11.L,
    1.L/11.L,  2.L/11.L,  3.L/11.L,  5.L/11.L,
    1.L/11.L,  2.L/11.L,  2.L/11.L,  6.L/11.L,
    1.L/11.L,  2.L/11.L,  1.L/11.L,  7.L/11.L,
    1.L/11.L,  1.L/11.L,  8.L/11.L,  1.L/11.L,
    1.L/11.L,  1.L/11.L,  7.L/11.L,  2.L/11.L,
    1.L/11.L,  1.L/11.L,  6.L/11.L,  3.L/11.L,
    1.L/11.L,  1.L/11.L,  5.L/11.L,  4.L/11.L,
    1.L/11.L,  1.L/11.L,  4.L/11.L,  5.L/11.L,
    1.L/11.L,  1.L/11.L,  3.L/11.L,  6.L/11.L,
    1.L/11.L,  1.L/11.L,  2.L/11.L,  7.L/11.L,
    1.L/11.L,  1.L/11.L,  1.L/11.L,  8.L/11.L
};
DOF_TYPE DOF_DG11_ = {
    DofCache, "DG11", DG11_points, NULL,
    DOF_DG10, DOF_P11, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 364, 11, -1, 1,
    0, 0, 0, 364
};

static FLOAT DG12_points[] = {
   12.L/12.L,  0.L/12.L,  0.L/12.L,  0.L/12.L,
    0.L/12.L, 12.L/12.L,  0.L/12.L,  0.L/12.L,
    0.L/12.L,  0.L/12.L, 12.L/12.L,  0.L/12.L,
    0.L/12.L,  0.L/12.L,  0.L/12.L, 12.L/12.L,
   11.L/12.L,  1.L/12.L,  0.L/12.L,  0.L/12.L,
   10.L/12.L,  2.L/12.L,  0.L/12.L,  0.L/12.L,
    9.L/12.L,  3.L/12.L,  0.L/12.L,  0.L/12.L,
    8.L/12.L,  4.L/12.L,  0.L/12.L,  0.L/12.L,
    7.L/12.L,  5.L/12.L,  0.L/12.L,  0.L/12.L,
    6.L/12.L,  6.L/12.L,  0.L/12.L,  0.L/12.L,
    5.L/12.L,  7.L/12.L,  0.L/12.L,  0.L/12.L,
    4.L/12.L,  8.L/12.L,  0.L/12.L,  0.L/12.L,
    3.L/12.L,  9.L/12.L,  0.L/12.L,  0.L/12.L,
    2.L/12.L, 10.L/12.L,  0.L/12.L,  0.L/12.L,
    1.L/12.L, 11.L/12.L,  0.L/12.L,  0.L/12.L,
   11.L/12.L,  0.L/12.L,  1.L/12.L,  0.L/12.L,
   10.L/12.L,  0.L/12.L,  2.L/12.L,  0.L/12.L,
    9.L/12.L,  0.L/12.L,  3.L/12.L,  0.L/12.L,
    8.L/12.L,  0.L/12.L,  4.L/12.L,  0.L/12.L,
    7.L/12.L,  0.L/12.L,  5.L/12.L,  0.L/12.L,
    6.L/12.L,  0.L/12.L,  6.L/12.L,  0.L/12.L,
    5.L/12.L,  0.L/12.L,  7.L/12.L,  0.L/12.L,
    4.L/12.L,  0.L/12.L,  8.L/12.L,  0.L/12.L,
    3.L/12.L,  0.L/12.L,  9.L/12.L,  0.L/12.L,
    2.L/12.L,  0.L/12.L, 10.L/12.L,  0.L/12.L,
    1.L/12.L,  0.L/12.L, 11.L/12.L,  0.L/12.L,
   11.L/12.L,  0.L/12.L,  0.L/12.L,  1.L/12.L,
   10.L/12.L,  0.L/12.L,  0.L/12.L,  2.L/12.L,
    9.L/12.L,  0.L/12.L,  0.L/12.L,  3.L/12.L,
    8.L/12.L,  0.L/12.L,  0.L/12.L,  4.L/12.L,
    7.L/12.L,  0.L/12.L,  0.L/12.L,  5.L/12.L,
    6.L/12.L,  0.L/12.L,  0.L/12.L,  6.L/12.L,
    5.L/12.L,  0.L/12.L,  0.L/12.L,  7.L/12.L,
    4.L/12.L,  0.L/12.L,  0.L/12.L,  8.L/12.L,
    3.L/12.L,  0.L/12.L,  0.L/12.L,  9.L/12.L,
    2.L/12.L,  0.L/12.L,  0.L/12.L, 10.L/12.L,
    1.L/12.L,  0.L/12.L,  0.L/12.L, 11.L/12.L,
    0.L/12.L, 11.L/12.L,  1.L/12.L,  0.L/12.L,
    0.L/12.L, 10.L/12.L,  2.L/12.L,  0.L/12.L,
    0.L/12.L,  9.L/12.L,  3.L/12.L,  0.L/12.L,
    0.L/12.L,  8.L/12.L,  4.L/12.L,  0.L/12.L,
    0.L/12.L,  7.L/12.L,  5.L/12.L,  0.L/12.L,
    0.L/12.L,  6.L/12.L,  6.L/12.L,  0.L/12.L,
    0.L/12.L,  5.L/12.L,  7.L/12.L,  0.L/12.L,
    0.L/12.L,  4.L/12.L,  8.L/12.L,  0.L/12.L,
    0.L/12.L,  3.L/12.L,  9.L/12.L,  0.L/12.L,
    0.L/12.L,  2.L/12.L, 10.L/12.L,  0.L/12.L,
    0.L/12.L,  1.L/12.L, 11.L/12.L,  0.L/12.L,
    0.L/12.L, 11.L/12.L,  0.L/12.L,  1.L/12.L,
    0.L/12.L, 10.L/12.L,  0.L/12.L,  2.L/12.L,
    0.L/12.L,  9.L/12.L,  0.L/12.L,  3.L/12.L,
    0.L/12.L,  8.L/12.L,  0.L/12.L,  4.L/12.L,
    0.L/12.L,  7.L/12.L,  0.L/12.L,  5.L/12.L,
    0.L/12.L,  6.L/12.L,  0.L/12.L,  6.L/12.L,
    0.L/12.L,  5.L/12.L,  0.L/12.L,  7.L/12.L,
    0.L/12.L,  4.L/12.L,  0.L/12.L,  8.L/12.L,
    0.L/12.L,  3.L/12.L,  0.L/12.L,  9.L/12.L,
    0.L/12.L,  2.L/12.L,  0.L/12.L, 10.L/12.L,
    0.L/12.L,  1.L/12.L,  0.L/12.L, 11.L/12.L,
    0.L/12.L,  0.L/12.L, 11.L/12.L,  1.L/12.L,
    0.L/12.L,  0.L/12.L, 10.L/12.L,  2.L/12.L,
    0.L/12.L,  0.L/12.L,  9.L/12.L,  3.L/12.L,
    0.L/12.L,  0.L/12.L,  8.L/12.L,  4.L/12.L,
    0.L/12.L,  0.L/12.L,  7.L/12.L,  5.L/12.L,
    0.L/12.L,  0.L/12.L,  6.L/12.L,  6.L/12.L,
    0.L/12.L,  0.L/12.L,  5.L/12.L,  7.L/12.L,
    0.L/12.L,  0.L/12.L,  4.L/12.L,  8.L/12.L,
    0.L/12.L,  0.L/12.L,  3.L/12.L,  9.L/12.L,
    0.L/12.L,  0.L/12.L,  2.L/12.L, 10.L/12.L,
    0.L/12.L,  0.L/12.L,  1.L/12.L, 11.L/12.L,
    0.L/12.L, 10.L/12.L,  1.L/12.L,  1.L/12.L,
    0.L/12.L,  9.L/12.L,  2.L/12.L,  1.L/12.L,
    0.L/12.L,  9.L/12.L,  1.L/12.L,  2.L/12.L,
    0.L/12.L,  8.L/12.L,  3.L/12.L,  1.L/12.L,
    0.L/12.L,  8.L/12.L,  2.L/12.L,  2.L/12.L,
    0.L/12.L,  8.L/12.L,  1.L/12.L,  3.L/12.L,
    0.L/12.L,  7.L/12.L,  4.L/12.L,  1.L/12.L,
    0.L/12.L,  7.L/12.L,  3.L/12.L,  2.L/12.L,
    0.L/12.L,  7.L/12.L,  2.L/12.L,  3.L/12.L,
    0.L/12.L,  7.L/12.L,  1.L/12.L,  4.L/12.L,
    0.L/12.L,  6.L/12.L,  5.L/12.L,  1.L/12.L,
    0.L/12.L,  6.L/12.L,  4.L/12.L,  2.L/12.L,
    0.L/12.L,  6.L/12.L,  3.L/12.L,  3.L/12.L,
    0.L/12.L,  6.L/12.L,  2.L/12.L,  4.L/12.L,
    0.L/12.L,  6.L/12.L,  1.L/12.L,  5.L/12.L,
    0.L/12.L,  5.L/12.L,  6.L/12.L,  1.L/12.L,
    0.L/12.L,  5.L/12.L,  5.L/12.L,  2.L/12.L,
    0.L/12.L,  5.L/12.L,  4.L/12.L,  3.L/12.L,
    0.L/12.L,  5.L/12.L,  3.L/12.L,  4.L/12.L,
    0.L/12.L,  5.L/12.L,  2.L/12.L,  5.L/12.L,
    0.L/12.L,  5.L/12.L,  1.L/12.L,  6.L/12.L,
    0.L/12.L,  4.L/12.L,  7.L/12.L,  1.L/12.L,
    0.L/12.L,  4.L/12.L,  6.L/12.L,  2.L/12.L,
    0.L/12.L,  4.L/12.L,  5.L/12.L,  3.L/12.L,
    0.L/12.L,  4.L/12.L,  4.L/12.L,  4.L/12.L,
    0.L/12.L,  4.L/12.L,  3.L/12.L,  5.L/12.L,
    0.L/12.L,  4.L/12.L,  2.L/12.L,  6.L/12.L,
    0.L/12.L,  4.L/12.L,  1.L/12.L,  7.L/12.L,
    0.L/12.L,  3.L/12.L,  8.L/12.L,  1.L/12.L,
    0.L/12.L,  3.L/12.L,  7.L/12.L,  2.L/12.L,
    0.L/12.L,  3.L/12.L,  6.L/12.L,  3.L/12.L,
    0.L/12.L,  3.L/12.L,  5.L/12.L,  4.L/12.L,
    0.L/12.L,  3.L/12.L,  4.L/12.L,  5.L/12.L,
    0.L/12.L,  3.L/12.L,  3.L/12.L,  6.L/12.L,
    0.L/12.L,  3.L/12.L,  2.L/12.L,  7.L/12.L,
    0.L/12.L,  3.L/12.L,  1.L/12.L,  8.L/12.L,
    0.L/12.L,  2.L/12.L,  9.L/12.L,  1.L/12.L,
    0.L/12.L,  2.L/12.L,  8.L/12.L,  2.L/12.L,
    0.L/12.L,  2.L/12.L,  7.L/12.L,  3.L/12.L,
    0.L/12.L,  2.L/12.L,  6.L/12.L,  4.L/12.L,
    0.L/12.L,  2.L/12.L,  5.L/12.L,  5.L/12.L,
    0.L/12.L,  2.L/12.L,  4.L/12.L,  6.L/12.L,
    0.L/12.L,  2.L/12.L,  3.L/12.L,  7.L/12.L,
    0.L/12.L,  2.L/12.L,  2.L/12.L,  8.L/12.L,
    0.L/12.L,  2.L/12.L,  1.L/12.L,  9.L/12.L,
    0.L/12.L,  1.L/12.L, 10.L/12.L,  1.L/12.L,
    0.L/12.L,  1.L/12.L,  9.L/12.L,  2.L/12.L,
    0.L/12.L,  1.L/12.L,  8.L/12.L,  3.L/12.L,
    0.L/12.L,  1.L/12.L,  7.L/12.L,  4.L/12.L,
    0.L/12.L,  1.L/12.L,  6.L/12.L,  5.L/12.L,
    0.L/12.L,  1.L/12.L,  5.L/12.L,  6.L/12.L,
    0.L/12.L,  1.L/12.L,  4.L/12.L,  7.L/12.L,
    0.L/12.L,  1.L/12.L,  3.L/12.L,  8.L/12.L,
    0.L/12.L,  1.L/12.L,  2.L/12.L,  9.L/12.L,
    0.L/12.L,  1.L/12.L,  1.L/12.L, 10.L/12.L,
   10.L/12.L,  0.L/12.L,  1.L/12.L,  1.L/12.L,
    9.L/12.L,  0.L/12.L,  2.L/12.L,  1.L/12.L,
    9.L/12.L,  0.L/12.L,  1.L/12.L,  2.L/12.L,
    8.L/12.L,  0.L/12.L,  3.L/12.L,  1.L/12.L,
    8.L/12.L,  0.L/12.L,  2.L/12.L,  2.L/12.L,
    8.L/12.L,  0.L/12.L,  1.L/12.L,  3.L/12.L,
    7.L/12.L,  0.L/12.L,  4.L/12.L,  1.L/12.L,
    7.L/12.L,  0.L/12.L,  3.L/12.L,  2.L/12.L,
    7.L/12.L,  0.L/12.L,  2.L/12.L,  3.L/12.L,
    7.L/12.L,  0.L/12.L,  1.L/12.L,  4.L/12.L,
    6.L/12.L,  0.L/12.L,  5.L/12.L,  1.L/12.L,
    6.L/12.L,  0.L/12.L,  4.L/12.L,  2.L/12.L,
    6.L/12.L,  0.L/12.L,  3.L/12.L,  3.L/12.L,
    6.L/12.L,  0.L/12.L,  2.L/12.L,  4.L/12.L,
    6.L/12.L,  0.L/12.L,  1.L/12.L,  5.L/12.L,
    5.L/12.L,  0.L/12.L,  6.L/12.L,  1.L/12.L,
    5.L/12.L,  0.L/12.L,  5.L/12.L,  2.L/12.L,
    5.L/12.L,  0.L/12.L,  4.L/12.L,  3.L/12.L,
    5.L/12.L,  0.L/12.L,  3.L/12.L,  4.L/12.L,
    5.L/12.L,  0.L/12.L,  2.L/12.L,  5.L/12.L,
    5.L/12.L,  0.L/12.L,  1.L/12.L,  6.L/12.L,
    4.L/12.L,  0.L/12.L,  7.L/12.L,  1.L/12.L,
    4.L/12.L,  0.L/12.L,  6.L/12.L,  2.L/12.L,
    4.L/12.L,  0.L/12.L,  5.L/12.L,  3.L/12.L,
    4.L/12.L,  0.L/12.L,  4.L/12.L,  4.L/12.L,
    4.L/12.L,  0.L/12.L,  3.L/12.L,  5.L/12.L,
    4.L/12.L,  0.L/12.L,  2.L/12.L,  6.L/12.L,
    4.L/12.L,  0.L/12.L,  1.L/12.L,  7.L/12.L,
    3.L/12.L,  0.L/12.L,  8.L/12.L,  1.L/12.L,
    3.L/12.L,  0.L/12.L,  7.L/12.L,  2.L/12.L,
    3.L/12.L,  0.L/12.L,  6.L/12.L,  3.L/12.L,
    3.L/12.L,  0.L/12.L,  5.L/12.L,  4.L/12.L,
    3.L/12.L,  0.L/12.L,  4.L/12.L,  5.L/12.L,
    3.L/12.L,  0.L/12.L,  3.L/12.L,  6.L/12.L,
    3.L/12.L,  0.L/12.L,  2.L/12.L,  7.L/12.L,
    3.L/12.L,  0.L/12.L,  1.L/12.L,  8.L/12.L,
    2.L/12.L,  0.L/12.L,  9.L/12.L,  1.L/12.L,
    2.L/12.L,  0.L/12.L,  8.L/12.L,  2.L/12.L,
    2.L/12.L,  0.L/12.L,  7.L/12.L,  3.L/12.L,
    2.L/12.L,  0.L/12.L,  6.L/12.L,  4.L/12.L,
    2.L/12.L,  0.L/12.L,  5.L/12.L,  5.L/12.L,
    2.L/12.L,  0.L/12.L,  4.L/12.L,  6.L/12.L,
    2.L/12.L,  0.L/12.L,  3.L/12.L,  7.L/12.L,
    2.L/12.L,  0.L/12.L,  2.L/12.L,  8.L/12.L,
    2.L/12.L,  0.L/12.L,  1.L/12.L,  9.L/12.L,
    1.L/12.L,  0.L/12.L, 10.L/12.L,  1.L/12.L,
    1.L/12.L,  0.L/12.L,  9.L/12.L,  2.L/12.L,
    1.L/12.L,  0.L/12.L,  8.L/12.L,  3.L/12.L,
    1.L/12.L,  0.L/12.L,  7.L/12.L,  4.L/12.L,
    1.L/12.L,  0.L/12.L,  6.L/12.L,  5.L/12.L,
    1.L/12.L,  0.L/12.L,  5.L/12.L,  6.L/12.L,
    1.L/12.L,  0.L/12.L,  4.L/12.L,  7.L/12.L,
    1.L/12.L,  0.L/12.L,  3.L/12.L,  8.L/12.L,
    1.L/12.L,  0.L/12.L,  2.L/12.L,  9.L/12.L,
    1.L/12.L,  0.L/12.L,  1.L/12.L, 10.L/12.L,
   10.L/12.L,  1.L/12.L,  0.L/12.L,  1.L/12.L,
    9.L/12.L,  2.L/12.L,  0.L/12.L,  1.L/12.L,
    9.L/12.L,  1.L/12.L,  0.L/12.L,  2.L/12.L,
    8.L/12.L,  3.L/12.L,  0.L/12.L,  1.L/12.L,
    8.L/12.L,  2.L/12.L,  0.L/12.L,  2.L/12.L,
    8.L/12.L,  1.L/12.L,  0.L/12.L,  3.L/12.L,
    7.L/12.L,  4.L/12.L,  0.L/12.L,  1.L/12.L,
    7.L/12.L,  3.L/12.L,  0.L/12.L,  2.L/12.L,
    7.L/12.L,  2.L/12.L,  0.L/12.L,  3.L/12.L,
    7.L/12.L,  1.L/12.L,  0.L/12.L,  4.L/12.L,
    6.L/12.L,  5.L/12.L,  0.L/12.L,  1.L/12.L,
    6.L/12.L,  4.L/12.L,  0.L/12.L,  2.L/12.L,
    6.L/12.L,  3.L/12.L,  0.L/12.L,  3.L/12.L,
    6.L/12.L,  2.L/12.L,  0.L/12.L,  4.L/12.L,
    6.L/12.L,  1.L/12.L,  0.L/12.L,  5.L/12.L,
    5.L/12.L,  6.L/12.L,  0.L/12.L,  1.L/12.L,
    5.L/12.L,  5.L/12.L,  0.L/12.L,  2.L/12.L,
    5.L/12.L,  4.L/12.L,  0.L/12.L,  3.L/12.L,
    5.L/12.L,  3.L/12.L,  0.L/12.L,  4.L/12.L,
    5.L/12.L,  2.L/12.L,  0.L/12.L,  5.L/12.L,
    5.L/12.L,  1.L/12.L,  0.L/12.L,  6.L/12.L,
    4.L/12.L,  7.L/12.L,  0.L/12.L,  1.L/12.L,
    4.L/12.L,  6.L/12.L,  0.L/12.L,  2.L/12.L,
    4.L/12.L,  5.L/12.L,  0.L/12.L,  3.L/12.L,
    4.L/12.L,  4.L/12.L,  0.L/12.L,  4.L/12.L,
    4.L/12.L,  3.L/12.L,  0.L/12.L,  5.L/12.L,
    4.L/12.L,  2.L/12.L,  0.L/12.L,  6.L/12.L,
    4.L/12.L,  1.L/12.L,  0.L/12.L,  7.L/12.L,
    3.L/12.L,  8.L/12.L,  0.L/12.L,  1.L/12.L,
    3.L/12.L,  7.L/12.L,  0.L/12.L,  2.L/12.L,
    3.L/12.L,  6.L/12.L,  0.L/12.L,  3.L/12.L,
    3.L/12.L,  5.L/12.L,  0.L/12.L,  4.L/12.L,
    3.L/12.L,  4.L/12.L,  0.L/12.L,  5.L/12.L,
    3.L/12.L,  3.L/12.L,  0.L/12.L,  6.L/12.L,
    3.L/12.L,  2.L/12.L,  0.L/12.L,  7.L/12.L,
    3.L/12.L,  1.L/12.L,  0.L/12.L,  8.L/12.L,
    2.L/12.L,  9.L/12.L,  0.L/12.L,  1.L/12.L,
    2.L/12.L,  8.L/12.L,  0.L/12.L,  2.L/12.L,
    2.L/12.L,  7.L/12.L,  0.L/12.L,  3.L/12.L,
    2.L/12.L,  6.L/12.L,  0.L/12.L,  4.L/12.L,
    2.L/12.L,  5.L/12.L,  0.L/12.L,  5.L/12.L,
    2.L/12.L,  4.L/12.L,  0.L/12.L,  6.L/12.L,
    2.L/12.L,  3.L/12.L,  0.L/12.L,  7.L/12.L,
    2.L/12.L,  2.L/12.L,  0.L/12.L,  8.L/12.L,
    2.L/12.L,  1.L/12.L,  0.L/12.L,  9.L/12.L,
    1.L/12.L, 10.L/12.L,  0.L/12.L,  1.L/12.L,
    1.L/12.L,  9.L/12.L,  0.L/12.L,  2.L/12.L,
    1.L/12.L,  8.L/12.L,  0.L/12.L,  3.L/12.L,
    1.L/12.L,  7.L/12.L,  0.L/12.L,  4.L/12.L,
    1.L/12.L,  6.L/12.L,  0.L/12.L,  5.L/12.L,
    1.L/12.L,  5.L/12.L,  0.L/12.L,  6.L/12.L,
    1.L/12.L,  4.L/12.L,  0.L/12.L,  7.L/12.L,
    1.L/12.L,  3.L/12.L,  0.L/12.L,  8.L/12.L,
    1.L/12.L,  2.L/12.L,  0.L/12.L,  9.L/12.L,
    1.L/12.L,  1.L/12.L,  0.L/12.L, 10.L/12.L,
   10.L/12.L,  1.L/12.L,  1.L/12.L,  0.L/12.L,
    9.L/12.L,  2.L/12.L,  1.L/12.L,  0.L/12.L,
    9.L/12.L,  1.L/12.L,  2.L/12.L,  0.L/12.L,
    8.L/12.L,  3.L/12.L,  1.L/12.L,  0.L/12.L,
    8.L/12.L,  2.L/12.L,  2.L/12.L,  0.L/12.L,
    8.L/12.L,  1.L/12.L,  3.L/12.L,  0.L/12.L,
    7.L/12.L,  4.L/12.L,  1.L/12.L,  0.L/12.L,
    7.L/12.L,  3.L/12.L,  2.L/12.L,  0.L/12.L,
    7.L/12.L,  2.L/12.L,  3.L/12.L,  0.L/12.L,
    7.L/12.L,  1.L/12.L,  4.L/12.L,  0.L/12.L,
    6.L/12.L,  5.L/12.L,  1.L/12.L,  0.L/12.L,
    6.L/12.L,  4.L/12.L,  2.L/12.L,  0.L/12.L,
    6.L/12.L,  3.L/12.L,  3.L/12.L,  0.L/12.L,
    6.L/12.L,  2.L/12.L,  4.L/12.L,  0.L/12.L,
    6.L/12.L,  1.L/12.L,  5.L/12.L,  0.L/12.L,
    5.L/12.L,  6.L/12.L,  1.L/12.L,  0.L/12.L,
    5.L/12.L,  5.L/12.L,  2.L/12.L,  0.L/12.L,
    5.L/12.L,  4.L/12.L,  3.L/12.L,  0.L/12.L,
    5.L/12.L,  3.L/12.L,  4.L/12.L,  0.L/12.L,
    5.L/12.L,  2.L/12.L,  5.L/12.L,  0.L/12.L,
    5.L/12.L,  1.L/12.L,  6.L/12.L,  0.L/12.L,
    4.L/12.L,  7.L/12.L,  1.L/12.L,  0.L/12.L,
    4.L/12.L,  6.L/12.L,  2.L/12.L,  0.L/12.L,
    4.L/12.L,  5.L/12.L,  3.L/12.L,  0.L/12.L,
    4.L/12.L,  4.L/12.L,  4.L/12.L,  0.L/12.L,
    4.L/12.L,  3.L/12.L,  5.L/12.L,  0.L/12.L,
    4.L/12.L,  2.L/12.L,  6.L/12.L,  0.L/12.L,
    4.L/12.L,  1.L/12.L,  7.L/12.L,  0.L/12.L,
    3.L/12.L,  8.L/12.L,  1.L/12.L,  0.L/12.L,
    3.L/12.L,  7.L/12.L,  2.L/12.L,  0.L/12.L,
    3.L/12.L,  6.L/12.L,  3.L/12.L,  0.L/12.L,
    3.L/12.L,  5.L/12.L,  4.L/12.L,  0.L/12.L,
    3.L/12.L,  4.L/12.L,  5.L/12.L,  0.L/12.L,
    3.L/12.L,  3.L/12.L,  6.L/12.L,  0.L/12.L,
    3.L/12.L,  2.L/12.L,  7.L/12.L,  0.L/12.L,
    3.L/12.L,  1.L/12.L,  8.L/12.L,  0.L/12.L,
    2.L/12.L,  9.L/12.L,  1.L/12.L,  0.L/12.L,
    2.L/12.L,  8.L/12.L,  2.L/12.L,  0.L/12.L,
    2.L/12.L,  7.L/12.L,  3.L/12.L,  0.L/12.L,
    2.L/12.L,  6.L/12.L,  4.L/12.L,  0.L/12.L,
    2.L/12.L,  5.L/12.L,  5.L/12.L,  0.L/12.L,
    2.L/12.L,  4.L/12.L,  6.L/12.L,  0.L/12.L,
    2.L/12.L,  3.L/12.L,  7.L/12.L,  0.L/12.L,
    2.L/12.L,  2.L/12.L,  8.L/12.L,  0.L/12.L,
    2.L/12.L,  1.L/12.L,  9.L/12.L,  0.L/12.L,
    1.L/12.L, 10.L/12.L,  1.L/12.L,  0.L/12.L,
    1.L/12.L,  9.L/12.L,  2.L/12.L,  0.L/12.L,
    1.L/12.L,  8.L/12.L,  3.L/12.L,  0.L/12.L,
    1.L/12.L,  7.L/12.L,  4.L/12.L,  0.L/12.L,
    1.L/12.L,  6.L/12.L,  5.L/12.L,  0.L/12.L,
    1.L/12.L,  5.L/12.L,  6.L/12.L,  0.L/12.L,
    1.L/12.L,  4.L/12.L,  7.L/12.L,  0.L/12.L,
    1.L/12.L,  3.L/12.L,  8.L/12.L,  0.L/12.L,
    1.L/12.L,  2.L/12.L,  9.L/12.L,  0.L/12.L,
    1.L/12.L,  1.L/12.L, 10.L/12.L,  0.L/12.L,
    9.L/12.L,  1.L/12.L,  1.L/12.L,  1.L/12.L,
    8.L/12.L,  2.L/12.L,  1.L/12.L,  1.L/12.L,
    8.L/12.L,  1.L/12.L,  2.L/12.L,  1.L/12.L,
    8.L/12.L,  1.L/12.L,  1.L/12.L,  2.L/12.L,
    7.L/12.L,  3.L/12.L,  1.L/12.L,  1.L/12.L,
    7.L/12.L,  2.L/12.L,  2.L/12.L,  1.L/12.L,
    7.L/12.L,  2.L/12.L,  1.L/12.L,  2.L/12.L,
    7.L/12.L,  1.L/12.L,  3.L/12.L,  1.L/12.L,
    7.L/12.L,  1.L/12.L,  2.L/12.L,  2.L/12.L,
    7.L/12.L,  1.L/12.L,  1.L/12.L,  3.L/12.L,
    6.L/12.L,  4.L/12.L,  1.L/12.L,  1.L/12.L,
    6.L/12.L,  3.L/12.L,  2.L/12.L,  1.L/12.L,
    6.L/12.L,  3.L/12.L,  1.L/12.L,  2.L/12.L,
    6.L/12.L,  2.L/12.L,  3.L/12.L,  1.L/12.L,
    6.L/12.L,  2.L/12.L,  2.L/12.L,  2.L/12.L,
    6.L/12.L,  2.L/12.L,  1.L/12.L,  3.L/12.L,
    6.L/12.L,  1.L/12.L,  4.L/12.L,  1.L/12.L,
    6.L/12.L,  1.L/12.L,  3.L/12.L,  2.L/12.L,
    6.L/12.L,  1.L/12.L,  2.L/12.L,  3.L/12.L,
    6.L/12.L,  1.L/12.L,  1.L/12.L,  4.L/12.L,
    5.L/12.L,  5.L/12.L,  1.L/12.L,  1.L/12.L,
    5.L/12.L,  4.L/12.L,  2.L/12.L,  1.L/12.L,
    5.L/12.L,  4.L/12.L,  1.L/12.L,  2.L/12.L,
    5.L/12.L,  3.L/12.L,  3.L/12.L,  1.L/12.L,
    5.L/12.L,  3.L/12.L,  2.L/12.L,  2.L/12.L,
    5.L/12.L,  3.L/12.L,  1.L/12.L,  3.L/12.L,
    5.L/12.L,  2.L/12.L,  4.L/12.L,  1.L/12.L,
    5.L/12.L,  2.L/12.L,  3.L/12.L,  2.L/12.L,
    5.L/12.L,  2.L/12.L,  2.L/12.L,  3.L/12.L,
    5.L/12.L,  2.L/12.L,  1.L/12.L,  4.L/12.L,
    5.L/12.L,  1.L/12.L,  5.L/12.L,  1.L/12.L,
    5.L/12.L,  1.L/12.L,  4.L/12.L,  2.L/12.L,
    5.L/12.L,  1.L/12.L,  3.L/12.L,  3.L/12.L,
    5.L/12.L,  1.L/12.L,  2.L/12.L,  4.L/12.L,
    5.L/12.L,  1.L/12.L,  1.L/12.L,  5.L/12.L,
    4.L/12.L,  6.L/12.L,  1.L/12.L,  1.L/12.L,
    4.L/12.L,  5.L/12.L,  2.L/12.L,  1.L/12.L,
    4.L/12.L,  5.L/12.L,  1.L/12.L,  2.L/12.L,
    4.L/12.L,  4.L/12.L,  3.L/12.L,  1.L/12.L,
    4.L/12.L,  4.L/12.L,  2.L/12.L,  2.L/12.L,
    4.L/12.L,  4.L/12.L,  1.L/12.L,  3.L/12.L,
    4.L/12.L,  3.L/12.L,  4.L/12.L,  1.L/12.L,
    4.L/12.L,  3.L/12.L,  3.L/12.L,  2.L/12.L,
    4.L/12.L,  3.L/12.L,  2.L/12.L,  3.L/12.L,
    4.L/12.L,  3.L/12.L,  1.L/12.L,  4.L/12.L,
    4.L/12.L,  2.L/12.L,  5.L/12.L,  1.L/12.L,
    4.L/12.L,  2.L/12.L,  4.L/12.L,  2.L/12.L,
    4.L/12.L,  2.L/12.L,  3.L/12.L,  3.L/12.L,
    4.L/12.L,  2.L/12.L,  2.L/12.L,  4.L/12.L,
    4.L/12.L,  2.L/12.L,  1.L/12.L,  5.L/12.L,
    4.L/12.L,  1.L/12.L,  6.L/12.L,  1.L/12.L,
    4.L/12.L,  1.L/12.L,  5.L/12.L,  2.L/12.L,
    4.L/12.L,  1.L/12.L,  4.L/12.L,  3.L/12.L,
    4.L/12.L,  1.L/12.L,  3.L/12.L,  4.L/12.L,
    4.L/12.L,  1.L/12.L,  2.L/12.L,  5.L/12.L,
    4.L/12.L,  1.L/12.L,  1.L/12.L,  6.L/12.L,
    3.L/12.L,  7.L/12.L,  1.L/12.L,  1.L/12.L,
    3.L/12.L,  6.L/12.L,  2.L/12.L,  1.L/12.L,
    3.L/12.L,  6.L/12.L,  1.L/12.L,  2.L/12.L,
    3.L/12.L,  5.L/12.L,  3.L/12.L,  1.L/12.L,
    3.L/12.L,  5.L/12.L,  2.L/12.L,  2.L/12.L,
    3.L/12.L,  5.L/12.L,  1.L/12.L,  3.L/12.L,
    3.L/12.L,  4.L/12.L,  4.L/12.L,  1.L/12.L,
    3.L/12.L,  4.L/12.L,  3.L/12.L,  2.L/12.L,
    3.L/12.L,  4.L/12.L,  2.L/12.L,  3.L/12.L,
    3.L/12.L,  4.L/12.L,  1.L/12.L,  4.L/12.L,
    3.L/12.L,  3.L/12.L,  5.L/12.L,  1.L/12.L,
    3.L/12.L,  3.L/12.L,  4.L/12.L,  2.L/12.L,
    3.L/12.L,  3.L/12.L,  3.L/12.L,  3.L/12.L,
    3.L/12.L,  3.L/12.L,  2.L/12.L,  4.L/12.L,
    3.L/12.L,  3.L/12.L,  1.L/12.L,  5.L/12.L,
    3.L/12.L,  2.L/12.L,  6.L/12.L,  1.L/12.L,
    3.L/12.L,  2.L/12.L,  5.L/12.L,  2.L/12.L,
    3.L/12.L,  2.L/12.L,  4.L/12.L,  3.L/12.L,
    3.L/12.L,  2.L/12.L,  3.L/12.L,  4.L/12.L,
    3.L/12.L,  2.L/12.L,  2.L/12.L,  5.L/12.L,
    3.L/12.L,  2.L/12.L,  1.L/12.L,  6.L/12.L,
    3.L/12.L,  1.L/12.L,  7.L/12.L,  1.L/12.L,
    3.L/12.L,  1.L/12.L,  6.L/12.L,  2.L/12.L,
    3.L/12.L,  1.L/12.L,  5.L/12.L,  3.L/12.L,
    3.L/12.L,  1.L/12.L,  4.L/12.L,  4.L/12.L,
    3.L/12.L,  1.L/12.L,  3.L/12.L,  5.L/12.L,
    3.L/12.L,  1.L/12.L,  2.L/12.L,  6.L/12.L,
    3.L/12.L,  1.L/12.L,  1.L/12.L,  7.L/12.L,
    2.L/12.L,  8.L/12.L,  1.L/12.L,  1.L/12.L,
    2.L/12.L,  7.L/12.L,  2.L/12.L,  1.L/12.L,
    2.L/12.L,  7.L/12.L,  1.L/12.L,  2.L/12.L,
    2.L/12.L,  6.L/12.L,  3.L/12.L,  1.L/12.L,
    2.L/12.L,  6.L/12.L,  2.L/12.L,  2.L/12.L,
    2.L/12.L,  6.L/12.L,  1.L/12.L,  3.L/12.L,
    2.L/12.L,  5.L/12.L,  4.L/12.L,  1.L/12.L,
    2.L/12.L,  5.L/12.L,  3.L/12.L,  2.L/12.L,
    2.L/12.L,  5.L/12.L,  2.L/12.L,  3.L/12.L,
    2.L/12.L,  5.L/12.L,  1.L/12.L,  4.L/12.L,
    2.L/12.L,  4.L/12.L,  5.L/12.L,  1.L/12.L,
    2.L/12.L,  4.L/12.L,  4.L/12.L,  2.L/12.L,
    2.L/12.L,  4.L/12.L,  3.L/12.L,  3.L/12.L,
    2.L/12.L,  4.L/12.L,  2.L/12.L,  4.L/12.L,
    2.L/12.L,  4.L/12.L,  1.L/12.L,  5.L/12.L,
    2.L/12.L,  3.L/12.L,  6.L/12.L,  1.L/12.L,
    2.L/12.L,  3.L/12.L,  5.L/12.L,  2.L/12.L,
    2.L/12.L,  3.L/12.L,  4.L/12.L,  3.L/12.L,
    2.L/12.L,  3.L/12.L,  3.L/12.L,  4.L/12.L,
    2.L/12.L,  3.L/12.L,  2.L/12.L,  5.L/12.L,
    2.L/12.L,  3.L/12.L,  1.L/12.L,  6.L/12.L,
    2.L/12.L,  2.L/12.L,  7.L/12.L,  1.L/12.L,
    2.L/12.L,  2.L/12.L,  6.L/12.L,  2.L/12.L,
    2.L/12.L,  2.L/12.L,  5.L/12.L,  3.L/12.L,
    2.L/12.L,  2.L/12.L,  4.L/12.L,  4.L/12.L,
    2.L/12.L,  2.L/12.L,  3.L/12.L,  5.L/12.L,
    2.L/12.L,  2.L/12.L,  2.L/12.L,  6.L/12.L,
    2.L/12.L,  2.L/12.L,  1.L/12.L,  7.L/12.L,
    2.L/12.L,  1.L/12.L,  8.L/12.L,  1.L/12.L,
    2.L/12.L,  1.L/12.L,  7.L/12.L,  2.L/12.L,
    2.L/12.L,  1.L/12.L,  6.L/12.L,  3.L/12.L,
    2.L/12.L,  1.L/12.L,  5.L/12.L,  4.L/12.L,
    2.L/12.L,  1.L/12.L,  4.L/12.L,  5.L/12.L,
    2.L/12.L,  1.L/12.L,  3.L/12.L,  6.L/12.L,
    2.L/12.L,  1.L/12.L,  2.L/12.L,  7.L/12.L,
    2.L/12.L,  1.L/12.L,  1.L/12.L,  8.L/12.L,
    1.L/12.L,  9.L/12.L,  1.L/12.L,  1.L/12.L,
    1.L/12.L,  8.L/12.L,  2.L/12.L,  1.L/12.L,
    1.L/12.L,  8.L/12.L,  1.L/12.L,  2.L/12.L,
    1.L/12.L,  7.L/12.L,  3.L/12.L,  1.L/12.L,
    1.L/12.L,  7.L/12.L,  2.L/12.L,  2.L/12.L,
    1.L/12.L,  7.L/12.L,  1.L/12.L,  3.L/12.L,
    1.L/12.L,  6.L/12.L,  4.L/12.L,  1.L/12.L,
    1.L/12.L,  6.L/12.L,  3.L/12.L,  2.L/12.L,
    1.L/12.L,  6.L/12.L,  2.L/12.L,  3.L/12.L,
    1.L/12.L,  6.L/12.L,  1.L/12.L,  4.L/12.L,
    1.L/12.L,  5.L/12.L,  5.L/12.L,  1.L/12.L,
    1.L/12.L,  5.L/12.L,  4.L/12.L,  2.L/12.L,
    1.L/12.L,  5.L/12.L,  3.L/12.L,  3.L/12.L,
    1.L/12.L,  5.L/12.L,  2.L/12.L,  4.L/12.L,
    1.L/12.L,  5.L/12.L,  1.L/12.L,  5.L/12.L,
    1.L/12.L,  4.L/12.L,  6.L/12.L,  1.L/12.L,
    1.L/12.L,  4.L/12.L,  5.L/12.L,  2.L/12.L,
    1.L/12.L,  4.L/12.L,  4.L/12.L,  3.L/12.L,
    1.L/12.L,  4.L/12.L,  3.L/12.L,  4.L/12.L,
    1.L/12.L,  4.L/12.L,  2.L/12.L,  5.L/12.L,
    1.L/12.L,  4.L/12.L,  1.L/12.L,  6.L/12.L,
    1.L/12.L,  3.L/12.L,  7.L/12.L,  1.L/12.L,
    1.L/12.L,  3.L/12.L,  6.L/12.L,  2.L/12.L,
    1.L/12.L,  3.L/12.L,  5.L/12.L,  3.L/12.L,
    1.L/12.L,  3.L/12.L,  4.L/12.L,  4.L/12.L,
    1.L/12.L,  3.L/12.L,  3.L/12.L,  5.L/12.L,
    1.L/12.L,  3.L/12.L,  2.L/12.L,  6.L/12.L,
    1.L/12.L,  3.L/12.L,  1.L/12.L,  7.L/12.L,
    1.L/12.L,  2.L/12.L,  8.L/12.L,  1.L/12.L,
    1.L/12.L,  2.L/12.L,  7.L/12.L,  2.L/12.L,
    1.L/12.L,  2.L/12.L,  6.L/12.L,  3.L/12.L,
    1.L/12.L,  2.L/12.L,  5.L/12.L,  4.L/12.L,
    1.L/12.L,  2.L/12.L,  4.L/12.L,  5.L/12.L,
    1.L/12.L,  2.L/12.L,  3.L/12.L,  6.L/12.L,
    1.L/12.L,  2.L/12.L,  2.L/12.L,  7.L/12.L,
    1.L/12.L,  2.L/12.L,  1.L/12.L,  8.L/12.L,
    1.L/12.L,  1.L/12.L,  9.L/12.L,  1.L/12.L,
    1.L/12.L,  1.L/12.L,  8.L/12.L,  2.L/12.L,
    1.L/12.L,  1.L/12.L,  7.L/12.L,  3.L/12.L,
    1.L/12.L,  1.L/12.L,  6.L/12.L,  4.L/12.L,
    1.L/12.L,  1.L/12.L,  5.L/12.L,  5.L/12.L,
    1.L/12.L,  1.L/12.L,  4.L/12.L,  6.L/12.L,
    1.L/12.L,  1.L/12.L,  3.L/12.L,  7.L/12.L,
    1.L/12.L,  1.L/12.L,  2.L/12.L,  8.L/12.L,
    1.L/12.L,  1.L/12.L,  1.L/12.L,  9.L/12.L
};
DOF_TYPE DOF_DG12_ = {
    DofCache, "DG12", DG12_points, NULL,
    DOF_DG11, DOF_P12, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 455, 12, -1, 1,
    0, 0, 0, 455
};

static FLOAT DG13_points[] = {
   13.L/13.L,  0.L/13.L,  0.L/13.L,  0.L/13.L,
    0.L/13.L, 13.L/13.L,  0.L/13.L,  0.L/13.L,
    0.L/13.L,  0.L/13.L, 13.L/13.L,  0.L/13.L,
    0.L/13.L,  0.L/13.L,  0.L/13.L, 13.L/13.L,
   12.L/13.L,  1.L/13.L,  0.L/13.L,  0.L/13.L,
   11.L/13.L,  2.L/13.L,  0.L/13.L,  0.L/13.L,
   10.L/13.L,  3.L/13.L,  0.L/13.L,  0.L/13.L,
    9.L/13.L,  4.L/13.L,  0.L/13.L,  0.L/13.L,
    8.L/13.L,  5.L/13.L,  0.L/13.L,  0.L/13.L,
    7.L/13.L,  6.L/13.L,  0.L/13.L,  0.L/13.L,
    6.L/13.L,  7.L/13.L,  0.L/13.L,  0.L/13.L,
    5.L/13.L,  8.L/13.L,  0.L/13.L,  0.L/13.L,
    4.L/13.L,  9.L/13.L,  0.L/13.L,  0.L/13.L,
    3.L/13.L, 10.L/13.L,  0.L/13.L,  0.L/13.L,
    2.L/13.L, 11.L/13.L,  0.L/13.L,  0.L/13.L,
    1.L/13.L, 12.L/13.L,  0.L/13.L,  0.L/13.L,
   12.L/13.L,  0.L/13.L,  1.L/13.L,  0.L/13.L,
   11.L/13.L,  0.L/13.L,  2.L/13.L,  0.L/13.L,
   10.L/13.L,  0.L/13.L,  3.L/13.L,  0.L/13.L,
    9.L/13.L,  0.L/13.L,  4.L/13.L,  0.L/13.L,
    8.L/13.L,  0.L/13.L,  5.L/13.L,  0.L/13.L,
    7.L/13.L,  0.L/13.L,  6.L/13.L,  0.L/13.L,
    6.L/13.L,  0.L/13.L,  7.L/13.L,  0.L/13.L,
    5.L/13.L,  0.L/13.L,  8.L/13.L,  0.L/13.L,
    4.L/13.L,  0.L/13.L,  9.L/13.L,  0.L/13.L,
    3.L/13.L,  0.L/13.L, 10.L/13.L,  0.L/13.L,
    2.L/13.L,  0.L/13.L, 11.L/13.L,  0.L/13.L,
    1.L/13.L,  0.L/13.L, 12.L/13.L,  0.L/13.L,
   12.L/13.L,  0.L/13.L,  0.L/13.L,  1.L/13.L,
   11.L/13.L,  0.L/13.L,  0.L/13.L,  2.L/13.L,
   10.L/13.L,  0.L/13.L,  0.L/13.L,  3.L/13.L,
    9.L/13.L,  0.L/13.L,  0.L/13.L,  4.L/13.L,
    8.L/13.L,  0.L/13.L,  0.L/13.L,  5.L/13.L,
    7.L/13.L,  0.L/13.L,  0.L/13.L,  6.L/13.L,
    6.L/13.L,  0.L/13.L,  0.L/13.L,  7.L/13.L,
    5.L/13.L,  0.L/13.L,  0.L/13.L,  8.L/13.L,
    4.L/13.L,  0.L/13.L,  0.L/13.L,  9.L/13.L,
    3.L/13.L,  0.L/13.L,  0.L/13.L, 10.L/13.L,
    2.L/13.L,  0.L/13.L,  0.L/13.L, 11.L/13.L,
    1.L/13.L,  0.L/13.L,  0.L/13.L, 12.L/13.L,
    0.L/13.L, 12.L/13.L,  1.L/13.L,  0.L/13.L,
    0.L/13.L, 11.L/13.L,  2.L/13.L,  0.L/13.L,
    0.L/13.L, 10.L/13.L,  3.L/13.L,  0.L/13.L,
    0.L/13.L,  9.L/13.L,  4.L/13.L,  0.L/13.L,
    0.L/13.L,  8.L/13.L,  5.L/13.L,  0.L/13.L,
    0.L/13.L,  7.L/13.L,  6.L/13.L,  0.L/13.L,
    0.L/13.L,  6.L/13.L,  7.L/13.L,  0.L/13.L,
    0.L/13.L,  5.L/13.L,  8.L/13.L,  0.L/13.L,
    0.L/13.L,  4.L/13.L,  9.L/13.L,  0.L/13.L,
    0.L/13.L,  3.L/13.L, 10.L/13.L,  0.L/13.L,
    0.L/13.L,  2.L/13.L, 11.L/13.L,  0.L/13.L,
    0.L/13.L,  1.L/13.L, 12.L/13.L,  0.L/13.L,
    0.L/13.L, 12.L/13.L,  0.L/13.L,  1.L/13.L,
    0.L/13.L, 11.L/13.L,  0.L/13.L,  2.L/13.L,
    0.L/13.L, 10.L/13.L,  0.L/13.L,  3.L/13.L,
    0.L/13.L,  9.L/13.L,  0.L/13.L,  4.L/13.L,
    0.L/13.L,  8.L/13.L,  0.L/13.L,  5.L/13.L,
    0.L/13.L,  7.L/13.L,  0.L/13.L,  6.L/13.L,
    0.L/13.L,  6.L/13.L,  0.L/13.L,  7.L/13.L,
    0.L/13.L,  5.L/13.L,  0.L/13.L,  8.L/13.L,
    0.L/13.L,  4.L/13.L,  0.L/13.L,  9.L/13.L,
    0.L/13.L,  3.L/13.L,  0.L/13.L, 10.L/13.L,
    0.L/13.L,  2.L/13.L,  0.L/13.L, 11.L/13.L,
    0.L/13.L,  1.L/13.L,  0.L/13.L, 12.L/13.L,
    0.L/13.L,  0.L/13.L, 12.L/13.L,  1.L/13.L,
    0.L/13.L,  0.L/13.L, 11.L/13.L,  2.L/13.L,
    0.L/13.L,  0.L/13.L, 10.L/13.L,  3.L/13.L,
    0.L/13.L,  0.L/13.L,  9.L/13.L,  4.L/13.L,
    0.L/13.L,  0.L/13.L,  8.L/13.L,  5.L/13.L,
    0.L/13.L,  0.L/13.L,  7.L/13.L,  6.L/13.L,
    0.L/13.L,  0.L/13.L,  6.L/13.L,  7.L/13.L,
    0.L/13.L,  0.L/13.L,  5.L/13.L,  8.L/13.L,
    0.L/13.L,  0.L/13.L,  4.L/13.L,  9.L/13.L,
    0.L/13.L,  0.L/13.L,  3.L/13.L, 10.L/13.L,
    0.L/13.L,  0.L/13.L,  2.L/13.L, 11.L/13.L,
    0.L/13.L,  0.L/13.L,  1.L/13.L, 12.L/13.L,
    0.L/13.L, 11.L/13.L,  1.L/13.L,  1.L/13.L,
    0.L/13.L, 10.L/13.L,  2.L/13.L,  1.L/13.L,
    0.L/13.L, 10.L/13.L,  1.L/13.L,  2.L/13.L,
    0.L/13.L,  9.L/13.L,  3.L/13.L,  1.L/13.L,
    0.L/13.L,  9.L/13.L,  2.L/13.L,  2.L/13.L,
    0.L/13.L,  9.L/13.L,  1.L/13.L,  3.L/13.L,
    0.L/13.L,  8.L/13.L,  4.L/13.L,  1.L/13.L,
    0.L/13.L,  8.L/13.L,  3.L/13.L,  2.L/13.L,
    0.L/13.L,  8.L/13.L,  2.L/13.L,  3.L/13.L,
    0.L/13.L,  8.L/13.L,  1.L/13.L,  4.L/13.L,
    0.L/13.L,  7.L/13.L,  5.L/13.L,  1.L/13.L,
    0.L/13.L,  7.L/13.L,  4.L/13.L,  2.L/13.L,
    0.L/13.L,  7.L/13.L,  3.L/13.L,  3.L/13.L,
    0.L/13.L,  7.L/13.L,  2.L/13.L,  4.L/13.L,
    0.L/13.L,  7.L/13.L,  1.L/13.L,  5.L/13.L,
    0.L/13.L,  6.L/13.L,  6.L/13.L,  1.L/13.L,
    0.L/13.L,  6.L/13.L,  5.L/13.L,  2.L/13.L,
    0.L/13.L,  6.L/13.L,  4.L/13.L,  3.L/13.L,
    0.L/13.L,  6.L/13.L,  3.L/13.L,  4.L/13.L,
    0.L/13.L,  6.L/13.L,  2.L/13.L,  5.L/13.L,
    0.L/13.L,  6.L/13.L,  1.L/13.L,  6.L/13.L,
    0.L/13.L,  5.L/13.L,  7.L/13.L,  1.L/13.L,
    0.L/13.L,  5.L/13.L,  6.L/13.L,  2.L/13.L,
    0.L/13.L,  5.L/13.L,  5.L/13.L,  3.L/13.L,
    0.L/13.L,  5.L/13.L,  4.L/13.L,  4.L/13.L,
    0.L/13.L,  5.L/13.L,  3.L/13.L,  5.L/13.L,
    0.L/13.L,  5.L/13.L,  2.L/13.L,  6.L/13.L,
    0.L/13.L,  5.L/13.L,  1.L/13.L,  7.L/13.L,
    0.L/13.L,  4.L/13.L,  8.L/13.L,  1.L/13.L,
    0.L/13.L,  4.L/13.L,  7.L/13.L,  2.L/13.L,
    0.L/13.L,  4.L/13.L,  6.L/13.L,  3.L/13.L,
    0.L/13.L,  4.L/13.L,  5.L/13.L,  4.L/13.L,
    0.L/13.L,  4.L/13.L,  4.L/13.L,  5.L/13.L,
    0.L/13.L,  4.L/13.L,  3.L/13.L,  6.L/13.L,
    0.L/13.L,  4.L/13.L,  2.L/13.L,  7.L/13.L,
    0.L/13.L,  4.L/13.L,  1.L/13.L,  8.L/13.L,
    0.L/13.L,  3.L/13.L,  9.L/13.L,  1.L/13.L,
    0.L/13.L,  3.L/13.L,  8.L/13.L,  2.L/13.L,
    0.L/13.L,  3.L/13.L,  7.L/13.L,  3.L/13.L,
    0.L/13.L,  3.L/13.L,  6.L/13.L,  4.L/13.L,
    0.L/13.L,  3.L/13.L,  5.L/13.L,  5.L/13.L,
    0.L/13.L,  3.L/13.L,  4.L/13.L,  6.L/13.L,
    0.L/13.L,  3.L/13.L,  3.L/13.L,  7.L/13.L,
    0.L/13.L,  3.L/13.L,  2.L/13.L,  8.L/13.L,
    0.L/13.L,  3.L/13.L,  1.L/13.L,  9.L/13.L,
    0.L/13.L,  2.L/13.L, 10.L/13.L,  1.L/13.L,
    0.L/13.L,  2.L/13.L,  9.L/13.L,  2.L/13.L,
    0.L/13.L,  2.L/13.L,  8.L/13.L,  3.L/13.L,
    0.L/13.L,  2.L/13.L,  7.L/13.L,  4.L/13.L,
    0.L/13.L,  2.L/13.L,  6.L/13.L,  5.L/13.L,
    0.L/13.L,  2.L/13.L,  5.L/13.L,  6.L/13.L,
    0.L/13.L,  2.L/13.L,  4.L/13.L,  7.L/13.L,
    0.L/13.L,  2.L/13.L,  3.L/13.L,  8.L/13.L,
    0.L/13.L,  2.L/13.L,  2.L/13.L,  9.L/13.L,
    0.L/13.L,  2.L/13.L,  1.L/13.L, 10.L/13.L,
    0.L/13.L,  1.L/13.L, 11.L/13.L,  1.L/13.L,
    0.L/13.L,  1.L/13.L, 10.L/13.L,  2.L/13.L,
    0.L/13.L,  1.L/13.L,  9.L/13.L,  3.L/13.L,
    0.L/13.L,  1.L/13.L,  8.L/13.L,  4.L/13.L,
    0.L/13.L,  1.L/13.L,  7.L/13.L,  5.L/13.L,
    0.L/13.L,  1.L/13.L,  6.L/13.L,  6.L/13.L,
    0.L/13.L,  1.L/13.L,  5.L/13.L,  7.L/13.L,
    0.L/13.L,  1.L/13.L,  4.L/13.L,  8.L/13.L,
    0.L/13.L,  1.L/13.L,  3.L/13.L,  9.L/13.L,
    0.L/13.L,  1.L/13.L,  2.L/13.L, 10.L/13.L,
    0.L/13.L,  1.L/13.L,  1.L/13.L, 11.L/13.L,
   11.L/13.L,  0.L/13.L,  1.L/13.L,  1.L/13.L,
   10.L/13.L,  0.L/13.L,  2.L/13.L,  1.L/13.L,
   10.L/13.L,  0.L/13.L,  1.L/13.L,  2.L/13.L,
    9.L/13.L,  0.L/13.L,  3.L/13.L,  1.L/13.L,
    9.L/13.L,  0.L/13.L,  2.L/13.L,  2.L/13.L,
    9.L/13.L,  0.L/13.L,  1.L/13.L,  3.L/13.L,
    8.L/13.L,  0.L/13.L,  4.L/13.L,  1.L/13.L,
    8.L/13.L,  0.L/13.L,  3.L/13.L,  2.L/13.L,
    8.L/13.L,  0.L/13.L,  2.L/13.L,  3.L/13.L,
    8.L/13.L,  0.L/13.L,  1.L/13.L,  4.L/13.L,
    7.L/13.L,  0.L/13.L,  5.L/13.L,  1.L/13.L,
    7.L/13.L,  0.L/13.L,  4.L/13.L,  2.L/13.L,
    7.L/13.L,  0.L/13.L,  3.L/13.L,  3.L/13.L,
    7.L/13.L,  0.L/13.L,  2.L/13.L,  4.L/13.L,
    7.L/13.L,  0.L/13.L,  1.L/13.L,  5.L/13.L,
    6.L/13.L,  0.L/13.L,  6.L/13.L,  1.L/13.L,
    6.L/13.L,  0.L/13.L,  5.L/13.L,  2.L/13.L,
    6.L/13.L,  0.L/13.L,  4.L/13.L,  3.L/13.L,
    6.L/13.L,  0.L/13.L,  3.L/13.L,  4.L/13.L,
    6.L/13.L,  0.L/13.L,  2.L/13.L,  5.L/13.L,
    6.L/13.L,  0.L/13.L,  1.L/13.L,  6.L/13.L,
    5.L/13.L,  0.L/13.L,  7.L/13.L,  1.L/13.L,
    5.L/13.L,  0.L/13.L,  6.L/13.L,  2.L/13.L,
    5.L/13.L,  0.L/13.L,  5.L/13.L,  3.L/13.L,
    5.L/13.L,  0.L/13.L,  4.L/13.L,  4.L/13.L,
    5.L/13.L,  0.L/13.L,  3.L/13.L,  5.L/13.L,
    5.L/13.L,  0.L/13.L,  2.L/13.L,  6.L/13.L,
    5.L/13.L,  0.L/13.L,  1.L/13.L,  7.L/13.L,
    4.L/13.L,  0.L/13.L,  8.L/13.L,  1.L/13.L,
    4.L/13.L,  0.L/13.L,  7.L/13.L,  2.L/13.L,
    4.L/13.L,  0.L/13.L,  6.L/13.L,  3.L/13.L,
    4.L/13.L,  0.L/13.L,  5.L/13.L,  4.L/13.L,
    4.L/13.L,  0.L/13.L,  4.L/13.L,  5.L/13.L,
    4.L/13.L,  0.L/13.L,  3.L/13.L,  6.L/13.L,
    4.L/13.L,  0.L/13.L,  2.L/13.L,  7.L/13.L,
    4.L/13.L,  0.L/13.L,  1.L/13.L,  8.L/13.L,
    3.L/13.L,  0.L/13.L,  9.L/13.L,  1.L/13.L,
    3.L/13.L,  0.L/13.L,  8.L/13.L,  2.L/13.L,
    3.L/13.L,  0.L/13.L,  7.L/13.L,  3.L/13.L,
    3.L/13.L,  0.L/13.L,  6.L/13.L,  4.L/13.L,
    3.L/13.L,  0.L/13.L,  5.L/13.L,  5.L/13.L,
    3.L/13.L,  0.L/13.L,  4.L/13.L,  6.L/13.L,
    3.L/13.L,  0.L/13.L,  3.L/13.L,  7.L/13.L,
    3.L/13.L,  0.L/13.L,  2.L/13.L,  8.L/13.L,
    3.L/13.L,  0.L/13.L,  1.L/13.L,  9.L/13.L,
    2.L/13.L,  0.L/13.L, 10.L/13.L,  1.L/13.L,
    2.L/13.L,  0.L/13.L,  9.L/13.L,  2.L/13.L,
    2.L/13.L,  0.L/13.L,  8.L/13.L,  3.L/13.L,
    2.L/13.L,  0.L/13.L,  7.L/13.L,  4.L/13.L,
    2.L/13.L,  0.L/13.L,  6.L/13.L,  5.L/13.L,
    2.L/13.L,  0.L/13.L,  5.L/13.L,  6.L/13.L,
    2.L/13.L,  0.L/13.L,  4.L/13.L,  7.L/13.L,
    2.L/13.L,  0.L/13.L,  3.L/13.L,  8.L/13.L,
    2.L/13.L,  0.L/13.L,  2.L/13.L,  9.L/13.L,
    2.L/13.L,  0.L/13.L,  1.L/13.L, 10.L/13.L,
    1.L/13.L,  0.L/13.L, 11.L/13.L,  1.L/13.L,
    1.L/13.L,  0.L/13.L, 10.L/13.L,  2.L/13.L,
    1.L/13.L,  0.L/13.L,  9.L/13.L,  3.L/13.L,
    1.L/13.L,  0.L/13.L,  8.L/13.L,  4.L/13.L,
    1.L/13.L,  0.L/13.L,  7.L/13.L,  5.L/13.L,
    1.L/13.L,  0.L/13.L,  6.L/13.L,  6.L/13.L,
    1.L/13.L,  0.L/13.L,  5.L/13.L,  7.L/13.L,
    1.L/13.L,  0.L/13.L,  4.L/13.L,  8.L/13.L,
    1.L/13.L,  0.L/13.L,  3.L/13.L,  9.L/13.L,
    1.L/13.L,  0.L/13.L,  2.L/13.L, 10.L/13.L,
    1.L/13.L,  0.L/13.L,  1.L/13.L, 11.L/13.L,
   11.L/13.L,  1.L/13.L,  0.L/13.L,  1.L/13.L,
   10.L/13.L,  2.L/13.L,  0.L/13.L,  1.L/13.L,
   10.L/13.L,  1.L/13.L,  0.L/13.L,  2.L/13.L,
    9.L/13.L,  3.L/13.L,  0.L/13.L,  1.L/13.L,
    9.L/13.L,  2.L/13.L,  0.L/13.L,  2.L/13.L,
    9.L/13.L,  1.L/13.L,  0.L/13.L,  3.L/13.L,
    8.L/13.L,  4.L/13.L,  0.L/13.L,  1.L/13.L,
    8.L/13.L,  3.L/13.L,  0.L/13.L,  2.L/13.L,
    8.L/13.L,  2.L/13.L,  0.L/13.L,  3.L/13.L,
    8.L/13.L,  1.L/13.L,  0.L/13.L,  4.L/13.L,
    7.L/13.L,  5.L/13.L,  0.L/13.L,  1.L/13.L,
    7.L/13.L,  4.L/13.L,  0.L/13.L,  2.L/13.L,
    7.L/13.L,  3.L/13.L,  0.L/13.L,  3.L/13.L,
    7.L/13.L,  2.L/13.L,  0.L/13.L,  4.L/13.L,
    7.L/13.L,  1.L/13.L,  0.L/13.L,  5.L/13.L,
    6.L/13.L,  6.L/13.L,  0.L/13.L,  1.L/13.L,
    6.L/13.L,  5.L/13.L,  0.L/13.L,  2.L/13.L,
    6.L/13.L,  4.L/13.L,  0.L/13.L,  3.L/13.L,
    6.L/13.L,  3.L/13.L,  0.L/13.L,  4.L/13.L,
    6.L/13.L,  2.L/13.L,  0.L/13.L,  5.L/13.L,
    6.L/13.L,  1.L/13.L,  0.L/13.L,  6.L/13.L,
    5.L/13.L,  7.L/13.L,  0.L/13.L,  1.L/13.L,
    5.L/13.L,  6.L/13.L,  0.L/13.L,  2.L/13.L,
    5.L/13.L,  5.L/13.L,  0.L/13.L,  3.L/13.L,
    5.L/13.L,  4.L/13.L,  0.L/13.L,  4.L/13.L,
    5.L/13.L,  3.L/13.L,  0.L/13.L,  5.L/13.L,
    5.L/13.L,  2.L/13.L,  0.L/13.L,  6.L/13.L,
    5.L/13.L,  1.L/13.L,  0.L/13.L,  7.L/13.L,
    4.L/13.L,  8.L/13.L,  0.L/13.L,  1.L/13.L,
    4.L/13.L,  7.L/13.L,  0.L/13.L,  2.L/13.L,
    4.L/13.L,  6.L/13.L,  0.L/13.L,  3.L/13.L,
    4.L/13.L,  5.L/13.L,  0.L/13.L,  4.L/13.L,
    4.L/13.L,  4.L/13.L,  0.L/13.L,  5.L/13.L,
    4.L/13.L,  3.L/13.L,  0.L/13.L,  6.L/13.L,
    4.L/13.L,  2.L/13.L,  0.L/13.L,  7.L/13.L,
    4.L/13.L,  1.L/13.L,  0.L/13.L,  8.L/13.L,
    3.L/13.L,  9.L/13.L,  0.L/13.L,  1.L/13.L,
    3.L/13.L,  8.L/13.L,  0.L/13.L,  2.L/13.L,
    3.L/13.L,  7.L/13.L,  0.L/13.L,  3.L/13.L,
    3.L/13.L,  6.L/13.L,  0.L/13.L,  4.L/13.L,
    3.L/13.L,  5.L/13.L,  0.L/13.L,  5.L/13.L,
    3.L/13.L,  4.L/13.L,  0.L/13.L,  6.L/13.L,
    3.L/13.L,  3.L/13.L,  0.L/13.L,  7.L/13.L,
    3.L/13.L,  2.L/13.L,  0.L/13.L,  8.L/13.L,
    3.L/13.L,  1.L/13.L,  0.L/13.L,  9.L/13.L,
    2.L/13.L, 10.L/13.L,  0.L/13.L,  1.L/13.L,
    2.L/13.L,  9.L/13.L,  0.L/13.L,  2.L/13.L,
    2.L/13.L,  8.L/13.L,  0.L/13.L,  3.L/13.L,
    2.L/13.L,  7.L/13.L,  0.L/13.L,  4.L/13.L,
    2.L/13.L,  6.L/13.L,  0.L/13.L,  5.L/13.L,
    2.L/13.L,  5.L/13.L,  0.L/13.L,  6.L/13.L,
    2.L/13.L,  4.L/13.L,  0.L/13.L,  7.L/13.L,
    2.L/13.L,  3.L/13.L,  0.L/13.L,  8.L/13.L,
    2.L/13.L,  2.L/13.L,  0.L/13.L,  9.L/13.L,
    2.L/13.L,  1.L/13.L,  0.L/13.L, 10.L/13.L,
    1.L/13.L, 11.L/13.L,  0.L/13.L,  1.L/13.L,
    1.L/13.L, 10.L/13.L,  0.L/13.L,  2.L/13.L,
    1.L/13.L,  9.L/13.L,  0.L/13.L,  3.L/13.L,
    1.L/13.L,  8.L/13.L,  0.L/13.L,  4.L/13.L,
    1.L/13.L,  7.L/13.L,  0.L/13.L,  5.L/13.L,
    1.L/13.L,  6.L/13.L,  0.L/13.L,  6.L/13.L,
    1.L/13.L,  5.L/13.L,  0.L/13.L,  7.L/13.L,
    1.L/13.L,  4.L/13.L,  0.L/13.L,  8.L/13.L,
    1.L/13.L,  3.L/13.L,  0.L/13.L,  9.L/13.L,
    1.L/13.L,  2.L/13.L,  0.L/13.L, 10.L/13.L,
    1.L/13.L,  1.L/13.L,  0.L/13.L, 11.L/13.L,
   11.L/13.L,  1.L/13.L,  1.L/13.L,  0.L/13.L,
   10.L/13.L,  2.L/13.L,  1.L/13.L,  0.L/13.L,
   10.L/13.L,  1.L/13.L,  2.L/13.L,  0.L/13.L,
    9.L/13.L,  3.L/13.L,  1.L/13.L,  0.L/13.L,
    9.L/13.L,  2.L/13.L,  2.L/13.L,  0.L/13.L,
    9.L/13.L,  1.L/13.L,  3.L/13.L,  0.L/13.L,
    8.L/13.L,  4.L/13.L,  1.L/13.L,  0.L/13.L,
    8.L/13.L,  3.L/13.L,  2.L/13.L,  0.L/13.L,
    8.L/13.L,  2.L/13.L,  3.L/13.L,  0.L/13.L,
    8.L/13.L,  1.L/13.L,  4.L/13.L,  0.L/13.L,
    7.L/13.L,  5.L/13.L,  1.L/13.L,  0.L/13.L,
    7.L/13.L,  4.L/13.L,  2.L/13.L,  0.L/13.L,
    7.L/13.L,  3.L/13.L,  3.L/13.L,  0.L/13.L,
    7.L/13.L,  2.L/13.L,  4.L/13.L,  0.L/13.L,
    7.L/13.L,  1.L/13.L,  5.L/13.L,  0.L/13.L,
    6.L/13.L,  6.L/13.L,  1.L/13.L,  0.L/13.L,
    6.L/13.L,  5.L/13.L,  2.L/13.L,  0.L/13.L,
    6.L/13.L,  4.L/13.L,  3.L/13.L,  0.L/13.L,
    6.L/13.L,  3.L/13.L,  4.L/13.L,  0.L/13.L,
    6.L/13.L,  2.L/13.L,  5.L/13.L,  0.L/13.L,
    6.L/13.L,  1.L/13.L,  6.L/13.L,  0.L/13.L,
    5.L/13.L,  7.L/13.L,  1.L/13.L,  0.L/13.L,
    5.L/13.L,  6.L/13.L,  2.L/13.L,  0.L/13.L,
    5.L/13.L,  5.L/13.L,  3.L/13.L,  0.L/13.L,
    5.L/13.L,  4.L/13.L,  4.L/13.L,  0.L/13.L,
    5.L/13.L,  3.L/13.L,  5.L/13.L,  0.L/13.L,
    5.L/13.L,  2.L/13.L,  6.L/13.L,  0.L/13.L,
    5.L/13.L,  1.L/13.L,  7.L/13.L,  0.L/13.L,
    4.L/13.L,  8.L/13.L,  1.L/13.L,  0.L/13.L,
    4.L/13.L,  7.L/13.L,  2.L/13.L,  0.L/13.L,
    4.L/13.L,  6.L/13.L,  3.L/13.L,  0.L/13.L,
    4.L/13.L,  5.L/13.L,  4.L/13.L,  0.L/13.L,
    4.L/13.L,  4.L/13.L,  5.L/13.L,  0.L/13.L,
    4.L/13.L,  3.L/13.L,  6.L/13.L,  0.L/13.L,
    4.L/13.L,  2.L/13.L,  7.L/13.L,  0.L/13.L,
    4.L/13.L,  1.L/13.L,  8.L/13.L,  0.L/13.L,
    3.L/13.L,  9.L/13.L,  1.L/13.L,  0.L/13.L,
    3.L/13.L,  8.L/13.L,  2.L/13.L,  0.L/13.L,
    3.L/13.L,  7.L/13.L,  3.L/13.L,  0.L/13.L,
    3.L/13.L,  6.L/13.L,  4.L/13.L,  0.L/13.L,
    3.L/13.L,  5.L/13.L,  5.L/13.L,  0.L/13.L,
    3.L/13.L,  4.L/13.L,  6.L/13.L,  0.L/13.L,
    3.L/13.L,  3.L/13.L,  7.L/13.L,  0.L/13.L,
    3.L/13.L,  2.L/13.L,  8.L/13.L,  0.L/13.L,
    3.L/13.L,  1.L/13.L,  9.L/13.L,  0.L/13.L,
    2.L/13.L, 10.L/13.L,  1.L/13.L,  0.L/13.L,
    2.L/13.L,  9.L/13.L,  2.L/13.L,  0.L/13.L,
    2.L/13.L,  8.L/13.L,  3.L/13.L,  0.L/13.L,
    2.L/13.L,  7.L/13.L,  4.L/13.L,  0.L/13.L,
    2.L/13.L,  6.L/13.L,  5.L/13.L,  0.L/13.L,
    2.L/13.L,  5.L/13.L,  6.L/13.L,  0.L/13.L,
    2.L/13.L,  4.L/13.L,  7.L/13.L,  0.L/13.L,
    2.L/13.L,  3.L/13.L,  8.L/13.L,  0.L/13.L,
    2.L/13.L,  2.L/13.L,  9.L/13.L,  0.L/13.L,
    2.L/13.L,  1.L/13.L, 10.L/13.L,  0.L/13.L,
    1.L/13.L, 11.L/13.L,  1.L/13.L,  0.L/13.L,
    1.L/13.L, 10.L/13.L,  2.L/13.L,  0.L/13.L,
    1.L/13.L,  9.L/13.L,  3.L/13.L,  0.L/13.L,
    1.L/13.L,  8.L/13.L,  4.L/13.L,  0.L/13.L,
    1.L/13.L,  7.L/13.L,  5.L/13.L,  0.L/13.L,
    1.L/13.L,  6.L/13.L,  6.L/13.L,  0.L/13.L,
    1.L/13.L,  5.L/13.L,  7.L/13.L,  0.L/13.L,
    1.L/13.L,  4.L/13.L,  8.L/13.L,  0.L/13.L,
    1.L/13.L,  3.L/13.L,  9.L/13.L,  0.L/13.L,
    1.L/13.L,  2.L/13.L, 10.L/13.L,  0.L/13.L,
    1.L/13.L,  1.L/13.L, 11.L/13.L,  0.L/13.L,
   10.L/13.L,  1.L/13.L,  1.L/13.L,  1.L/13.L,
    9.L/13.L,  2.L/13.L,  1.L/13.L,  1.L/13.L,
    9.L/13.L,  1.L/13.L,  2.L/13.L,  1.L/13.L,
    9.L/13.L,  1.L/13.L,  1.L/13.L,  2.L/13.L,
    8.L/13.L,  3.L/13.L,  1.L/13.L,  1.L/13.L,
    8.L/13.L,  2.L/13.L,  2.L/13.L,  1.L/13.L,
    8.L/13.L,  2.L/13.L,  1.L/13.L,  2.L/13.L,
    8.L/13.L,  1.L/13.L,  3.L/13.L,  1.L/13.L,
    8.L/13.L,  1.L/13.L,  2.L/13.L,  2.L/13.L,
    8.L/13.L,  1.L/13.L,  1.L/13.L,  3.L/13.L,
    7.L/13.L,  4.L/13.L,  1.L/13.L,  1.L/13.L,
    7.L/13.L,  3.L/13.L,  2.L/13.L,  1.L/13.L,
    7.L/13.L,  3.L/13.L,  1.L/13.L,  2.L/13.L,
    7.L/13.L,  2.L/13.L,  3.L/13.L,  1.L/13.L,
    7.L/13.L,  2.L/13.L,  2.L/13.L,  2.L/13.L,
    7.L/13.L,  2.L/13.L,  1.L/13.L,  3.L/13.L,
    7.L/13.L,  1.L/13.L,  4.L/13.L,  1.L/13.L,
    7.L/13.L,  1.L/13.L,  3.L/13.L,  2.L/13.L,
    7.L/13.L,  1.L/13.L,  2.L/13.L,  3.L/13.L,
    7.L/13.L,  1.L/13.L,  1.L/13.L,  4.L/13.L,
    6.L/13.L,  5.L/13.L,  1.L/13.L,  1.L/13.L,
    6.L/13.L,  4.L/13.L,  2.L/13.L,  1.L/13.L,
    6.L/13.L,  4.L/13.L,  1.L/13.L,  2.L/13.L,
    6.L/13.L,  3.L/13.L,  3.L/13.L,  1.L/13.L,
    6.L/13.L,  3.L/13.L,  2.L/13.L,  2.L/13.L,
    6.L/13.L,  3.L/13.L,  1.L/13.L,  3.L/13.L,
    6.L/13.L,  2.L/13.L,  4.L/13.L,  1.L/13.L,
    6.L/13.L,  2.L/13.L,  3.L/13.L,  2.L/13.L,
    6.L/13.L,  2.L/13.L,  2.L/13.L,  3.L/13.L,
    6.L/13.L,  2.L/13.L,  1.L/13.L,  4.L/13.L,
    6.L/13.L,  1.L/13.L,  5.L/13.L,  1.L/13.L,
    6.L/13.L,  1.L/13.L,  4.L/13.L,  2.L/13.L,
    6.L/13.L,  1.L/13.L,  3.L/13.L,  3.L/13.L,
    6.L/13.L,  1.L/13.L,  2.L/13.L,  4.L/13.L,
    6.L/13.L,  1.L/13.L,  1.L/13.L,  5.L/13.L,
    5.L/13.L,  6.L/13.L,  1.L/13.L,  1.L/13.L,
    5.L/13.L,  5.L/13.L,  2.L/13.L,  1.L/13.L,
    5.L/13.L,  5.L/13.L,  1.L/13.L,  2.L/13.L,
    5.L/13.L,  4.L/13.L,  3.L/13.L,  1.L/13.L,
    5.L/13.L,  4.L/13.L,  2.L/13.L,  2.L/13.L,
    5.L/13.L,  4.L/13.L,  1.L/13.L,  3.L/13.L,
    5.L/13.L,  3.L/13.L,  4.L/13.L,  1.L/13.L,
    5.L/13.L,  3.L/13.L,  3.L/13.L,  2.L/13.L,
    5.L/13.L,  3.L/13.L,  2.L/13.L,  3.L/13.L,
    5.L/13.L,  3.L/13.L,  1.L/13.L,  4.L/13.L,
    5.L/13.L,  2.L/13.L,  5.L/13.L,  1.L/13.L,
    5.L/13.L,  2.L/13.L,  4.L/13.L,  2.L/13.L,
    5.L/13.L,  2.L/13.L,  3.L/13.L,  3.L/13.L,
    5.L/13.L,  2.L/13.L,  2.L/13.L,  4.L/13.L,
    5.L/13.L,  2.L/13.L,  1.L/13.L,  5.L/13.L,
    5.L/13.L,  1.L/13.L,  6.L/13.L,  1.L/13.L,
    5.L/13.L,  1.L/13.L,  5.L/13.L,  2.L/13.L,
    5.L/13.L,  1.L/13.L,  4.L/13.L,  3.L/13.L,
    5.L/13.L,  1.L/13.L,  3.L/13.L,  4.L/13.L,
    5.L/13.L,  1.L/13.L,  2.L/13.L,  5.L/13.L,
    5.L/13.L,  1.L/13.L,  1.L/13.L,  6.L/13.L,
    4.L/13.L,  7.L/13.L,  1.L/13.L,  1.L/13.L,
    4.L/13.L,  6.L/13.L,  2.L/13.L,  1.L/13.L,
    4.L/13.L,  6.L/13.L,  1.L/13.L,  2.L/13.L,
    4.L/13.L,  5.L/13.L,  3.L/13.L,  1.L/13.L,
    4.L/13.L,  5.L/13.L,  2.L/13.L,  2.L/13.L,
    4.L/13.L,  5.L/13.L,  1.L/13.L,  3.L/13.L,
    4.L/13.L,  4.L/13.L,  4.L/13.L,  1.L/13.L,
    4.L/13.L,  4.L/13.L,  3.L/13.L,  2.L/13.L,
    4.L/13.L,  4.L/13.L,  2.L/13.L,  3.L/13.L,
    4.L/13.L,  4.L/13.L,  1.L/13.L,  4.L/13.L,
    4.L/13.L,  3.L/13.L,  5.L/13.L,  1.L/13.L,
    4.L/13.L,  3.L/13.L,  4.L/13.L,  2.L/13.L,
    4.L/13.L,  3.L/13.L,  3.L/13.L,  3.L/13.L,
    4.L/13.L,  3.L/13.L,  2.L/13.L,  4.L/13.L,
    4.L/13.L,  3.L/13.L,  1.L/13.L,  5.L/13.L,
    4.L/13.L,  2.L/13.L,  6.L/13.L,  1.L/13.L,
    4.L/13.L,  2.L/13.L,  5.L/13.L,  2.L/13.L,
    4.L/13.L,  2.L/13.L,  4.L/13.L,  3.L/13.L,
    4.L/13.L,  2.L/13.L,  3.L/13.L,  4.L/13.L,
    4.L/13.L,  2.L/13.L,  2.L/13.L,  5.L/13.L,
    4.L/13.L,  2.L/13.L,  1.L/13.L,  6.L/13.L,
    4.L/13.L,  1.L/13.L,  7.L/13.L,  1.L/13.L,
    4.L/13.L,  1.L/13.L,  6.L/13.L,  2.L/13.L,
    4.L/13.L,  1.L/13.L,  5.L/13.L,  3.L/13.L,
    4.L/13.L,  1.L/13.L,  4.L/13.L,  4.L/13.L,
    4.L/13.L,  1.L/13.L,  3.L/13.L,  5.L/13.L,
    4.L/13.L,  1.L/13.L,  2.L/13.L,  6.L/13.L,
    4.L/13.L,  1.L/13.L,  1.L/13.L,  7.L/13.L,
    3.L/13.L,  8.L/13.L,  1.L/13.L,  1.L/13.L,
    3.L/13.L,  7.L/13.L,  2.L/13.L,  1.L/13.L,
    3.L/13.L,  7.L/13.L,  1.L/13.L,  2.L/13.L,
    3.L/13.L,  6.L/13.L,  3.L/13.L,  1.L/13.L,
    3.L/13.L,  6.L/13.L,  2.L/13.L,  2.L/13.L,
    3.L/13.L,  6.L/13.L,  1.L/13.L,  3.L/13.L,
    3.L/13.L,  5.L/13.L,  4.L/13.L,  1.L/13.L,
    3.L/13.L,  5.L/13.L,  3.L/13.L,  2.L/13.L,
    3.L/13.L,  5.L/13.L,  2.L/13.L,  3.L/13.L,
    3.L/13.L,  5.L/13.L,  1.L/13.L,  4.L/13.L,
    3.L/13.L,  4.L/13.L,  5.L/13.L,  1.L/13.L,
    3.L/13.L,  4.L/13.L,  4.L/13.L,  2.L/13.L,
    3.L/13.L,  4.L/13.L,  3.L/13.L,  3.L/13.L,
    3.L/13.L,  4.L/13.L,  2.L/13.L,  4.L/13.L,
    3.L/13.L,  4.L/13.L,  1.L/13.L,  5.L/13.L,
    3.L/13.L,  3.L/13.L,  6.L/13.L,  1.L/13.L,
    3.L/13.L,  3.L/13.L,  5.L/13.L,  2.L/13.L,
    3.L/13.L,  3.L/13.L,  4.L/13.L,  3.L/13.L,
    3.L/13.L,  3.L/13.L,  3.L/13.L,  4.L/13.L,
    3.L/13.L,  3.L/13.L,  2.L/13.L,  5.L/13.L,
    3.L/13.L,  3.L/13.L,  1.L/13.L,  6.L/13.L,
    3.L/13.L,  2.L/13.L,  7.L/13.L,  1.L/13.L,
    3.L/13.L,  2.L/13.L,  6.L/13.L,  2.L/13.L,
    3.L/13.L,  2.L/13.L,  5.L/13.L,  3.L/13.L,
    3.L/13.L,  2.L/13.L,  4.L/13.L,  4.L/13.L,
    3.L/13.L,  2.L/13.L,  3.L/13.L,  5.L/13.L,
    3.L/13.L,  2.L/13.L,  2.L/13.L,  6.L/13.L,
    3.L/13.L,  2.L/13.L,  1.L/13.L,  7.L/13.L,
    3.L/13.L,  1.L/13.L,  8.L/13.L,  1.L/13.L,
    3.L/13.L,  1.L/13.L,  7.L/13.L,  2.L/13.L,
    3.L/13.L,  1.L/13.L,  6.L/13.L,  3.L/13.L,
    3.L/13.L,  1.L/13.L,  5.L/13.L,  4.L/13.L,
    3.L/13.L,  1.L/13.L,  4.L/13.L,  5.L/13.L,
    3.L/13.L,  1.L/13.L,  3.L/13.L,  6.L/13.L,
    3.L/13.L,  1.L/13.L,  2.L/13.L,  7.L/13.L,
    3.L/13.L,  1.L/13.L,  1.L/13.L,  8.L/13.L,
    2.L/13.L,  9.L/13.L,  1.L/13.L,  1.L/13.L,
    2.L/13.L,  8.L/13.L,  2.L/13.L,  1.L/13.L,
    2.L/13.L,  8.L/13.L,  1.L/13.L,  2.L/13.L,
    2.L/13.L,  7.L/13.L,  3.L/13.L,  1.L/13.L,
    2.L/13.L,  7.L/13.L,  2.L/13.L,  2.L/13.L,
    2.L/13.L,  7.L/13.L,  1.L/13.L,  3.L/13.L,
    2.L/13.L,  6.L/13.L,  4.L/13.L,  1.L/13.L,
    2.L/13.L,  6.L/13.L,  3.L/13.L,  2.L/13.L,
    2.L/13.L,  6.L/13.L,  2.L/13.L,  3.L/13.L,
    2.L/13.L,  6.L/13.L,  1.L/13.L,  4.L/13.L,
    2.L/13.L,  5.L/13.L,  5.L/13.L,  1.L/13.L,
    2.L/13.L,  5.L/13.L,  4.L/13.L,  2.L/13.L,
    2.L/13.L,  5.L/13.L,  3.L/13.L,  3.L/13.L,
    2.L/13.L,  5.L/13.L,  2.L/13.L,  4.L/13.L,
    2.L/13.L,  5.L/13.L,  1.L/13.L,  5.L/13.L,
    2.L/13.L,  4.L/13.L,  6.L/13.L,  1.L/13.L,
    2.L/13.L,  4.L/13.L,  5.L/13.L,  2.L/13.L,
    2.L/13.L,  4.L/13.L,  4.L/13.L,  3.L/13.L,
    2.L/13.L,  4.L/13.L,  3.L/13.L,  4.L/13.L,
    2.L/13.L,  4.L/13.L,  2.L/13.L,  5.L/13.L,
    2.L/13.L,  4.L/13.L,  1.L/13.L,  6.L/13.L,
    2.L/13.L,  3.L/13.L,  7.L/13.L,  1.L/13.L,
    2.L/13.L,  3.L/13.L,  6.L/13.L,  2.L/13.L,
    2.L/13.L,  3.L/13.L,  5.L/13.L,  3.L/13.L,
    2.L/13.L,  3.L/13.L,  4.L/13.L,  4.L/13.L,
    2.L/13.L,  3.L/13.L,  3.L/13.L,  5.L/13.L,
    2.L/13.L,  3.L/13.L,  2.L/13.L,  6.L/13.L,
    2.L/13.L,  3.L/13.L,  1.L/13.L,  7.L/13.L,
    2.L/13.L,  2.L/13.L,  8.L/13.L,  1.L/13.L,
    2.L/13.L,  2.L/13.L,  7.L/13.L,  2.L/13.L,
    2.L/13.L,  2.L/13.L,  6.L/13.L,  3.L/13.L,
    2.L/13.L,  2.L/13.L,  5.L/13.L,  4.L/13.L,
    2.L/13.L,  2.L/13.L,  4.L/13.L,  5.L/13.L,
    2.L/13.L,  2.L/13.L,  3.L/13.L,  6.L/13.L,
    2.L/13.L,  2.L/13.L,  2.L/13.L,  7.L/13.L,
    2.L/13.L,  2.L/13.L,  1.L/13.L,  8.L/13.L,
    2.L/13.L,  1.L/13.L,  9.L/13.L,  1.L/13.L,
    2.L/13.L,  1.L/13.L,  8.L/13.L,  2.L/13.L,
    2.L/13.L,  1.L/13.L,  7.L/13.L,  3.L/13.L,
    2.L/13.L,  1.L/13.L,  6.L/13.L,  4.L/13.L,
    2.L/13.L,  1.L/13.L,  5.L/13.L,  5.L/13.L,
    2.L/13.L,  1.L/13.L,  4.L/13.L,  6.L/13.L,
    2.L/13.L,  1.L/13.L,  3.L/13.L,  7.L/13.L,
    2.L/13.L,  1.L/13.L,  2.L/13.L,  8.L/13.L,
    2.L/13.L,  1.L/13.L,  1.L/13.L,  9.L/13.L,
    1.L/13.L, 10.L/13.L,  1.L/13.L,  1.L/13.L,
    1.L/13.L,  9.L/13.L,  2.L/13.L,  1.L/13.L,
    1.L/13.L,  9.L/13.L,  1.L/13.L,  2.L/13.L,
    1.L/13.L,  8.L/13.L,  3.L/13.L,  1.L/13.L,
    1.L/13.L,  8.L/13.L,  2.L/13.L,  2.L/13.L,
    1.L/13.L,  8.L/13.L,  1.L/13.L,  3.L/13.L,
    1.L/13.L,  7.L/13.L,  4.L/13.L,  1.L/13.L,
    1.L/13.L,  7.L/13.L,  3.L/13.L,  2.L/13.L,
    1.L/13.L,  7.L/13.L,  2.L/13.L,  3.L/13.L,
    1.L/13.L,  7.L/13.L,  1.L/13.L,  4.L/13.L,
    1.L/13.L,  6.L/13.L,  5.L/13.L,  1.L/13.L,
    1.L/13.L,  6.L/13.L,  4.L/13.L,  2.L/13.L,
    1.L/13.L,  6.L/13.L,  3.L/13.L,  3.L/13.L,
    1.L/13.L,  6.L/13.L,  2.L/13.L,  4.L/13.L,
    1.L/13.L,  6.L/13.L,  1.L/13.L,  5.L/13.L,
    1.L/13.L,  5.L/13.L,  6.L/13.L,  1.L/13.L,
    1.L/13.L,  5.L/13.L,  5.L/13.L,  2.L/13.L,
    1.L/13.L,  5.L/13.L,  4.L/13.L,  3.L/13.L,
    1.L/13.L,  5.L/13.L,  3.L/13.L,  4.L/13.L,
    1.L/13.L,  5.L/13.L,  2.L/13.L,  5.L/13.L,
    1.L/13.L,  5.L/13.L,  1.L/13.L,  6.L/13.L,
    1.L/13.L,  4.L/13.L,  7.L/13.L,  1.L/13.L,
    1.L/13.L,  4.L/13.L,  6.L/13.L,  2.L/13.L,
    1.L/13.L,  4.L/13.L,  5.L/13.L,  3.L/13.L,
    1.L/13.L,  4.L/13.L,  4.L/13.L,  4.L/13.L,
    1.L/13.L,  4.L/13.L,  3.L/13.L,  5.L/13.L,
    1.L/13.L,  4.L/13.L,  2.L/13.L,  6.L/13.L,
    1.L/13.L,  4.L/13.L,  1.L/13.L,  7.L/13.L,
    1.L/13.L,  3.L/13.L,  8.L/13.L,  1.L/13.L,
    1.L/13.L,  3.L/13.L,  7.L/13.L,  2.L/13.L,
    1.L/13.L,  3.L/13.L,  6.L/13.L,  3.L/13.L,
    1.L/13.L,  3.L/13.L,  5.L/13.L,  4.L/13.L,
    1.L/13.L,  3.L/13.L,  4.L/13.L,  5.L/13.L,
    1.L/13.L,  3.L/13.L,  3.L/13.L,  6.L/13.L,
    1.L/13.L,  3.L/13.L,  2.L/13.L,  7.L/13.L,
    1.L/13.L,  3.L/13.L,  1.L/13.L,  8.L/13.L,
    1.L/13.L,  2.L/13.L,  9.L/13.L,  1.L/13.L,
    1.L/13.L,  2.L/13.L,  8.L/13.L,  2.L/13.L,
    1.L/13.L,  2.L/13.L,  7.L/13.L,  3.L/13.L,
    1.L/13.L,  2.L/13.L,  6.L/13.L,  4.L/13.L,
    1.L/13.L,  2.L/13.L,  5.L/13.L,  5.L/13.L,
    1.L/13.L,  2.L/13.L,  4.L/13.L,  6.L/13.L,
    1.L/13.L,  2.L/13.L,  3.L/13.L,  7.L/13.L,
    1.L/13.L,  2.L/13.L,  2.L/13.L,  8.L/13.L,
    1.L/13.L,  2.L/13.L,  1.L/13.L,  9.L/13.L,
    1.L/13.L,  1.L/13.L, 10.L/13.L,  1.L/13.L,
    1.L/13.L,  1.L/13.L,  9.L/13.L,  2.L/13.L,
    1.L/13.L,  1.L/13.L,  8.L/13.L,  3.L/13.L,
    1.L/13.L,  1.L/13.L,  7.L/13.L,  4.L/13.L,
    1.L/13.L,  1.L/13.L,  6.L/13.L,  5.L/13.L,
    1.L/13.L,  1.L/13.L,  5.L/13.L,  6.L/13.L,
    1.L/13.L,  1.L/13.L,  4.L/13.L,  7.L/13.L,
    1.L/13.L,  1.L/13.L,  3.L/13.L,  8.L/13.L,
    1.L/13.L,  1.L/13.L,  2.L/13.L,  9.L/13.L,
    1.L/13.L,  1.L/13.L,  1.L/13.L, 10.L/13.L
};
DOF_TYPE DOF_DG13_ = {
    DofCache, "DG13", DG13_points, NULL,
    DOF_DG12, DOF_P13, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 560, 13, -1, 1,
    0, 0, 0, 560
};

static FLOAT DG14_points[] = {
   14.L/14.L,  0.L/14.L,  0.L/14.L,  0.L/14.L,
    0.L/14.L, 14.L/14.L,  0.L/14.L,  0.L/14.L,
    0.L/14.L,  0.L/14.L, 14.L/14.L,  0.L/14.L,
    0.L/14.L,  0.L/14.L,  0.L/14.L, 14.L/14.L,
   13.L/14.L,  1.L/14.L,  0.L/14.L,  0.L/14.L,
   12.L/14.L,  2.L/14.L,  0.L/14.L,  0.L/14.L,
   11.L/14.L,  3.L/14.L,  0.L/14.L,  0.L/14.L,
   10.L/14.L,  4.L/14.L,  0.L/14.L,  0.L/14.L,
    9.L/14.L,  5.L/14.L,  0.L/14.L,  0.L/14.L,
    8.L/14.L,  6.L/14.L,  0.L/14.L,  0.L/14.L,
    7.L/14.L,  7.L/14.L,  0.L/14.L,  0.L/14.L,
    6.L/14.L,  8.L/14.L,  0.L/14.L,  0.L/14.L,
    5.L/14.L,  9.L/14.L,  0.L/14.L,  0.L/14.L,
    4.L/14.L, 10.L/14.L,  0.L/14.L,  0.L/14.L,
    3.L/14.L, 11.L/14.L,  0.L/14.L,  0.L/14.L,
    2.L/14.L, 12.L/14.L,  0.L/14.L,  0.L/14.L,
    1.L/14.L, 13.L/14.L,  0.L/14.L,  0.L/14.L,
   13.L/14.L,  0.L/14.L,  1.L/14.L,  0.L/14.L,
   12.L/14.L,  0.L/14.L,  2.L/14.L,  0.L/14.L,
   11.L/14.L,  0.L/14.L,  3.L/14.L,  0.L/14.L,
   10.L/14.L,  0.L/14.L,  4.L/14.L,  0.L/14.L,
    9.L/14.L,  0.L/14.L,  5.L/14.L,  0.L/14.L,
    8.L/14.L,  0.L/14.L,  6.L/14.L,  0.L/14.L,
    7.L/14.L,  0.L/14.L,  7.L/14.L,  0.L/14.L,
    6.L/14.L,  0.L/14.L,  8.L/14.L,  0.L/14.L,
    5.L/14.L,  0.L/14.L,  9.L/14.L,  0.L/14.L,
    4.L/14.L,  0.L/14.L, 10.L/14.L,  0.L/14.L,
    3.L/14.L,  0.L/14.L, 11.L/14.L,  0.L/14.L,
    2.L/14.L,  0.L/14.L, 12.L/14.L,  0.L/14.L,
    1.L/14.L,  0.L/14.L, 13.L/14.L,  0.L/14.L,
   13.L/14.L,  0.L/14.L,  0.L/14.L,  1.L/14.L,
   12.L/14.L,  0.L/14.L,  0.L/14.L,  2.L/14.L,
   11.L/14.L,  0.L/14.L,  0.L/14.L,  3.L/14.L,
   10.L/14.L,  0.L/14.L,  0.L/14.L,  4.L/14.L,
    9.L/14.L,  0.L/14.L,  0.L/14.L,  5.L/14.L,
    8.L/14.L,  0.L/14.L,  0.L/14.L,  6.L/14.L,
    7.L/14.L,  0.L/14.L,  0.L/14.L,  7.L/14.L,
    6.L/14.L,  0.L/14.L,  0.L/14.L,  8.L/14.L,
    5.L/14.L,  0.L/14.L,  0.L/14.L,  9.L/14.L,
    4.L/14.L,  0.L/14.L,  0.L/14.L, 10.L/14.L,
    3.L/14.L,  0.L/14.L,  0.L/14.L, 11.L/14.L,
    2.L/14.L,  0.L/14.L,  0.L/14.L, 12.L/14.L,
    1.L/14.L,  0.L/14.L,  0.L/14.L, 13.L/14.L,
    0.L/14.L, 13.L/14.L,  1.L/14.L,  0.L/14.L,
    0.L/14.L, 12.L/14.L,  2.L/14.L,  0.L/14.L,
    0.L/14.L, 11.L/14.L,  3.L/14.L,  0.L/14.L,
    0.L/14.L, 10.L/14.L,  4.L/14.L,  0.L/14.L,
    0.L/14.L,  9.L/14.L,  5.L/14.L,  0.L/14.L,
    0.L/14.L,  8.L/14.L,  6.L/14.L,  0.L/14.L,
    0.L/14.L,  7.L/14.L,  7.L/14.L,  0.L/14.L,
    0.L/14.L,  6.L/14.L,  8.L/14.L,  0.L/14.L,
    0.L/14.L,  5.L/14.L,  9.L/14.L,  0.L/14.L,
    0.L/14.L,  4.L/14.L, 10.L/14.L,  0.L/14.L,
    0.L/14.L,  3.L/14.L, 11.L/14.L,  0.L/14.L,
    0.L/14.L,  2.L/14.L, 12.L/14.L,  0.L/14.L,
    0.L/14.L,  1.L/14.L, 13.L/14.L,  0.L/14.L,
    0.L/14.L, 13.L/14.L,  0.L/14.L,  1.L/14.L,
    0.L/14.L, 12.L/14.L,  0.L/14.L,  2.L/14.L,
    0.L/14.L, 11.L/14.L,  0.L/14.L,  3.L/14.L,
    0.L/14.L, 10.L/14.L,  0.L/14.L,  4.L/14.L,
    0.L/14.L,  9.L/14.L,  0.L/14.L,  5.L/14.L,
    0.L/14.L,  8.L/14.L,  0.L/14.L,  6.L/14.L,
    0.L/14.L,  7.L/14.L,  0.L/14.L,  7.L/14.L,
    0.L/14.L,  6.L/14.L,  0.L/14.L,  8.L/14.L,
    0.L/14.L,  5.L/14.L,  0.L/14.L,  9.L/14.L,
    0.L/14.L,  4.L/14.L,  0.L/14.L, 10.L/14.L,
    0.L/14.L,  3.L/14.L,  0.L/14.L, 11.L/14.L,
    0.L/14.L,  2.L/14.L,  0.L/14.L, 12.L/14.L,
    0.L/14.L,  1.L/14.L,  0.L/14.L, 13.L/14.L,
    0.L/14.L,  0.L/14.L, 13.L/14.L,  1.L/14.L,
    0.L/14.L,  0.L/14.L, 12.L/14.L,  2.L/14.L,
    0.L/14.L,  0.L/14.L, 11.L/14.L,  3.L/14.L,
    0.L/14.L,  0.L/14.L, 10.L/14.L,  4.L/14.L,
    0.L/14.L,  0.L/14.L,  9.L/14.L,  5.L/14.L,
    0.L/14.L,  0.L/14.L,  8.L/14.L,  6.L/14.L,
    0.L/14.L,  0.L/14.L,  7.L/14.L,  7.L/14.L,
    0.L/14.L,  0.L/14.L,  6.L/14.L,  8.L/14.L,
    0.L/14.L,  0.L/14.L,  5.L/14.L,  9.L/14.L,
    0.L/14.L,  0.L/14.L,  4.L/14.L, 10.L/14.L,
    0.L/14.L,  0.L/14.L,  3.L/14.L, 11.L/14.L,
    0.L/14.L,  0.L/14.L,  2.L/14.L, 12.L/14.L,
    0.L/14.L,  0.L/14.L,  1.L/14.L, 13.L/14.L,
    0.L/14.L, 12.L/14.L,  1.L/14.L,  1.L/14.L,
    0.L/14.L, 11.L/14.L,  2.L/14.L,  1.L/14.L,
    0.L/14.L, 11.L/14.L,  1.L/14.L,  2.L/14.L,
    0.L/14.L, 10.L/14.L,  3.L/14.L,  1.L/14.L,
    0.L/14.L, 10.L/14.L,  2.L/14.L,  2.L/14.L,
    0.L/14.L, 10.L/14.L,  1.L/14.L,  3.L/14.L,
    0.L/14.L,  9.L/14.L,  4.L/14.L,  1.L/14.L,
    0.L/14.L,  9.L/14.L,  3.L/14.L,  2.L/14.L,
    0.L/14.L,  9.L/14.L,  2.L/14.L,  3.L/14.L,
    0.L/14.L,  9.L/14.L,  1.L/14.L,  4.L/14.L,
    0.L/14.L,  8.L/14.L,  5.L/14.L,  1.L/14.L,
    0.L/14.L,  8.L/14.L,  4.L/14.L,  2.L/14.L,
    0.L/14.L,  8.L/14.L,  3.L/14.L,  3.L/14.L,
    0.L/14.L,  8.L/14.L,  2.L/14.L,  4.L/14.L,
    0.L/14.L,  8.L/14.L,  1.L/14.L,  5.L/14.L,
    0.L/14.L,  7.L/14.L,  6.L/14.L,  1.L/14.L,
    0.L/14.L,  7.L/14.L,  5.L/14.L,  2.L/14.L,
    0.L/14.L,  7.L/14.L,  4.L/14.L,  3.L/14.L,
    0.L/14.L,  7.L/14.L,  3.L/14.L,  4.L/14.L,
    0.L/14.L,  7.L/14.L,  2.L/14.L,  5.L/14.L,
    0.L/14.L,  7.L/14.L,  1.L/14.L,  6.L/14.L,
    0.L/14.L,  6.L/14.L,  7.L/14.L,  1.L/14.L,
    0.L/14.L,  6.L/14.L,  6.L/14.L,  2.L/14.L,
    0.L/14.L,  6.L/14.L,  5.L/14.L,  3.L/14.L,
    0.L/14.L,  6.L/14.L,  4.L/14.L,  4.L/14.L,
    0.L/14.L,  6.L/14.L,  3.L/14.L,  5.L/14.L,
    0.L/14.L,  6.L/14.L,  2.L/14.L,  6.L/14.L,
    0.L/14.L,  6.L/14.L,  1.L/14.L,  7.L/14.L,
    0.L/14.L,  5.L/14.L,  8.L/14.L,  1.L/14.L,
    0.L/14.L,  5.L/14.L,  7.L/14.L,  2.L/14.L,
    0.L/14.L,  5.L/14.L,  6.L/14.L,  3.L/14.L,
    0.L/14.L,  5.L/14.L,  5.L/14.L,  4.L/14.L,
    0.L/14.L,  5.L/14.L,  4.L/14.L,  5.L/14.L,
    0.L/14.L,  5.L/14.L,  3.L/14.L,  6.L/14.L,
    0.L/14.L,  5.L/14.L,  2.L/14.L,  7.L/14.L,
    0.L/14.L,  5.L/14.L,  1.L/14.L,  8.L/14.L,
    0.L/14.L,  4.L/14.L,  9.L/14.L,  1.L/14.L,
    0.L/14.L,  4.L/14.L,  8.L/14.L,  2.L/14.L,
    0.L/14.L,  4.L/14.L,  7.L/14.L,  3.L/14.L,
    0.L/14.L,  4.L/14.L,  6.L/14.L,  4.L/14.L,
    0.L/14.L,  4.L/14.L,  5.L/14.L,  5.L/14.L,
    0.L/14.L,  4.L/14.L,  4.L/14.L,  6.L/14.L,
    0.L/14.L,  4.L/14.L,  3.L/14.L,  7.L/14.L,
    0.L/14.L,  4.L/14.L,  2.L/14.L,  8.L/14.L,
    0.L/14.L,  4.L/14.L,  1.L/14.L,  9.L/14.L,
    0.L/14.L,  3.L/14.L, 10.L/14.L,  1.L/14.L,
    0.L/14.L,  3.L/14.L,  9.L/14.L,  2.L/14.L,
    0.L/14.L,  3.L/14.L,  8.L/14.L,  3.L/14.L,
    0.L/14.L,  3.L/14.L,  7.L/14.L,  4.L/14.L,
    0.L/14.L,  3.L/14.L,  6.L/14.L,  5.L/14.L,
    0.L/14.L,  3.L/14.L,  5.L/14.L,  6.L/14.L,
    0.L/14.L,  3.L/14.L,  4.L/14.L,  7.L/14.L,
    0.L/14.L,  3.L/14.L,  3.L/14.L,  8.L/14.L,
    0.L/14.L,  3.L/14.L,  2.L/14.L,  9.L/14.L,
    0.L/14.L,  3.L/14.L,  1.L/14.L, 10.L/14.L,
    0.L/14.L,  2.L/14.L, 11.L/14.L,  1.L/14.L,
    0.L/14.L,  2.L/14.L, 10.L/14.L,  2.L/14.L,
    0.L/14.L,  2.L/14.L,  9.L/14.L,  3.L/14.L,
    0.L/14.L,  2.L/14.L,  8.L/14.L,  4.L/14.L,
    0.L/14.L,  2.L/14.L,  7.L/14.L,  5.L/14.L,
    0.L/14.L,  2.L/14.L,  6.L/14.L,  6.L/14.L,
    0.L/14.L,  2.L/14.L,  5.L/14.L,  7.L/14.L,
    0.L/14.L,  2.L/14.L,  4.L/14.L,  8.L/14.L,
    0.L/14.L,  2.L/14.L,  3.L/14.L,  9.L/14.L,
    0.L/14.L,  2.L/14.L,  2.L/14.L, 10.L/14.L,
    0.L/14.L,  2.L/14.L,  1.L/14.L, 11.L/14.L,
    0.L/14.L,  1.L/14.L, 12.L/14.L,  1.L/14.L,
    0.L/14.L,  1.L/14.L, 11.L/14.L,  2.L/14.L,
    0.L/14.L,  1.L/14.L, 10.L/14.L,  3.L/14.L,
    0.L/14.L,  1.L/14.L,  9.L/14.L,  4.L/14.L,
    0.L/14.L,  1.L/14.L,  8.L/14.L,  5.L/14.L,
    0.L/14.L,  1.L/14.L,  7.L/14.L,  6.L/14.L,
    0.L/14.L,  1.L/14.L,  6.L/14.L,  7.L/14.L,
    0.L/14.L,  1.L/14.L,  5.L/14.L,  8.L/14.L,
    0.L/14.L,  1.L/14.L,  4.L/14.L,  9.L/14.L,
    0.L/14.L,  1.L/14.L,  3.L/14.L, 10.L/14.L,
    0.L/14.L,  1.L/14.L,  2.L/14.L, 11.L/14.L,
    0.L/14.L,  1.L/14.L,  1.L/14.L, 12.L/14.L,
   12.L/14.L,  0.L/14.L,  1.L/14.L,  1.L/14.L,
   11.L/14.L,  0.L/14.L,  2.L/14.L,  1.L/14.L,
   11.L/14.L,  0.L/14.L,  1.L/14.L,  2.L/14.L,
   10.L/14.L,  0.L/14.L,  3.L/14.L,  1.L/14.L,
   10.L/14.L,  0.L/14.L,  2.L/14.L,  2.L/14.L,
   10.L/14.L,  0.L/14.L,  1.L/14.L,  3.L/14.L,
    9.L/14.L,  0.L/14.L,  4.L/14.L,  1.L/14.L,
    9.L/14.L,  0.L/14.L,  3.L/14.L,  2.L/14.L,
    9.L/14.L,  0.L/14.L,  2.L/14.L,  3.L/14.L,
    9.L/14.L,  0.L/14.L,  1.L/14.L,  4.L/14.L,
    8.L/14.L,  0.L/14.L,  5.L/14.L,  1.L/14.L,
    8.L/14.L,  0.L/14.L,  4.L/14.L,  2.L/14.L,
    8.L/14.L,  0.L/14.L,  3.L/14.L,  3.L/14.L,
    8.L/14.L,  0.L/14.L,  2.L/14.L,  4.L/14.L,
    8.L/14.L,  0.L/14.L,  1.L/14.L,  5.L/14.L,
    7.L/14.L,  0.L/14.L,  6.L/14.L,  1.L/14.L,
    7.L/14.L,  0.L/14.L,  5.L/14.L,  2.L/14.L,
    7.L/14.L,  0.L/14.L,  4.L/14.L,  3.L/14.L,
    7.L/14.L,  0.L/14.L,  3.L/14.L,  4.L/14.L,
    7.L/14.L,  0.L/14.L,  2.L/14.L,  5.L/14.L,
    7.L/14.L,  0.L/14.L,  1.L/14.L,  6.L/14.L,
    6.L/14.L,  0.L/14.L,  7.L/14.L,  1.L/14.L,
    6.L/14.L,  0.L/14.L,  6.L/14.L,  2.L/14.L,
    6.L/14.L,  0.L/14.L,  5.L/14.L,  3.L/14.L,
    6.L/14.L,  0.L/14.L,  4.L/14.L,  4.L/14.L,
    6.L/14.L,  0.L/14.L,  3.L/14.L,  5.L/14.L,
    6.L/14.L,  0.L/14.L,  2.L/14.L,  6.L/14.L,
    6.L/14.L,  0.L/14.L,  1.L/14.L,  7.L/14.L,
    5.L/14.L,  0.L/14.L,  8.L/14.L,  1.L/14.L,
    5.L/14.L,  0.L/14.L,  7.L/14.L,  2.L/14.L,
    5.L/14.L,  0.L/14.L,  6.L/14.L,  3.L/14.L,
    5.L/14.L,  0.L/14.L,  5.L/14.L,  4.L/14.L,
    5.L/14.L,  0.L/14.L,  4.L/14.L,  5.L/14.L,
    5.L/14.L,  0.L/14.L,  3.L/14.L,  6.L/14.L,
    5.L/14.L,  0.L/14.L,  2.L/14.L,  7.L/14.L,
    5.L/14.L,  0.L/14.L,  1.L/14.L,  8.L/14.L,
    4.L/14.L,  0.L/14.L,  9.L/14.L,  1.L/14.L,
    4.L/14.L,  0.L/14.L,  8.L/14.L,  2.L/14.L,
    4.L/14.L,  0.L/14.L,  7.L/14.L,  3.L/14.L,
    4.L/14.L,  0.L/14.L,  6.L/14.L,  4.L/14.L,
    4.L/14.L,  0.L/14.L,  5.L/14.L,  5.L/14.L,
    4.L/14.L,  0.L/14.L,  4.L/14.L,  6.L/14.L,
    4.L/14.L,  0.L/14.L,  3.L/14.L,  7.L/14.L,
    4.L/14.L,  0.L/14.L,  2.L/14.L,  8.L/14.L,
    4.L/14.L,  0.L/14.L,  1.L/14.L,  9.L/14.L,
    3.L/14.L,  0.L/14.L, 10.L/14.L,  1.L/14.L,
    3.L/14.L,  0.L/14.L,  9.L/14.L,  2.L/14.L,
    3.L/14.L,  0.L/14.L,  8.L/14.L,  3.L/14.L,
    3.L/14.L,  0.L/14.L,  7.L/14.L,  4.L/14.L,
    3.L/14.L,  0.L/14.L,  6.L/14.L,  5.L/14.L,
    3.L/14.L,  0.L/14.L,  5.L/14.L,  6.L/14.L,
    3.L/14.L,  0.L/14.L,  4.L/14.L,  7.L/14.L,
    3.L/14.L,  0.L/14.L,  3.L/14.L,  8.L/14.L,
    3.L/14.L,  0.L/14.L,  2.L/14.L,  9.L/14.L,
    3.L/14.L,  0.L/14.L,  1.L/14.L, 10.L/14.L,
    2.L/14.L,  0.L/14.L, 11.L/14.L,  1.L/14.L,
    2.L/14.L,  0.L/14.L, 10.L/14.L,  2.L/14.L,
    2.L/14.L,  0.L/14.L,  9.L/14.L,  3.L/14.L,
    2.L/14.L,  0.L/14.L,  8.L/14.L,  4.L/14.L,
    2.L/14.L,  0.L/14.L,  7.L/14.L,  5.L/14.L,
    2.L/14.L,  0.L/14.L,  6.L/14.L,  6.L/14.L,
    2.L/14.L,  0.L/14.L,  5.L/14.L,  7.L/14.L,
    2.L/14.L,  0.L/14.L,  4.L/14.L,  8.L/14.L,
    2.L/14.L,  0.L/14.L,  3.L/14.L,  9.L/14.L,
    2.L/14.L,  0.L/14.L,  2.L/14.L, 10.L/14.L,
    2.L/14.L,  0.L/14.L,  1.L/14.L, 11.L/14.L,
    1.L/14.L,  0.L/14.L, 12.L/14.L,  1.L/14.L,
    1.L/14.L,  0.L/14.L, 11.L/14.L,  2.L/14.L,
    1.L/14.L,  0.L/14.L, 10.L/14.L,  3.L/14.L,
    1.L/14.L,  0.L/14.L,  9.L/14.L,  4.L/14.L,
    1.L/14.L,  0.L/14.L,  8.L/14.L,  5.L/14.L,
    1.L/14.L,  0.L/14.L,  7.L/14.L,  6.L/14.L,
    1.L/14.L,  0.L/14.L,  6.L/14.L,  7.L/14.L,
    1.L/14.L,  0.L/14.L,  5.L/14.L,  8.L/14.L,
    1.L/14.L,  0.L/14.L,  4.L/14.L,  9.L/14.L,
    1.L/14.L,  0.L/14.L,  3.L/14.L, 10.L/14.L,
    1.L/14.L,  0.L/14.L,  2.L/14.L, 11.L/14.L,
    1.L/14.L,  0.L/14.L,  1.L/14.L, 12.L/14.L,
   12.L/14.L,  1.L/14.L,  0.L/14.L,  1.L/14.L,
   11.L/14.L,  2.L/14.L,  0.L/14.L,  1.L/14.L,
   11.L/14.L,  1.L/14.L,  0.L/14.L,  2.L/14.L,
   10.L/14.L,  3.L/14.L,  0.L/14.L,  1.L/14.L,
   10.L/14.L,  2.L/14.L,  0.L/14.L,  2.L/14.L,
   10.L/14.L,  1.L/14.L,  0.L/14.L,  3.L/14.L,
    9.L/14.L,  4.L/14.L,  0.L/14.L,  1.L/14.L,
    9.L/14.L,  3.L/14.L,  0.L/14.L,  2.L/14.L,
    9.L/14.L,  2.L/14.L,  0.L/14.L,  3.L/14.L,
    9.L/14.L,  1.L/14.L,  0.L/14.L,  4.L/14.L,
    8.L/14.L,  5.L/14.L,  0.L/14.L,  1.L/14.L,
    8.L/14.L,  4.L/14.L,  0.L/14.L,  2.L/14.L,
    8.L/14.L,  3.L/14.L,  0.L/14.L,  3.L/14.L,
    8.L/14.L,  2.L/14.L,  0.L/14.L,  4.L/14.L,
    8.L/14.L,  1.L/14.L,  0.L/14.L,  5.L/14.L,
    7.L/14.L,  6.L/14.L,  0.L/14.L,  1.L/14.L,
    7.L/14.L,  5.L/14.L,  0.L/14.L,  2.L/14.L,
    7.L/14.L,  4.L/14.L,  0.L/14.L,  3.L/14.L,
    7.L/14.L,  3.L/14.L,  0.L/14.L,  4.L/14.L,
    7.L/14.L,  2.L/14.L,  0.L/14.L,  5.L/14.L,
    7.L/14.L,  1.L/14.L,  0.L/14.L,  6.L/14.L,
    6.L/14.L,  7.L/14.L,  0.L/14.L,  1.L/14.L,
    6.L/14.L,  6.L/14.L,  0.L/14.L,  2.L/14.L,
    6.L/14.L,  5.L/14.L,  0.L/14.L,  3.L/14.L,
    6.L/14.L,  4.L/14.L,  0.L/14.L,  4.L/14.L,
    6.L/14.L,  3.L/14.L,  0.L/14.L,  5.L/14.L,
    6.L/14.L,  2.L/14.L,  0.L/14.L,  6.L/14.L,
    6.L/14.L,  1.L/14.L,  0.L/14.L,  7.L/14.L,
    5.L/14.L,  8.L/14.L,  0.L/14.L,  1.L/14.L,
    5.L/14.L,  7.L/14.L,  0.L/14.L,  2.L/14.L,
    5.L/14.L,  6.L/14.L,  0.L/14.L,  3.L/14.L,
    5.L/14.L,  5.L/14.L,  0.L/14.L,  4.L/14.L,
    5.L/14.L,  4.L/14.L,  0.L/14.L,  5.L/14.L,
    5.L/14.L,  3.L/14.L,  0.L/14.L,  6.L/14.L,
    5.L/14.L,  2.L/14.L,  0.L/14.L,  7.L/14.L,
    5.L/14.L,  1.L/14.L,  0.L/14.L,  8.L/14.L,
    4.L/14.L,  9.L/14.L,  0.L/14.L,  1.L/14.L,
    4.L/14.L,  8.L/14.L,  0.L/14.L,  2.L/14.L,
    4.L/14.L,  7.L/14.L,  0.L/14.L,  3.L/14.L,
    4.L/14.L,  6.L/14.L,  0.L/14.L,  4.L/14.L,
    4.L/14.L,  5.L/14.L,  0.L/14.L,  5.L/14.L,
    4.L/14.L,  4.L/14.L,  0.L/14.L,  6.L/14.L,
    4.L/14.L,  3.L/14.L,  0.L/14.L,  7.L/14.L,
    4.L/14.L,  2.L/14.L,  0.L/14.L,  8.L/14.L,
    4.L/14.L,  1.L/14.L,  0.L/14.L,  9.L/14.L,
    3.L/14.L, 10.L/14.L,  0.L/14.L,  1.L/14.L,
    3.L/14.L,  9.L/14.L,  0.L/14.L,  2.L/14.L,
    3.L/14.L,  8.L/14.L,  0.L/14.L,  3.L/14.L,
    3.L/14.L,  7.L/14.L,  0.L/14.L,  4.L/14.L,
    3.L/14.L,  6.L/14.L,  0.L/14.L,  5.L/14.L,
    3.L/14.L,  5.L/14.L,  0.L/14.L,  6.L/14.L,
    3.L/14.L,  4.L/14.L,  0.L/14.L,  7.L/14.L,
    3.L/14.L,  3.L/14.L,  0.L/14.L,  8.L/14.L,
    3.L/14.L,  2.L/14.L,  0.L/14.L,  9.L/14.L,
    3.L/14.L,  1.L/14.L,  0.L/14.L, 10.L/14.L,
    2.L/14.L, 11.L/14.L,  0.L/14.L,  1.L/14.L,
    2.L/14.L, 10.L/14.L,  0.L/14.L,  2.L/14.L,
    2.L/14.L,  9.L/14.L,  0.L/14.L,  3.L/14.L,
    2.L/14.L,  8.L/14.L,  0.L/14.L,  4.L/14.L,
    2.L/14.L,  7.L/14.L,  0.L/14.L,  5.L/14.L,
    2.L/14.L,  6.L/14.L,  0.L/14.L,  6.L/14.L,
    2.L/14.L,  5.L/14.L,  0.L/14.L,  7.L/14.L,
    2.L/14.L,  4.L/14.L,  0.L/14.L,  8.L/14.L,
    2.L/14.L,  3.L/14.L,  0.L/14.L,  9.L/14.L,
    2.L/14.L,  2.L/14.L,  0.L/14.L, 10.L/14.L,
    2.L/14.L,  1.L/14.L,  0.L/14.L, 11.L/14.L,
    1.L/14.L, 12.L/14.L,  0.L/14.L,  1.L/14.L,
    1.L/14.L, 11.L/14.L,  0.L/14.L,  2.L/14.L,
    1.L/14.L, 10.L/14.L,  0.L/14.L,  3.L/14.L,
    1.L/14.L,  9.L/14.L,  0.L/14.L,  4.L/14.L,
    1.L/14.L,  8.L/14.L,  0.L/14.L,  5.L/14.L,
    1.L/14.L,  7.L/14.L,  0.L/14.L,  6.L/14.L,
    1.L/14.L,  6.L/14.L,  0.L/14.L,  7.L/14.L,
    1.L/14.L,  5.L/14.L,  0.L/14.L,  8.L/14.L,
    1.L/14.L,  4.L/14.L,  0.L/14.L,  9.L/14.L,
    1.L/14.L,  3.L/14.L,  0.L/14.L, 10.L/14.L,
    1.L/14.L,  2.L/14.L,  0.L/14.L, 11.L/14.L,
    1.L/14.L,  1.L/14.L,  0.L/14.L, 12.L/14.L,
   12.L/14.L,  1.L/14.L,  1.L/14.L,  0.L/14.L,
   11.L/14.L,  2.L/14.L,  1.L/14.L,  0.L/14.L,
   11.L/14.L,  1.L/14.L,  2.L/14.L,  0.L/14.L,
   10.L/14.L,  3.L/14.L,  1.L/14.L,  0.L/14.L,
   10.L/14.L,  2.L/14.L,  2.L/14.L,  0.L/14.L,
   10.L/14.L,  1.L/14.L,  3.L/14.L,  0.L/14.L,
    9.L/14.L,  4.L/14.L,  1.L/14.L,  0.L/14.L,
    9.L/14.L,  3.L/14.L,  2.L/14.L,  0.L/14.L,
    9.L/14.L,  2.L/14.L,  3.L/14.L,  0.L/14.L,
    9.L/14.L,  1.L/14.L,  4.L/14.L,  0.L/14.L,
    8.L/14.L,  5.L/14.L,  1.L/14.L,  0.L/14.L,
    8.L/14.L,  4.L/14.L,  2.L/14.L,  0.L/14.L,
    8.L/14.L,  3.L/14.L,  3.L/14.L,  0.L/14.L,
    8.L/14.L,  2.L/14.L,  4.L/14.L,  0.L/14.L,
    8.L/14.L,  1.L/14.L,  5.L/14.L,  0.L/14.L,
    7.L/14.L,  6.L/14.L,  1.L/14.L,  0.L/14.L,
    7.L/14.L,  5.L/14.L,  2.L/14.L,  0.L/14.L,
    7.L/14.L,  4.L/14.L,  3.L/14.L,  0.L/14.L,
    7.L/14.L,  3.L/14.L,  4.L/14.L,  0.L/14.L,
    7.L/14.L,  2.L/14.L,  5.L/14.L,  0.L/14.L,
    7.L/14.L,  1.L/14.L,  6.L/14.L,  0.L/14.L,
    6.L/14.L,  7.L/14.L,  1.L/14.L,  0.L/14.L,
    6.L/14.L,  6.L/14.L,  2.L/14.L,  0.L/14.L,
    6.L/14.L,  5.L/14.L,  3.L/14.L,  0.L/14.L,
    6.L/14.L,  4.L/14.L,  4.L/14.L,  0.L/14.L,
    6.L/14.L,  3.L/14.L,  5.L/14.L,  0.L/14.L,
    6.L/14.L,  2.L/14.L,  6.L/14.L,  0.L/14.L,
    6.L/14.L,  1.L/14.L,  7.L/14.L,  0.L/14.L,
    5.L/14.L,  8.L/14.L,  1.L/14.L,  0.L/14.L,
    5.L/14.L,  7.L/14.L,  2.L/14.L,  0.L/14.L,
    5.L/14.L,  6.L/14.L,  3.L/14.L,  0.L/14.L,
    5.L/14.L,  5.L/14.L,  4.L/14.L,  0.L/14.L,
    5.L/14.L,  4.L/14.L,  5.L/14.L,  0.L/14.L,
    5.L/14.L,  3.L/14.L,  6.L/14.L,  0.L/14.L,
    5.L/14.L,  2.L/14.L,  7.L/14.L,  0.L/14.L,
    5.L/14.L,  1.L/14.L,  8.L/14.L,  0.L/14.L,
    4.L/14.L,  9.L/14.L,  1.L/14.L,  0.L/14.L,
    4.L/14.L,  8.L/14.L,  2.L/14.L,  0.L/14.L,
    4.L/14.L,  7.L/14.L,  3.L/14.L,  0.L/14.L,
    4.L/14.L,  6.L/14.L,  4.L/14.L,  0.L/14.L,
    4.L/14.L,  5.L/14.L,  5.L/14.L,  0.L/14.L,
    4.L/14.L,  4.L/14.L,  6.L/14.L,  0.L/14.L,
    4.L/14.L,  3.L/14.L,  7.L/14.L,  0.L/14.L,
    4.L/14.L,  2.L/14.L,  8.L/14.L,  0.L/14.L,
    4.L/14.L,  1.L/14.L,  9.L/14.L,  0.L/14.L,
    3.L/14.L, 10.L/14.L,  1.L/14.L,  0.L/14.L,
    3.L/14.L,  9.L/14.L,  2.L/14.L,  0.L/14.L,
    3.L/14.L,  8.L/14.L,  3.L/14.L,  0.L/14.L,
    3.L/14.L,  7.L/14.L,  4.L/14.L,  0.L/14.L,
    3.L/14.L,  6.L/14.L,  5.L/14.L,  0.L/14.L,
    3.L/14.L,  5.L/14.L,  6.L/14.L,  0.L/14.L,
    3.L/14.L,  4.L/14.L,  7.L/14.L,  0.L/14.L,
    3.L/14.L,  3.L/14.L,  8.L/14.L,  0.L/14.L,
    3.L/14.L,  2.L/14.L,  9.L/14.L,  0.L/14.L,
    3.L/14.L,  1.L/14.L, 10.L/14.L,  0.L/14.L,
    2.L/14.L, 11.L/14.L,  1.L/14.L,  0.L/14.L,
    2.L/14.L, 10.L/14.L,  2.L/14.L,  0.L/14.L,
    2.L/14.L,  9.L/14.L,  3.L/14.L,  0.L/14.L,
    2.L/14.L,  8.L/14.L,  4.L/14.L,  0.L/14.L,
    2.L/14.L,  7.L/14.L,  5.L/14.L,  0.L/14.L,
    2.L/14.L,  6.L/14.L,  6.L/14.L,  0.L/14.L,
    2.L/14.L,  5.L/14.L,  7.L/14.L,  0.L/14.L,
    2.L/14.L,  4.L/14.L,  8.L/14.L,  0.L/14.L,
    2.L/14.L,  3.L/14.L,  9.L/14.L,  0.L/14.L,
    2.L/14.L,  2.L/14.L, 10.L/14.L,  0.L/14.L,
    2.L/14.L,  1.L/14.L, 11.L/14.L,  0.L/14.L,
    1.L/14.L, 12.L/14.L,  1.L/14.L,  0.L/14.L,
    1.L/14.L, 11.L/14.L,  2.L/14.L,  0.L/14.L,
    1.L/14.L, 10.L/14.L,  3.L/14.L,  0.L/14.L,
    1.L/14.L,  9.L/14.L,  4.L/14.L,  0.L/14.L,
    1.L/14.L,  8.L/14.L,  5.L/14.L,  0.L/14.L,
    1.L/14.L,  7.L/14.L,  6.L/14.L,  0.L/14.L,
    1.L/14.L,  6.L/14.L,  7.L/14.L,  0.L/14.L,
    1.L/14.L,  5.L/14.L,  8.L/14.L,  0.L/14.L,
    1.L/14.L,  4.L/14.L,  9.L/14.L,  0.L/14.L,
    1.L/14.L,  3.L/14.L, 10.L/14.L,  0.L/14.L,
    1.L/14.L,  2.L/14.L, 11.L/14.L,  0.L/14.L,
    1.L/14.L,  1.L/14.L, 12.L/14.L,  0.L/14.L,
   11.L/14.L,  1.L/14.L,  1.L/14.L,  1.L/14.L,
   10.L/14.L,  2.L/14.L,  1.L/14.L,  1.L/14.L,
   10.L/14.L,  1.L/14.L,  2.L/14.L,  1.L/14.L,
   10.L/14.L,  1.L/14.L,  1.L/14.L,  2.L/14.L,
    9.L/14.L,  3.L/14.L,  1.L/14.L,  1.L/14.L,
    9.L/14.L,  2.L/14.L,  2.L/14.L,  1.L/14.L,
    9.L/14.L,  2.L/14.L,  1.L/14.L,  2.L/14.L,
    9.L/14.L,  1.L/14.L,  3.L/14.L,  1.L/14.L,
    9.L/14.L,  1.L/14.L,  2.L/14.L,  2.L/14.L,
    9.L/14.L,  1.L/14.L,  1.L/14.L,  3.L/14.L,
    8.L/14.L,  4.L/14.L,  1.L/14.L,  1.L/14.L,
    8.L/14.L,  3.L/14.L,  2.L/14.L,  1.L/14.L,
    8.L/14.L,  3.L/14.L,  1.L/14.L,  2.L/14.L,
    8.L/14.L,  2.L/14.L,  3.L/14.L,  1.L/14.L,
    8.L/14.L,  2.L/14.L,  2.L/14.L,  2.L/14.L,
    8.L/14.L,  2.L/14.L,  1.L/14.L,  3.L/14.L,
    8.L/14.L,  1.L/14.L,  4.L/14.L,  1.L/14.L,
    8.L/14.L,  1.L/14.L,  3.L/14.L,  2.L/14.L,
    8.L/14.L,  1.L/14.L,  2.L/14.L,  3.L/14.L,
    8.L/14.L,  1.L/14.L,  1.L/14.L,  4.L/14.L,
    7.L/14.L,  5.L/14.L,  1.L/14.L,  1.L/14.L,
    7.L/14.L,  4.L/14.L,  2.L/14.L,  1.L/14.L,
    7.L/14.L,  4.L/14.L,  1.L/14.L,  2.L/14.L,
    7.L/14.L,  3.L/14.L,  3.L/14.L,  1.L/14.L,
    7.L/14.L,  3.L/14.L,  2.L/14.L,  2.L/14.L,
    7.L/14.L,  3.L/14.L,  1.L/14.L,  3.L/14.L,
    7.L/14.L,  2.L/14.L,  4.L/14.L,  1.L/14.L,
    7.L/14.L,  2.L/14.L,  3.L/14.L,  2.L/14.L,
    7.L/14.L,  2.L/14.L,  2.L/14.L,  3.L/14.L,
    7.L/14.L,  2.L/14.L,  1.L/14.L,  4.L/14.L,
    7.L/14.L,  1.L/14.L,  5.L/14.L,  1.L/14.L,
    7.L/14.L,  1.L/14.L,  4.L/14.L,  2.L/14.L,
    7.L/14.L,  1.L/14.L,  3.L/14.L,  3.L/14.L,
    7.L/14.L,  1.L/14.L,  2.L/14.L,  4.L/14.L,
    7.L/14.L,  1.L/14.L,  1.L/14.L,  5.L/14.L,
    6.L/14.L,  6.L/14.L,  1.L/14.L,  1.L/14.L,
    6.L/14.L,  5.L/14.L,  2.L/14.L,  1.L/14.L,
    6.L/14.L,  5.L/14.L,  1.L/14.L,  2.L/14.L,
    6.L/14.L,  4.L/14.L,  3.L/14.L,  1.L/14.L,
    6.L/14.L,  4.L/14.L,  2.L/14.L,  2.L/14.L,
    6.L/14.L,  4.L/14.L,  1.L/14.L,  3.L/14.L,
    6.L/14.L,  3.L/14.L,  4.L/14.L,  1.L/14.L,
    6.L/14.L,  3.L/14.L,  3.L/14.L,  2.L/14.L,
    6.L/14.L,  3.L/14.L,  2.L/14.L,  3.L/14.L,
    6.L/14.L,  3.L/14.L,  1.L/14.L,  4.L/14.L,
    6.L/14.L,  2.L/14.L,  5.L/14.L,  1.L/14.L,
    6.L/14.L,  2.L/14.L,  4.L/14.L,  2.L/14.L,
    6.L/14.L,  2.L/14.L,  3.L/14.L,  3.L/14.L,
    6.L/14.L,  2.L/14.L,  2.L/14.L,  4.L/14.L,
    6.L/14.L,  2.L/14.L,  1.L/14.L,  5.L/14.L,
    6.L/14.L,  1.L/14.L,  6.L/14.L,  1.L/14.L,
    6.L/14.L,  1.L/14.L,  5.L/14.L,  2.L/14.L,
    6.L/14.L,  1.L/14.L,  4.L/14.L,  3.L/14.L,
    6.L/14.L,  1.L/14.L,  3.L/14.L,  4.L/14.L,
    6.L/14.L,  1.L/14.L,  2.L/14.L,  5.L/14.L,
    6.L/14.L,  1.L/14.L,  1.L/14.L,  6.L/14.L,
    5.L/14.L,  7.L/14.L,  1.L/14.L,  1.L/14.L,
    5.L/14.L,  6.L/14.L,  2.L/14.L,  1.L/14.L,
    5.L/14.L,  6.L/14.L,  1.L/14.L,  2.L/14.L,
    5.L/14.L,  5.L/14.L,  3.L/14.L,  1.L/14.L,
    5.L/14.L,  5.L/14.L,  2.L/14.L,  2.L/14.L,
    5.L/14.L,  5.L/14.L,  1.L/14.L,  3.L/14.L,
    5.L/14.L,  4.L/14.L,  4.L/14.L,  1.L/14.L,
    5.L/14.L,  4.L/14.L,  3.L/14.L,  2.L/14.L,
    5.L/14.L,  4.L/14.L,  2.L/14.L,  3.L/14.L,
    5.L/14.L,  4.L/14.L,  1.L/14.L,  4.L/14.L,
    5.L/14.L,  3.L/14.L,  5.L/14.L,  1.L/14.L,
    5.L/14.L,  3.L/14.L,  4.L/14.L,  2.L/14.L,
    5.L/14.L,  3.L/14.L,  3.L/14.L,  3.L/14.L,
    5.L/14.L,  3.L/14.L,  2.L/14.L,  4.L/14.L,
    5.L/14.L,  3.L/14.L,  1.L/14.L,  5.L/14.L,
    5.L/14.L,  2.L/14.L,  6.L/14.L,  1.L/14.L,
    5.L/14.L,  2.L/14.L,  5.L/14.L,  2.L/14.L,
    5.L/14.L,  2.L/14.L,  4.L/14.L,  3.L/14.L,
    5.L/14.L,  2.L/14.L,  3.L/14.L,  4.L/14.L,
    5.L/14.L,  2.L/14.L,  2.L/14.L,  5.L/14.L,
    5.L/14.L,  2.L/14.L,  1.L/14.L,  6.L/14.L,
    5.L/14.L,  1.L/14.L,  7.L/14.L,  1.L/14.L,
    5.L/14.L,  1.L/14.L,  6.L/14.L,  2.L/14.L,
    5.L/14.L,  1.L/14.L,  5.L/14.L,  3.L/14.L,
    5.L/14.L,  1.L/14.L,  4.L/14.L,  4.L/14.L,
    5.L/14.L,  1.L/14.L,  3.L/14.L,  5.L/14.L,
    5.L/14.L,  1.L/14.L,  2.L/14.L,  6.L/14.L,
    5.L/14.L,  1.L/14.L,  1.L/14.L,  7.L/14.L,
    4.L/14.L,  8.L/14.L,  1.L/14.L,  1.L/14.L,
    4.L/14.L,  7.L/14.L,  2.L/14.L,  1.L/14.L,
    4.L/14.L,  7.L/14.L,  1.L/14.L,  2.L/14.L,
    4.L/14.L,  6.L/14.L,  3.L/14.L,  1.L/14.L,
    4.L/14.L,  6.L/14.L,  2.L/14.L,  2.L/14.L,
    4.L/14.L,  6.L/14.L,  1.L/14.L,  3.L/14.L,
    4.L/14.L,  5.L/14.L,  4.L/14.L,  1.L/14.L,
    4.L/14.L,  5.L/14.L,  3.L/14.L,  2.L/14.L,
    4.L/14.L,  5.L/14.L,  2.L/14.L,  3.L/14.L,
    4.L/14.L,  5.L/14.L,  1.L/14.L,  4.L/14.L,
    4.L/14.L,  4.L/14.L,  5.L/14.L,  1.L/14.L,
    4.L/14.L,  4.L/14.L,  4.L/14.L,  2.L/14.L,
    4.L/14.L,  4.L/14.L,  3.L/14.L,  3.L/14.L,
    4.L/14.L,  4.L/14.L,  2.L/14.L,  4.L/14.L,
    4.L/14.L,  4.L/14.L,  1.L/14.L,  5.L/14.L,
    4.L/14.L,  3.L/14.L,  6.L/14.L,  1.L/14.L,
    4.L/14.L,  3.L/14.L,  5.L/14.L,  2.L/14.L,
    4.L/14.L,  3.L/14.L,  4.L/14.L,  3.L/14.L,
    4.L/14.L,  3.L/14.L,  3.L/14.L,  4.L/14.L,
    4.L/14.L,  3.L/14.L,  2.L/14.L,  5.L/14.L,
    4.L/14.L,  3.L/14.L,  1.L/14.L,  6.L/14.L,
    4.L/14.L,  2.L/14.L,  7.L/14.L,  1.L/14.L,
    4.L/14.L,  2.L/14.L,  6.L/14.L,  2.L/14.L,
    4.L/14.L,  2.L/14.L,  5.L/14.L,  3.L/14.L,
    4.L/14.L,  2.L/14.L,  4.L/14.L,  4.L/14.L,
    4.L/14.L,  2.L/14.L,  3.L/14.L,  5.L/14.L,
    4.L/14.L,  2.L/14.L,  2.L/14.L,  6.L/14.L,
    4.L/14.L,  2.L/14.L,  1.L/14.L,  7.L/14.L,
    4.L/14.L,  1.L/14.L,  8.L/14.L,  1.L/14.L,
    4.L/14.L,  1.L/14.L,  7.L/14.L,  2.L/14.L,
    4.L/14.L,  1.L/14.L,  6.L/14.L,  3.L/14.L,
    4.L/14.L,  1.L/14.L,  5.L/14.L,  4.L/14.L,
    4.L/14.L,  1.L/14.L,  4.L/14.L,  5.L/14.L,
    4.L/14.L,  1.L/14.L,  3.L/14.L,  6.L/14.L,
    4.L/14.L,  1.L/14.L,  2.L/14.L,  7.L/14.L,
    4.L/14.L,  1.L/14.L,  1.L/14.L,  8.L/14.L,
    3.L/14.L,  9.L/14.L,  1.L/14.L,  1.L/14.L,
    3.L/14.L,  8.L/14.L,  2.L/14.L,  1.L/14.L,
    3.L/14.L,  8.L/14.L,  1.L/14.L,  2.L/14.L,
    3.L/14.L,  7.L/14.L,  3.L/14.L,  1.L/14.L,
    3.L/14.L,  7.L/14.L,  2.L/14.L,  2.L/14.L,
    3.L/14.L,  7.L/14.L,  1.L/14.L,  3.L/14.L,
    3.L/14.L,  6.L/14.L,  4.L/14.L,  1.L/14.L,
    3.L/14.L,  6.L/14.L,  3.L/14.L,  2.L/14.L,
    3.L/14.L,  6.L/14.L,  2.L/14.L,  3.L/14.L,
    3.L/14.L,  6.L/14.L,  1.L/14.L,  4.L/14.L,
    3.L/14.L,  5.L/14.L,  5.L/14.L,  1.L/14.L,
    3.L/14.L,  5.L/14.L,  4.L/14.L,  2.L/14.L,
    3.L/14.L,  5.L/14.L,  3.L/14.L,  3.L/14.L,
    3.L/14.L,  5.L/14.L,  2.L/14.L,  4.L/14.L,
    3.L/14.L,  5.L/14.L,  1.L/14.L,  5.L/14.L,
    3.L/14.L,  4.L/14.L,  6.L/14.L,  1.L/14.L,
    3.L/14.L,  4.L/14.L,  5.L/14.L,  2.L/14.L,
    3.L/14.L,  4.L/14.L,  4.L/14.L,  3.L/14.L,
    3.L/14.L,  4.L/14.L,  3.L/14.L,  4.L/14.L,
    3.L/14.L,  4.L/14.L,  2.L/14.L,  5.L/14.L,
    3.L/14.L,  4.L/14.L,  1.L/14.L,  6.L/14.L,
    3.L/14.L,  3.L/14.L,  7.L/14.L,  1.L/14.L,
    3.L/14.L,  3.L/14.L,  6.L/14.L,  2.L/14.L,
    3.L/14.L,  3.L/14.L,  5.L/14.L,  3.L/14.L,
    3.L/14.L,  3.L/14.L,  4.L/14.L,  4.L/14.L,
    3.L/14.L,  3.L/14.L,  3.L/14.L,  5.L/14.L,
    3.L/14.L,  3.L/14.L,  2.L/14.L,  6.L/14.L,
    3.L/14.L,  3.L/14.L,  1.L/14.L,  7.L/14.L,
    3.L/14.L,  2.L/14.L,  8.L/14.L,  1.L/14.L,
    3.L/14.L,  2.L/14.L,  7.L/14.L,  2.L/14.L,
    3.L/14.L,  2.L/14.L,  6.L/14.L,  3.L/14.L,
    3.L/14.L,  2.L/14.L,  5.L/14.L,  4.L/14.L,
    3.L/14.L,  2.L/14.L,  4.L/14.L,  5.L/14.L,
    3.L/14.L,  2.L/14.L,  3.L/14.L,  6.L/14.L,
    3.L/14.L,  2.L/14.L,  2.L/14.L,  7.L/14.L,
    3.L/14.L,  2.L/14.L,  1.L/14.L,  8.L/14.L,
    3.L/14.L,  1.L/14.L,  9.L/14.L,  1.L/14.L,
    3.L/14.L,  1.L/14.L,  8.L/14.L,  2.L/14.L,
    3.L/14.L,  1.L/14.L,  7.L/14.L,  3.L/14.L,
    3.L/14.L,  1.L/14.L,  6.L/14.L,  4.L/14.L,
    3.L/14.L,  1.L/14.L,  5.L/14.L,  5.L/14.L,
    3.L/14.L,  1.L/14.L,  4.L/14.L,  6.L/14.L,
    3.L/14.L,  1.L/14.L,  3.L/14.L,  7.L/14.L,
    3.L/14.L,  1.L/14.L,  2.L/14.L,  8.L/14.L,
    3.L/14.L,  1.L/14.L,  1.L/14.L,  9.L/14.L,
    2.L/14.L, 10.L/14.L,  1.L/14.L,  1.L/14.L,
    2.L/14.L,  9.L/14.L,  2.L/14.L,  1.L/14.L,
    2.L/14.L,  9.L/14.L,  1.L/14.L,  2.L/14.L,
    2.L/14.L,  8.L/14.L,  3.L/14.L,  1.L/14.L,
    2.L/14.L,  8.L/14.L,  2.L/14.L,  2.L/14.L,
    2.L/14.L,  8.L/14.L,  1.L/14.L,  3.L/14.L,
    2.L/14.L,  7.L/14.L,  4.L/14.L,  1.L/14.L,
    2.L/14.L,  7.L/14.L,  3.L/14.L,  2.L/14.L,
    2.L/14.L,  7.L/14.L,  2.L/14.L,  3.L/14.L,
    2.L/14.L,  7.L/14.L,  1.L/14.L,  4.L/14.L,
    2.L/14.L,  6.L/14.L,  5.L/14.L,  1.L/14.L,
    2.L/14.L,  6.L/14.L,  4.L/14.L,  2.L/14.L,
    2.L/14.L,  6.L/14.L,  3.L/14.L,  3.L/14.L,
    2.L/14.L,  6.L/14.L,  2.L/14.L,  4.L/14.L,
    2.L/14.L,  6.L/14.L,  1.L/14.L,  5.L/14.L,
    2.L/14.L,  5.L/14.L,  6.L/14.L,  1.L/14.L,
    2.L/14.L,  5.L/14.L,  5.L/14.L,  2.L/14.L,
    2.L/14.L,  5.L/14.L,  4.L/14.L,  3.L/14.L,
    2.L/14.L,  5.L/14.L,  3.L/14.L,  4.L/14.L,
    2.L/14.L,  5.L/14.L,  2.L/14.L,  5.L/14.L,
    2.L/14.L,  5.L/14.L,  1.L/14.L,  6.L/14.L,
    2.L/14.L,  4.L/14.L,  7.L/14.L,  1.L/14.L,
    2.L/14.L,  4.L/14.L,  6.L/14.L,  2.L/14.L,
    2.L/14.L,  4.L/14.L,  5.L/14.L,  3.L/14.L,
    2.L/14.L,  4.L/14.L,  4.L/14.L,  4.L/14.L,
    2.L/14.L,  4.L/14.L,  3.L/14.L,  5.L/14.L,
    2.L/14.L,  4.L/14.L,  2.L/14.L,  6.L/14.L,
    2.L/14.L,  4.L/14.L,  1.L/14.L,  7.L/14.L,
    2.L/14.L,  3.L/14.L,  8.L/14.L,  1.L/14.L,
    2.L/14.L,  3.L/14.L,  7.L/14.L,  2.L/14.L,
    2.L/14.L,  3.L/14.L,  6.L/14.L,  3.L/14.L,
    2.L/14.L,  3.L/14.L,  5.L/14.L,  4.L/14.L,
    2.L/14.L,  3.L/14.L,  4.L/14.L,  5.L/14.L,
    2.L/14.L,  3.L/14.L,  3.L/14.L,  6.L/14.L,
    2.L/14.L,  3.L/14.L,  2.L/14.L,  7.L/14.L,
    2.L/14.L,  3.L/14.L,  1.L/14.L,  8.L/14.L,
    2.L/14.L,  2.L/14.L,  9.L/14.L,  1.L/14.L,
    2.L/14.L,  2.L/14.L,  8.L/14.L,  2.L/14.L,
    2.L/14.L,  2.L/14.L,  7.L/14.L,  3.L/14.L,
    2.L/14.L,  2.L/14.L,  6.L/14.L,  4.L/14.L,
    2.L/14.L,  2.L/14.L,  5.L/14.L,  5.L/14.L,
    2.L/14.L,  2.L/14.L,  4.L/14.L,  6.L/14.L,
    2.L/14.L,  2.L/14.L,  3.L/14.L,  7.L/14.L,
    2.L/14.L,  2.L/14.L,  2.L/14.L,  8.L/14.L,
    2.L/14.L,  2.L/14.L,  1.L/14.L,  9.L/14.L,
    2.L/14.L,  1.L/14.L, 10.L/14.L,  1.L/14.L,
    2.L/14.L,  1.L/14.L,  9.L/14.L,  2.L/14.L,
    2.L/14.L,  1.L/14.L,  8.L/14.L,  3.L/14.L,
    2.L/14.L,  1.L/14.L,  7.L/14.L,  4.L/14.L,
    2.L/14.L,  1.L/14.L,  6.L/14.L,  5.L/14.L,
    2.L/14.L,  1.L/14.L,  5.L/14.L,  6.L/14.L,
    2.L/14.L,  1.L/14.L,  4.L/14.L,  7.L/14.L,
    2.L/14.L,  1.L/14.L,  3.L/14.L,  8.L/14.L,
    2.L/14.L,  1.L/14.L,  2.L/14.L,  9.L/14.L,
    2.L/14.L,  1.L/14.L,  1.L/14.L, 10.L/14.L,
    1.L/14.L, 11.L/14.L,  1.L/14.L,  1.L/14.L,
    1.L/14.L, 10.L/14.L,  2.L/14.L,  1.L/14.L,
    1.L/14.L, 10.L/14.L,  1.L/14.L,  2.L/14.L,
    1.L/14.L,  9.L/14.L,  3.L/14.L,  1.L/14.L,
    1.L/14.L,  9.L/14.L,  2.L/14.L,  2.L/14.L,
    1.L/14.L,  9.L/14.L,  1.L/14.L,  3.L/14.L,
    1.L/14.L,  8.L/14.L,  4.L/14.L,  1.L/14.L,
    1.L/14.L,  8.L/14.L,  3.L/14.L,  2.L/14.L,
    1.L/14.L,  8.L/14.L,  2.L/14.L,  3.L/14.L,
    1.L/14.L,  8.L/14.L,  1.L/14.L,  4.L/14.L,
    1.L/14.L,  7.L/14.L,  5.L/14.L,  1.L/14.L,
    1.L/14.L,  7.L/14.L,  4.L/14.L,  2.L/14.L,
    1.L/14.L,  7.L/14.L,  3.L/14.L,  3.L/14.L,
    1.L/14.L,  7.L/14.L,  2.L/14.L,  4.L/14.L,
    1.L/14.L,  7.L/14.L,  1.L/14.L,  5.L/14.L,
    1.L/14.L,  6.L/14.L,  6.L/14.L,  1.L/14.L,
    1.L/14.L,  6.L/14.L,  5.L/14.L,  2.L/14.L,
    1.L/14.L,  6.L/14.L,  4.L/14.L,  3.L/14.L,
    1.L/14.L,  6.L/14.L,  3.L/14.L,  4.L/14.L,
    1.L/14.L,  6.L/14.L,  2.L/14.L,  5.L/14.L,
    1.L/14.L,  6.L/14.L,  1.L/14.L,  6.L/14.L,
    1.L/14.L,  5.L/14.L,  7.L/14.L,  1.L/14.L,
    1.L/14.L,  5.L/14.L,  6.L/14.L,  2.L/14.L,
    1.L/14.L,  5.L/14.L,  5.L/14.L,  3.L/14.L,
    1.L/14.L,  5.L/14.L,  4.L/14.L,  4.L/14.L,
    1.L/14.L,  5.L/14.L,  3.L/14.L,  5.L/14.L,
    1.L/14.L,  5.L/14.L,  2.L/14.L,  6.L/14.L,
    1.L/14.L,  5.L/14.L,  1.L/14.L,  7.L/14.L,
    1.L/14.L,  4.L/14.L,  8.L/14.L,  1.L/14.L,
    1.L/14.L,  4.L/14.L,  7.L/14.L,  2.L/14.L,
    1.L/14.L,  4.L/14.L,  6.L/14.L,  3.L/14.L,
    1.L/14.L,  4.L/14.L,  5.L/14.L,  4.L/14.L,
    1.L/14.L,  4.L/14.L,  4.L/14.L,  5.L/14.L,
    1.L/14.L,  4.L/14.L,  3.L/14.L,  6.L/14.L,
    1.L/14.L,  4.L/14.L,  2.L/14.L,  7.L/14.L,
    1.L/14.L,  4.L/14.L,  1.L/14.L,  8.L/14.L,
    1.L/14.L,  3.L/14.L,  9.L/14.L,  1.L/14.L,
    1.L/14.L,  3.L/14.L,  8.L/14.L,  2.L/14.L,
    1.L/14.L,  3.L/14.L,  7.L/14.L,  3.L/14.L,
    1.L/14.L,  3.L/14.L,  6.L/14.L,  4.L/14.L,
    1.L/14.L,  3.L/14.L,  5.L/14.L,  5.L/14.L,
    1.L/14.L,  3.L/14.L,  4.L/14.L,  6.L/14.L,
    1.L/14.L,  3.L/14.L,  3.L/14.L,  7.L/14.L,
    1.L/14.L,  3.L/14.L,  2.L/14.L,  8.L/14.L,
    1.L/14.L,  3.L/14.L,  1.L/14.L,  9.L/14.L,
    1.L/14.L,  2.L/14.L, 10.L/14.L,  1.L/14.L,
    1.L/14.L,  2.L/14.L,  9.L/14.L,  2.L/14.L,
    1.L/14.L,  2.L/14.L,  8.L/14.L,  3.L/14.L,
    1.L/14.L,  2.L/14.L,  7.L/14.L,  4.L/14.L,
    1.L/14.L,  2.L/14.L,  6.L/14.L,  5.L/14.L,
    1.L/14.L,  2.L/14.L,  5.L/14.L,  6.L/14.L,
    1.L/14.L,  2.L/14.L,  4.L/14.L,  7.L/14.L,
    1.L/14.L,  2.L/14.L,  3.L/14.L,  8.L/14.L,
    1.L/14.L,  2.L/14.L,  2.L/14.L,  9.L/14.L,
    1.L/14.L,  2.L/14.L,  1.L/14.L, 10.L/14.L,
    1.L/14.L,  1.L/14.L, 11.L/14.L,  1.L/14.L,
    1.L/14.L,  1.L/14.L, 10.L/14.L,  2.L/14.L,
    1.L/14.L,  1.L/14.L,  9.L/14.L,  3.L/14.L,
    1.L/14.L,  1.L/14.L,  8.L/14.L,  4.L/14.L,
    1.L/14.L,  1.L/14.L,  7.L/14.L,  5.L/14.L,
    1.L/14.L,  1.L/14.L,  6.L/14.L,  6.L/14.L,
    1.L/14.L,  1.L/14.L,  5.L/14.L,  7.L/14.L,
    1.L/14.L,  1.L/14.L,  4.L/14.L,  8.L/14.L,
    1.L/14.L,  1.L/14.L,  3.L/14.L,  9.L/14.L,
    1.L/14.L,  1.L/14.L,  2.L/14.L, 10.L/14.L,
    1.L/14.L,  1.L/14.L,  1.L/14.L, 11.L/14.L
};
DOF_TYPE DOF_DG14_ = {
    DofCache, "DG14", DG14_points, NULL,
    DOF_DG13, DOF_P14, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 680, 14, -1, 1,
    0, 0, 0, 680
};

static FLOAT DG15_points[] = {
   15.L/15.L,  0.L/15.L,  0.L/15.L,  0.L/15.L,
    0.L/15.L, 15.L/15.L,  0.L/15.L,  0.L/15.L,
    0.L/15.L,  0.L/15.L, 15.L/15.L,  0.L/15.L,
    0.L/15.L,  0.L/15.L,  0.L/15.L, 15.L/15.L,
   14.L/15.L,  1.L/15.L,  0.L/15.L,  0.L/15.L,
   13.L/15.L,  2.L/15.L,  0.L/15.L,  0.L/15.L,
   12.L/15.L,  3.L/15.L,  0.L/15.L,  0.L/15.L,
   11.L/15.L,  4.L/15.L,  0.L/15.L,  0.L/15.L,
   10.L/15.L,  5.L/15.L,  0.L/15.L,  0.L/15.L,
    9.L/15.L,  6.L/15.L,  0.L/15.L,  0.L/15.L,
    8.L/15.L,  7.L/15.L,  0.L/15.L,  0.L/15.L,
    7.L/15.L,  8.L/15.L,  0.L/15.L,  0.L/15.L,
    6.L/15.L,  9.L/15.L,  0.L/15.L,  0.L/15.L,
    5.L/15.L, 10.L/15.L,  0.L/15.L,  0.L/15.L,
    4.L/15.L, 11.L/15.L,  0.L/15.L,  0.L/15.L,
    3.L/15.L, 12.L/15.L,  0.L/15.L,  0.L/15.L,
    2.L/15.L, 13.L/15.L,  0.L/15.L,  0.L/15.L,
    1.L/15.L, 14.L/15.L,  0.L/15.L,  0.L/15.L,
   14.L/15.L,  0.L/15.L,  1.L/15.L,  0.L/15.L,
   13.L/15.L,  0.L/15.L,  2.L/15.L,  0.L/15.L,
   12.L/15.L,  0.L/15.L,  3.L/15.L,  0.L/15.L,
   11.L/15.L,  0.L/15.L,  4.L/15.L,  0.L/15.L,
   10.L/15.L,  0.L/15.L,  5.L/15.L,  0.L/15.L,
    9.L/15.L,  0.L/15.L,  6.L/15.L,  0.L/15.L,
    8.L/15.L,  0.L/15.L,  7.L/15.L,  0.L/15.L,
    7.L/15.L,  0.L/15.L,  8.L/15.L,  0.L/15.L,
    6.L/15.L,  0.L/15.L,  9.L/15.L,  0.L/15.L,
    5.L/15.L,  0.L/15.L, 10.L/15.L,  0.L/15.L,
    4.L/15.L,  0.L/15.L, 11.L/15.L,  0.L/15.L,
    3.L/15.L,  0.L/15.L, 12.L/15.L,  0.L/15.L,
    2.L/15.L,  0.L/15.L, 13.L/15.L,  0.L/15.L,
    1.L/15.L,  0.L/15.L, 14.L/15.L,  0.L/15.L,
   14.L/15.L,  0.L/15.L,  0.L/15.L,  1.L/15.L,
   13.L/15.L,  0.L/15.L,  0.L/15.L,  2.L/15.L,
   12.L/15.L,  0.L/15.L,  0.L/15.L,  3.L/15.L,
   11.L/15.L,  0.L/15.L,  0.L/15.L,  4.L/15.L,
   10.L/15.L,  0.L/15.L,  0.L/15.L,  5.L/15.L,
    9.L/15.L,  0.L/15.L,  0.L/15.L,  6.L/15.L,
    8.L/15.L,  0.L/15.L,  0.L/15.L,  7.L/15.L,
    7.L/15.L,  0.L/15.L,  0.L/15.L,  8.L/15.L,
    6.L/15.L,  0.L/15.L,  0.L/15.L,  9.L/15.L,
    5.L/15.L,  0.L/15.L,  0.L/15.L, 10.L/15.L,
    4.L/15.L,  0.L/15.L,  0.L/15.L, 11.L/15.L,
    3.L/15.L,  0.L/15.L,  0.L/15.L, 12.L/15.L,
    2.L/15.L,  0.L/15.L,  0.L/15.L, 13.L/15.L,
    1.L/15.L,  0.L/15.L,  0.L/15.L, 14.L/15.L,
    0.L/15.L, 14.L/15.L,  1.L/15.L,  0.L/15.L,
    0.L/15.L, 13.L/15.L,  2.L/15.L,  0.L/15.L,
    0.L/15.L, 12.L/15.L,  3.L/15.L,  0.L/15.L,
    0.L/15.L, 11.L/15.L,  4.L/15.L,  0.L/15.L,
    0.L/15.L, 10.L/15.L,  5.L/15.L,  0.L/15.L,
    0.L/15.L,  9.L/15.L,  6.L/15.L,  0.L/15.L,
    0.L/15.L,  8.L/15.L,  7.L/15.L,  0.L/15.L,
    0.L/15.L,  7.L/15.L,  8.L/15.L,  0.L/15.L,
    0.L/15.L,  6.L/15.L,  9.L/15.L,  0.L/15.L,
    0.L/15.L,  5.L/15.L, 10.L/15.L,  0.L/15.L,
    0.L/15.L,  4.L/15.L, 11.L/15.L,  0.L/15.L,
    0.L/15.L,  3.L/15.L, 12.L/15.L,  0.L/15.L,
    0.L/15.L,  2.L/15.L, 13.L/15.L,  0.L/15.L,
    0.L/15.L,  1.L/15.L, 14.L/15.L,  0.L/15.L,
    0.L/15.L, 14.L/15.L,  0.L/15.L,  1.L/15.L,
    0.L/15.L, 13.L/15.L,  0.L/15.L,  2.L/15.L,
    0.L/15.L, 12.L/15.L,  0.L/15.L,  3.L/15.L,
    0.L/15.L, 11.L/15.L,  0.L/15.L,  4.L/15.L,
    0.L/15.L, 10.L/15.L,  0.L/15.L,  5.L/15.L,
    0.L/15.L,  9.L/15.L,  0.L/15.L,  6.L/15.L,
    0.L/15.L,  8.L/15.L,  0.L/15.L,  7.L/15.L,
    0.L/15.L,  7.L/15.L,  0.L/15.L,  8.L/15.L,
    0.L/15.L,  6.L/15.L,  0.L/15.L,  9.L/15.L,
    0.L/15.L,  5.L/15.L,  0.L/15.L, 10.L/15.L,
    0.L/15.L,  4.L/15.L,  0.L/15.L, 11.L/15.L,
    0.L/15.L,  3.L/15.L,  0.L/15.L, 12.L/15.L,
    0.L/15.L,  2.L/15.L,  0.L/15.L, 13.L/15.L,
    0.L/15.L,  1.L/15.L,  0.L/15.L, 14.L/15.L,
    0.L/15.L,  0.L/15.L, 14.L/15.L,  1.L/15.L,
    0.L/15.L,  0.L/15.L, 13.L/15.L,  2.L/15.L,
    0.L/15.L,  0.L/15.L, 12.L/15.L,  3.L/15.L,
    0.L/15.L,  0.L/15.L, 11.L/15.L,  4.L/15.L,
    0.L/15.L,  0.L/15.L, 10.L/15.L,  5.L/15.L,
    0.L/15.L,  0.L/15.L,  9.L/15.L,  6.L/15.L,
    0.L/15.L,  0.L/15.L,  8.L/15.L,  7.L/15.L,
    0.L/15.L,  0.L/15.L,  7.L/15.L,  8.L/15.L,
    0.L/15.L,  0.L/15.L,  6.L/15.L,  9.L/15.L,
    0.L/15.L,  0.L/15.L,  5.L/15.L, 10.L/15.L,
    0.L/15.L,  0.L/15.L,  4.L/15.L, 11.L/15.L,
    0.L/15.L,  0.L/15.L,  3.L/15.L, 12.L/15.L,
    0.L/15.L,  0.L/15.L,  2.L/15.L, 13.L/15.L,
    0.L/15.L,  0.L/15.L,  1.L/15.L, 14.L/15.L,
    0.L/15.L, 13.L/15.L,  1.L/15.L,  1.L/15.L,
    0.L/15.L, 12.L/15.L,  2.L/15.L,  1.L/15.L,
    0.L/15.L, 12.L/15.L,  1.L/15.L,  2.L/15.L,
    0.L/15.L, 11.L/15.L,  3.L/15.L,  1.L/15.L,
    0.L/15.L, 11.L/15.L,  2.L/15.L,  2.L/15.L,
    0.L/15.L, 11.L/15.L,  1.L/15.L,  3.L/15.L,
    0.L/15.L, 10.L/15.L,  4.L/15.L,  1.L/15.L,
    0.L/15.L, 10.L/15.L,  3.L/15.L,  2.L/15.L,
    0.L/15.L, 10.L/15.L,  2.L/15.L,  3.L/15.L,
    0.L/15.L, 10.L/15.L,  1.L/15.L,  4.L/15.L,
    0.L/15.L,  9.L/15.L,  5.L/15.L,  1.L/15.L,
    0.L/15.L,  9.L/15.L,  4.L/15.L,  2.L/15.L,
    0.L/15.L,  9.L/15.L,  3.L/15.L,  3.L/15.L,
    0.L/15.L,  9.L/15.L,  2.L/15.L,  4.L/15.L,
    0.L/15.L,  9.L/15.L,  1.L/15.L,  5.L/15.L,
    0.L/15.L,  8.L/15.L,  6.L/15.L,  1.L/15.L,
    0.L/15.L,  8.L/15.L,  5.L/15.L,  2.L/15.L,
    0.L/15.L,  8.L/15.L,  4.L/15.L,  3.L/15.L,
    0.L/15.L,  8.L/15.L,  3.L/15.L,  4.L/15.L,
    0.L/15.L,  8.L/15.L,  2.L/15.L,  5.L/15.L,
    0.L/15.L,  8.L/15.L,  1.L/15.L,  6.L/15.L,
    0.L/15.L,  7.L/15.L,  7.L/15.L,  1.L/15.L,
    0.L/15.L,  7.L/15.L,  6.L/15.L,  2.L/15.L,
    0.L/15.L,  7.L/15.L,  5.L/15.L,  3.L/15.L,
    0.L/15.L,  7.L/15.L,  4.L/15.L,  4.L/15.L,
    0.L/15.L,  7.L/15.L,  3.L/15.L,  5.L/15.L,
    0.L/15.L,  7.L/15.L,  2.L/15.L,  6.L/15.L,
    0.L/15.L,  7.L/15.L,  1.L/15.L,  7.L/15.L,
    0.L/15.L,  6.L/15.L,  8.L/15.L,  1.L/15.L,
    0.L/15.L,  6.L/15.L,  7.L/15.L,  2.L/15.L,
    0.L/15.L,  6.L/15.L,  6.L/15.L,  3.L/15.L,
    0.L/15.L,  6.L/15.L,  5.L/15.L,  4.L/15.L,
    0.L/15.L,  6.L/15.L,  4.L/15.L,  5.L/15.L,
    0.L/15.L,  6.L/15.L,  3.L/15.L,  6.L/15.L,
    0.L/15.L,  6.L/15.L,  2.L/15.L,  7.L/15.L,
    0.L/15.L,  6.L/15.L,  1.L/15.L,  8.L/15.L,
    0.L/15.L,  5.L/15.L,  9.L/15.L,  1.L/15.L,
    0.L/15.L,  5.L/15.L,  8.L/15.L,  2.L/15.L,
    0.L/15.L,  5.L/15.L,  7.L/15.L,  3.L/15.L,
    0.L/15.L,  5.L/15.L,  6.L/15.L,  4.L/15.L,
    0.L/15.L,  5.L/15.L,  5.L/15.L,  5.L/15.L,
    0.L/15.L,  5.L/15.L,  4.L/15.L,  6.L/15.L,
    0.L/15.L,  5.L/15.L,  3.L/15.L,  7.L/15.L,
    0.L/15.L,  5.L/15.L,  2.L/15.L,  8.L/15.L,
    0.L/15.L,  5.L/15.L,  1.L/15.L,  9.L/15.L,
    0.L/15.L,  4.L/15.L, 10.L/15.L,  1.L/15.L,
    0.L/15.L,  4.L/15.L,  9.L/15.L,  2.L/15.L,
    0.L/15.L,  4.L/15.L,  8.L/15.L,  3.L/15.L,
    0.L/15.L,  4.L/15.L,  7.L/15.L,  4.L/15.L,
    0.L/15.L,  4.L/15.L,  6.L/15.L,  5.L/15.L,
    0.L/15.L,  4.L/15.L,  5.L/15.L,  6.L/15.L,
    0.L/15.L,  4.L/15.L,  4.L/15.L,  7.L/15.L,
    0.L/15.L,  4.L/15.L,  3.L/15.L,  8.L/15.L,
    0.L/15.L,  4.L/15.L,  2.L/15.L,  9.L/15.L,
    0.L/15.L,  4.L/15.L,  1.L/15.L, 10.L/15.L,
    0.L/15.L,  3.L/15.L, 11.L/15.L,  1.L/15.L,
    0.L/15.L,  3.L/15.L, 10.L/15.L,  2.L/15.L,
    0.L/15.L,  3.L/15.L,  9.L/15.L,  3.L/15.L,
    0.L/15.L,  3.L/15.L,  8.L/15.L,  4.L/15.L,
    0.L/15.L,  3.L/15.L,  7.L/15.L,  5.L/15.L,
    0.L/15.L,  3.L/15.L,  6.L/15.L,  6.L/15.L,
    0.L/15.L,  3.L/15.L,  5.L/15.L,  7.L/15.L,
    0.L/15.L,  3.L/15.L,  4.L/15.L,  8.L/15.L,
    0.L/15.L,  3.L/15.L,  3.L/15.L,  9.L/15.L,
    0.L/15.L,  3.L/15.L,  2.L/15.L, 10.L/15.L,
    0.L/15.L,  3.L/15.L,  1.L/15.L, 11.L/15.L,
    0.L/15.L,  2.L/15.L, 12.L/15.L,  1.L/15.L,
    0.L/15.L,  2.L/15.L, 11.L/15.L,  2.L/15.L,
    0.L/15.L,  2.L/15.L, 10.L/15.L,  3.L/15.L,
    0.L/15.L,  2.L/15.L,  9.L/15.L,  4.L/15.L,
    0.L/15.L,  2.L/15.L,  8.L/15.L,  5.L/15.L,
    0.L/15.L,  2.L/15.L,  7.L/15.L,  6.L/15.L,
    0.L/15.L,  2.L/15.L,  6.L/15.L,  7.L/15.L,
    0.L/15.L,  2.L/15.L,  5.L/15.L,  8.L/15.L,
    0.L/15.L,  2.L/15.L,  4.L/15.L,  9.L/15.L,
    0.L/15.L,  2.L/15.L,  3.L/15.L, 10.L/15.L,
    0.L/15.L,  2.L/15.L,  2.L/15.L, 11.L/15.L,
    0.L/15.L,  2.L/15.L,  1.L/15.L, 12.L/15.L,
    0.L/15.L,  1.L/15.L, 13.L/15.L,  1.L/15.L,
    0.L/15.L,  1.L/15.L, 12.L/15.L,  2.L/15.L,
    0.L/15.L,  1.L/15.L, 11.L/15.L,  3.L/15.L,
    0.L/15.L,  1.L/15.L, 10.L/15.L,  4.L/15.L,
    0.L/15.L,  1.L/15.L,  9.L/15.L,  5.L/15.L,
    0.L/15.L,  1.L/15.L,  8.L/15.L,  6.L/15.L,
    0.L/15.L,  1.L/15.L,  7.L/15.L,  7.L/15.L,
    0.L/15.L,  1.L/15.L,  6.L/15.L,  8.L/15.L,
    0.L/15.L,  1.L/15.L,  5.L/15.L,  9.L/15.L,
    0.L/15.L,  1.L/15.L,  4.L/15.L, 10.L/15.L,
    0.L/15.L,  1.L/15.L,  3.L/15.L, 11.L/15.L,
    0.L/15.L,  1.L/15.L,  2.L/15.L, 12.L/15.L,
    0.L/15.L,  1.L/15.L,  1.L/15.L, 13.L/15.L,
   13.L/15.L,  0.L/15.L,  1.L/15.L,  1.L/15.L,
   12.L/15.L,  0.L/15.L,  2.L/15.L,  1.L/15.L,
   12.L/15.L,  0.L/15.L,  1.L/15.L,  2.L/15.L,
   11.L/15.L,  0.L/15.L,  3.L/15.L,  1.L/15.L,
   11.L/15.L,  0.L/15.L,  2.L/15.L,  2.L/15.L,
   11.L/15.L,  0.L/15.L,  1.L/15.L,  3.L/15.L,
   10.L/15.L,  0.L/15.L,  4.L/15.L,  1.L/15.L,
   10.L/15.L,  0.L/15.L,  3.L/15.L,  2.L/15.L,
   10.L/15.L,  0.L/15.L,  2.L/15.L,  3.L/15.L,
   10.L/15.L,  0.L/15.L,  1.L/15.L,  4.L/15.L,
    9.L/15.L,  0.L/15.L,  5.L/15.L,  1.L/15.L,
    9.L/15.L,  0.L/15.L,  4.L/15.L,  2.L/15.L,
    9.L/15.L,  0.L/15.L,  3.L/15.L,  3.L/15.L,
    9.L/15.L,  0.L/15.L,  2.L/15.L,  4.L/15.L,
    9.L/15.L,  0.L/15.L,  1.L/15.L,  5.L/15.L,
    8.L/15.L,  0.L/15.L,  6.L/15.L,  1.L/15.L,
    8.L/15.L,  0.L/15.L,  5.L/15.L,  2.L/15.L,
    8.L/15.L,  0.L/15.L,  4.L/15.L,  3.L/15.L,
    8.L/15.L,  0.L/15.L,  3.L/15.L,  4.L/15.L,
    8.L/15.L,  0.L/15.L,  2.L/15.L,  5.L/15.L,
    8.L/15.L,  0.L/15.L,  1.L/15.L,  6.L/15.L,
    7.L/15.L,  0.L/15.L,  7.L/15.L,  1.L/15.L,
    7.L/15.L,  0.L/15.L,  6.L/15.L,  2.L/15.L,
    7.L/15.L,  0.L/15.L,  5.L/15.L,  3.L/15.L,
    7.L/15.L,  0.L/15.L,  4.L/15.L,  4.L/15.L,
    7.L/15.L,  0.L/15.L,  3.L/15.L,  5.L/15.L,
    7.L/15.L,  0.L/15.L,  2.L/15.L,  6.L/15.L,
    7.L/15.L,  0.L/15.L,  1.L/15.L,  7.L/15.L,
    6.L/15.L,  0.L/15.L,  8.L/15.L,  1.L/15.L,
    6.L/15.L,  0.L/15.L,  7.L/15.L,  2.L/15.L,
    6.L/15.L,  0.L/15.L,  6.L/15.L,  3.L/15.L,
    6.L/15.L,  0.L/15.L,  5.L/15.L,  4.L/15.L,
    6.L/15.L,  0.L/15.L,  4.L/15.L,  5.L/15.L,
    6.L/15.L,  0.L/15.L,  3.L/15.L,  6.L/15.L,
    6.L/15.L,  0.L/15.L,  2.L/15.L,  7.L/15.L,
    6.L/15.L,  0.L/15.L,  1.L/15.L,  8.L/15.L,
    5.L/15.L,  0.L/15.L,  9.L/15.L,  1.L/15.L,
    5.L/15.L,  0.L/15.L,  8.L/15.L,  2.L/15.L,
    5.L/15.L,  0.L/15.L,  7.L/15.L,  3.L/15.L,
    5.L/15.L,  0.L/15.L,  6.L/15.L,  4.L/15.L,
    5.L/15.L,  0.L/15.L,  5.L/15.L,  5.L/15.L,
    5.L/15.L,  0.L/15.L,  4.L/15.L,  6.L/15.L,
    5.L/15.L,  0.L/15.L,  3.L/15.L,  7.L/15.L,
    5.L/15.L,  0.L/15.L,  2.L/15.L,  8.L/15.L,
    5.L/15.L,  0.L/15.L,  1.L/15.L,  9.L/15.L,
    4.L/15.L,  0.L/15.L, 10.L/15.L,  1.L/15.L,
    4.L/15.L,  0.L/15.L,  9.L/15.L,  2.L/15.L,
    4.L/15.L,  0.L/15.L,  8.L/15.L,  3.L/15.L,
    4.L/15.L,  0.L/15.L,  7.L/15.L,  4.L/15.L,
    4.L/15.L,  0.L/15.L,  6.L/15.L,  5.L/15.L,
    4.L/15.L,  0.L/15.L,  5.L/15.L,  6.L/15.L,
    4.L/15.L,  0.L/15.L,  4.L/15.L,  7.L/15.L,
    4.L/15.L,  0.L/15.L,  3.L/15.L,  8.L/15.L,
    4.L/15.L,  0.L/15.L,  2.L/15.L,  9.L/15.L,
    4.L/15.L,  0.L/15.L,  1.L/15.L, 10.L/15.L,
    3.L/15.L,  0.L/15.L, 11.L/15.L,  1.L/15.L,
    3.L/15.L,  0.L/15.L, 10.L/15.L,  2.L/15.L,
    3.L/15.L,  0.L/15.L,  9.L/15.L,  3.L/15.L,
    3.L/15.L,  0.L/15.L,  8.L/15.L,  4.L/15.L,
    3.L/15.L,  0.L/15.L,  7.L/15.L,  5.L/15.L,
    3.L/15.L,  0.L/15.L,  6.L/15.L,  6.L/15.L,
    3.L/15.L,  0.L/15.L,  5.L/15.L,  7.L/15.L,
    3.L/15.L,  0.L/15.L,  4.L/15.L,  8.L/15.L,
    3.L/15.L,  0.L/15.L,  3.L/15.L,  9.L/15.L,
    3.L/15.L,  0.L/15.L,  2.L/15.L, 10.L/15.L,
    3.L/15.L,  0.L/15.L,  1.L/15.L, 11.L/15.L,
    2.L/15.L,  0.L/15.L, 12.L/15.L,  1.L/15.L,
    2.L/15.L,  0.L/15.L, 11.L/15.L,  2.L/15.L,
    2.L/15.L,  0.L/15.L, 10.L/15.L,  3.L/15.L,
    2.L/15.L,  0.L/15.L,  9.L/15.L,  4.L/15.L,
    2.L/15.L,  0.L/15.L,  8.L/15.L,  5.L/15.L,
    2.L/15.L,  0.L/15.L,  7.L/15.L,  6.L/15.L,
    2.L/15.L,  0.L/15.L,  6.L/15.L,  7.L/15.L,
    2.L/15.L,  0.L/15.L,  5.L/15.L,  8.L/15.L,
    2.L/15.L,  0.L/15.L,  4.L/15.L,  9.L/15.L,
    2.L/15.L,  0.L/15.L,  3.L/15.L, 10.L/15.L,
    2.L/15.L,  0.L/15.L,  2.L/15.L, 11.L/15.L,
    2.L/15.L,  0.L/15.L,  1.L/15.L, 12.L/15.L,
    1.L/15.L,  0.L/15.L, 13.L/15.L,  1.L/15.L,
    1.L/15.L,  0.L/15.L, 12.L/15.L,  2.L/15.L,
    1.L/15.L,  0.L/15.L, 11.L/15.L,  3.L/15.L,
    1.L/15.L,  0.L/15.L, 10.L/15.L,  4.L/15.L,
    1.L/15.L,  0.L/15.L,  9.L/15.L,  5.L/15.L,
    1.L/15.L,  0.L/15.L,  8.L/15.L,  6.L/15.L,
    1.L/15.L,  0.L/15.L,  7.L/15.L,  7.L/15.L,
    1.L/15.L,  0.L/15.L,  6.L/15.L,  8.L/15.L,
    1.L/15.L,  0.L/15.L,  5.L/15.L,  9.L/15.L,
    1.L/15.L,  0.L/15.L,  4.L/15.L, 10.L/15.L,
    1.L/15.L,  0.L/15.L,  3.L/15.L, 11.L/15.L,
    1.L/15.L,  0.L/15.L,  2.L/15.L, 12.L/15.L,
    1.L/15.L,  0.L/15.L,  1.L/15.L, 13.L/15.L,
   13.L/15.L,  1.L/15.L,  0.L/15.L,  1.L/15.L,
   12.L/15.L,  2.L/15.L,  0.L/15.L,  1.L/15.L,
   12.L/15.L,  1.L/15.L,  0.L/15.L,  2.L/15.L,
   11.L/15.L,  3.L/15.L,  0.L/15.L,  1.L/15.L,
   11.L/15.L,  2.L/15.L,  0.L/15.L,  2.L/15.L,
   11.L/15.L,  1.L/15.L,  0.L/15.L,  3.L/15.L,
   10.L/15.L,  4.L/15.L,  0.L/15.L,  1.L/15.L,
   10.L/15.L,  3.L/15.L,  0.L/15.L,  2.L/15.L,
   10.L/15.L,  2.L/15.L,  0.L/15.L,  3.L/15.L,
   10.L/15.L,  1.L/15.L,  0.L/15.L,  4.L/15.L,
    9.L/15.L,  5.L/15.L,  0.L/15.L,  1.L/15.L,
    9.L/15.L,  4.L/15.L,  0.L/15.L,  2.L/15.L,
    9.L/15.L,  3.L/15.L,  0.L/15.L,  3.L/15.L,
    9.L/15.L,  2.L/15.L,  0.L/15.L,  4.L/15.L,
    9.L/15.L,  1.L/15.L,  0.L/15.L,  5.L/15.L,
    8.L/15.L,  6.L/15.L,  0.L/15.L,  1.L/15.L,
    8.L/15.L,  5.L/15.L,  0.L/15.L,  2.L/15.L,
    8.L/15.L,  4.L/15.L,  0.L/15.L,  3.L/15.L,
    8.L/15.L,  3.L/15.L,  0.L/15.L,  4.L/15.L,
    8.L/15.L,  2.L/15.L,  0.L/15.L,  5.L/15.L,
    8.L/15.L,  1.L/15.L,  0.L/15.L,  6.L/15.L,
    7.L/15.L,  7.L/15.L,  0.L/15.L,  1.L/15.L,
    7.L/15.L,  6.L/15.L,  0.L/15.L,  2.L/15.L,
    7.L/15.L,  5.L/15.L,  0.L/15.L,  3.L/15.L,
    7.L/15.L,  4.L/15.L,  0.L/15.L,  4.L/15.L,
    7.L/15.L,  3.L/15.L,  0.L/15.L,  5.L/15.L,
    7.L/15.L,  2.L/15.L,  0.L/15.L,  6.L/15.L,
    7.L/15.L,  1.L/15.L,  0.L/15.L,  7.L/15.L,
    6.L/15.L,  8.L/15.L,  0.L/15.L,  1.L/15.L,
    6.L/15.L,  7.L/15.L,  0.L/15.L,  2.L/15.L,
    6.L/15.L,  6.L/15.L,  0.L/15.L,  3.L/15.L,
    6.L/15.L,  5.L/15.L,  0.L/15.L,  4.L/15.L,
    6.L/15.L,  4.L/15.L,  0.L/15.L,  5.L/15.L,
    6.L/15.L,  3.L/15.L,  0.L/15.L,  6.L/15.L,
    6.L/15.L,  2.L/15.L,  0.L/15.L,  7.L/15.L,
    6.L/15.L,  1.L/15.L,  0.L/15.L,  8.L/15.L,
    5.L/15.L,  9.L/15.L,  0.L/15.L,  1.L/15.L,
    5.L/15.L,  8.L/15.L,  0.L/15.L,  2.L/15.L,
    5.L/15.L,  7.L/15.L,  0.L/15.L,  3.L/15.L,
    5.L/15.L,  6.L/15.L,  0.L/15.L,  4.L/15.L,
    5.L/15.L,  5.L/15.L,  0.L/15.L,  5.L/15.L,
    5.L/15.L,  4.L/15.L,  0.L/15.L,  6.L/15.L,
    5.L/15.L,  3.L/15.L,  0.L/15.L,  7.L/15.L,
    5.L/15.L,  2.L/15.L,  0.L/15.L,  8.L/15.L,
    5.L/15.L,  1.L/15.L,  0.L/15.L,  9.L/15.L,
    4.L/15.L, 10.L/15.L,  0.L/15.L,  1.L/15.L,
    4.L/15.L,  9.L/15.L,  0.L/15.L,  2.L/15.L,
    4.L/15.L,  8.L/15.L,  0.L/15.L,  3.L/15.L,
    4.L/15.L,  7.L/15.L,  0.L/15.L,  4.L/15.L,
    4.L/15.L,  6.L/15.L,  0.L/15.L,  5.L/15.L,
    4.L/15.L,  5.L/15.L,  0.L/15.L,  6.L/15.L,
    4.L/15.L,  4.L/15.L,  0.L/15.L,  7.L/15.L,
    4.L/15.L,  3.L/15.L,  0.L/15.L,  8.L/15.L,
    4.L/15.L,  2.L/15.L,  0.L/15.L,  9.L/15.L,
    4.L/15.L,  1.L/15.L,  0.L/15.L, 10.L/15.L,
    3.L/15.L, 11.L/15.L,  0.L/15.L,  1.L/15.L,
    3.L/15.L, 10.L/15.L,  0.L/15.L,  2.L/15.L,
    3.L/15.L,  9.L/15.L,  0.L/15.L,  3.L/15.L,
    3.L/15.L,  8.L/15.L,  0.L/15.L,  4.L/15.L,
    3.L/15.L,  7.L/15.L,  0.L/15.L,  5.L/15.L,
    3.L/15.L,  6.L/15.L,  0.L/15.L,  6.L/15.L,
    3.L/15.L,  5.L/15.L,  0.L/15.L,  7.L/15.L,
    3.L/15.L,  4.L/15.L,  0.L/15.L,  8.L/15.L,
    3.L/15.L,  3.L/15.L,  0.L/15.L,  9.L/15.L,
    3.L/15.L,  2.L/15.L,  0.L/15.L, 10.L/15.L,
    3.L/15.L,  1.L/15.L,  0.L/15.L, 11.L/15.L,
    2.L/15.L, 12.L/15.L,  0.L/15.L,  1.L/15.L,
    2.L/15.L, 11.L/15.L,  0.L/15.L,  2.L/15.L,
    2.L/15.L, 10.L/15.L,  0.L/15.L,  3.L/15.L,
    2.L/15.L,  9.L/15.L,  0.L/15.L,  4.L/15.L,
    2.L/15.L,  8.L/15.L,  0.L/15.L,  5.L/15.L,
    2.L/15.L,  7.L/15.L,  0.L/15.L,  6.L/15.L,
    2.L/15.L,  6.L/15.L,  0.L/15.L,  7.L/15.L,
    2.L/15.L,  5.L/15.L,  0.L/15.L,  8.L/15.L,
    2.L/15.L,  4.L/15.L,  0.L/15.L,  9.L/15.L,
    2.L/15.L,  3.L/15.L,  0.L/15.L, 10.L/15.L,
    2.L/15.L,  2.L/15.L,  0.L/15.L, 11.L/15.L,
    2.L/15.L,  1.L/15.L,  0.L/15.L, 12.L/15.L,
    1.L/15.L, 13.L/15.L,  0.L/15.L,  1.L/15.L,
    1.L/15.L, 12.L/15.L,  0.L/15.L,  2.L/15.L,
    1.L/15.L, 11.L/15.L,  0.L/15.L,  3.L/15.L,
    1.L/15.L, 10.L/15.L,  0.L/15.L,  4.L/15.L,
    1.L/15.L,  9.L/15.L,  0.L/15.L,  5.L/15.L,
    1.L/15.L,  8.L/15.L,  0.L/15.L,  6.L/15.L,
    1.L/15.L,  7.L/15.L,  0.L/15.L,  7.L/15.L,
    1.L/15.L,  6.L/15.L,  0.L/15.L,  8.L/15.L,
    1.L/15.L,  5.L/15.L,  0.L/15.L,  9.L/15.L,
    1.L/15.L,  4.L/15.L,  0.L/15.L, 10.L/15.L,
    1.L/15.L,  3.L/15.L,  0.L/15.L, 11.L/15.L,
    1.L/15.L,  2.L/15.L,  0.L/15.L, 12.L/15.L,
    1.L/15.L,  1.L/15.L,  0.L/15.L, 13.L/15.L,
   13.L/15.L,  1.L/15.L,  1.L/15.L,  0.L/15.L,
   12.L/15.L,  2.L/15.L,  1.L/15.L,  0.L/15.L,
   12.L/15.L,  1.L/15.L,  2.L/15.L,  0.L/15.L,
   11.L/15.L,  3.L/15.L,  1.L/15.L,  0.L/15.L,
   11.L/15.L,  2.L/15.L,  2.L/15.L,  0.L/15.L,
   11.L/15.L,  1.L/15.L,  3.L/15.L,  0.L/15.L,
   10.L/15.L,  4.L/15.L,  1.L/15.L,  0.L/15.L,
   10.L/15.L,  3.L/15.L,  2.L/15.L,  0.L/15.L,
   10.L/15.L,  2.L/15.L,  3.L/15.L,  0.L/15.L,
   10.L/15.L,  1.L/15.L,  4.L/15.L,  0.L/15.L,
    9.L/15.L,  5.L/15.L,  1.L/15.L,  0.L/15.L,
    9.L/15.L,  4.L/15.L,  2.L/15.L,  0.L/15.L,
    9.L/15.L,  3.L/15.L,  3.L/15.L,  0.L/15.L,
    9.L/15.L,  2.L/15.L,  4.L/15.L,  0.L/15.L,
    9.L/15.L,  1.L/15.L,  5.L/15.L,  0.L/15.L,
    8.L/15.L,  6.L/15.L,  1.L/15.L,  0.L/15.L,
    8.L/15.L,  5.L/15.L,  2.L/15.L,  0.L/15.L,
    8.L/15.L,  4.L/15.L,  3.L/15.L,  0.L/15.L,
    8.L/15.L,  3.L/15.L,  4.L/15.L,  0.L/15.L,
    8.L/15.L,  2.L/15.L,  5.L/15.L,  0.L/15.L,
    8.L/15.L,  1.L/15.L,  6.L/15.L,  0.L/15.L,
    7.L/15.L,  7.L/15.L,  1.L/15.L,  0.L/15.L,
    7.L/15.L,  6.L/15.L,  2.L/15.L,  0.L/15.L,
    7.L/15.L,  5.L/15.L,  3.L/15.L,  0.L/15.L,
    7.L/15.L,  4.L/15.L,  4.L/15.L,  0.L/15.L,
    7.L/15.L,  3.L/15.L,  5.L/15.L,  0.L/15.L,
    7.L/15.L,  2.L/15.L,  6.L/15.L,  0.L/15.L,
    7.L/15.L,  1.L/15.L,  7.L/15.L,  0.L/15.L,
    6.L/15.L,  8.L/15.L,  1.L/15.L,  0.L/15.L,
    6.L/15.L,  7.L/15.L,  2.L/15.L,  0.L/15.L,
    6.L/15.L,  6.L/15.L,  3.L/15.L,  0.L/15.L,
    6.L/15.L,  5.L/15.L,  4.L/15.L,  0.L/15.L,
    6.L/15.L,  4.L/15.L,  5.L/15.L,  0.L/15.L,
    6.L/15.L,  3.L/15.L,  6.L/15.L,  0.L/15.L,
    6.L/15.L,  2.L/15.L,  7.L/15.L,  0.L/15.L,
    6.L/15.L,  1.L/15.L,  8.L/15.L,  0.L/15.L,
    5.L/15.L,  9.L/15.L,  1.L/15.L,  0.L/15.L,
    5.L/15.L,  8.L/15.L,  2.L/15.L,  0.L/15.L,
    5.L/15.L,  7.L/15.L,  3.L/15.L,  0.L/15.L,
    5.L/15.L,  6.L/15.L,  4.L/15.L,  0.L/15.L,
    5.L/15.L,  5.L/15.L,  5.L/15.L,  0.L/15.L,
    5.L/15.L,  4.L/15.L,  6.L/15.L,  0.L/15.L,
    5.L/15.L,  3.L/15.L,  7.L/15.L,  0.L/15.L,
    5.L/15.L,  2.L/15.L,  8.L/15.L,  0.L/15.L,
    5.L/15.L,  1.L/15.L,  9.L/15.L,  0.L/15.L,
    4.L/15.L, 10.L/15.L,  1.L/15.L,  0.L/15.L,
    4.L/15.L,  9.L/15.L,  2.L/15.L,  0.L/15.L,
    4.L/15.L,  8.L/15.L,  3.L/15.L,  0.L/15.L,
    4.L/15.L,  7.L/15.L,  4.L/15.L,  0.L/15.L,
    4.L/15.L,  6.L/15.L,  5.L/15.L,  0.L/15.L,
    4.L/15.L,  5.L/15.L,  6.L/15.L,  0.L/15.L,
    4.L/15.L,  4.L/15.L,  7.L/15.L,  0.L/15.L,
    4.L/15.L,  3.L/15.L,  8.L/15.L,  0.L/15.L,
    4.L/15.L,  2.L/15.L,  9.L/15.L,  0.L/15.L,
    4.L/15.L,  1.L/15.L, 10.L/15.L,  0.L/15.L,
    3.L/15.L, 11.L/15.L,  1.L/15.L,  0.L/15.L,
    3.L/15.L, 10.L/15.L,  2.L/15.L,  0.L/15.L,
    3.L/15.L,  9.L/15.L,  3.L/15.L,  0.L/15.L,
    3.L/15.L,  8.L/15.L,  4.L/15.L,  0.L/15.L,
    3.L/15.L,  7.L/15.L,  5.L/15.L,  0.L/15.L,
    3.L/15.L,  6.L/15.L,  6.L/15.L,  0.L/15.L,
    3.L/15.L,  5.L/15.L,  7.L/15.L,  0.L/15.L,
    3.L/15.L,  4.L/15.L,  8.L/15.L,  0.L/15.L,
    3.L/15.L,  3.L/15.L,  9.L/15.L,  0.L/15.L,
    3.L/15.L,  2.L/15.L, 10.L/15.L,  0.L/15.L,
    3.L/15.L,  1.L/15.L, 11.L/15.L,  0.L/15.L,
    2.L/15.L, 12.L/15.L,  1.L/15.L,  0.L/15.L,
    2.L/15.L, 11.L/15.L,  2.L/15.L,  0.L/15.L,
    2.L/15.L, 10.L/15.L,  3.L/15.L,  0.L/15.L,
    2.L/15.L,  9.L/15.L,  4.L/15.L,  0.L/15.L,
    2.L/15.L,  8.L/15.L,  5.L/15.L,  0.L/15.L,
    2.L/15.L,  7.L/15.L,  6.L/15.L,  0.L/15.L,
    2.L/15.L,  6.L/15.L,  7.L/15.L,  0.L/15.L,
    2.L/15.L,  5.L/15.L,  8.L/15.L,  0.L/15.L,
    2.L/15.L,  4.L/15.L,  9.L/15.L,  0.L/15.L,
    2.L/15.L,  3.L/15.L, 10.L/15.L,  0.L/15.L,
    2.L/15.L,  2.L/15.L, 11.L/15.L,  0.L/15.L,
    2.L/15.L,  1.L/15.L, 12.L/15.L,  0.L/15.L,
    1.L/15.L, 13.L/15.L,  1.L/15.L,  0.L/15.L,
    1.L/15.L, 12.L/15.L,  2.L/15.L,  0.L/15.L,
    1.L/15.L, 11.L/15.L,  3.L/15.L,  0.L/15.L,
    1.L/15.L, 10.L/15.L,  4.L/15.L,  0.L/15.L,
    1.L/15.L,  9.L/15.L,  5.L/15.L,  0.L/15.L,
    1.L/15.L,  8.L/15.L,  6.L/15.L,  0.L/15.L,
    1.L/15.L,  7.L/15.L,  7.L/15.L,  0.L/15.L,
    1.L/15.L,  6.L/15.L,  8.L/15.L,  0.L/15.L,
    1.L/15.L,  5.L/15.L,  9.L/15.L,  0.L/15.L,
    1.L/15.L,  4.L/15.L, 10.L/15.L,  0.L/15.L,
    1.L/15.L,  3.L/15.L, 11.L/15.L,  0.L/15.L,
    1.L/15.L,  2.L/15.L, 12.L/15.L,  0.L/15.L,
    1.L/15.L,  1.L/15.L, 13.L/15.L,  0.L/15.L,
   12.L/15.L,  1.L/15.L,  1.L/15.L,  1.L/15.L,
   11.L/15.L,  2.L/15.L,  1.L/15.L,  1.L/15.L,
   11.L/15.L,  1.L/15.L,  2.L/15.L,  1.L/15.L,
   11.L/15.L,  1.L/15.L,  1.L/15.L,  2.L/15.L,
   10.L/15.L,  3.L/15.L,  1.L/15.L,  1.L/15.L,
   10.L/15.L,  2.L/15.L,  2.L/15.L,  1.L/15.L,
   10.L/15.L,  2.L/15.L,  1.L/15.L,  2.L/15.L,
   10.L/15.L,  1.L/15.L,  3.L/15.L,  1.L/15.L,
   10.L/15.L,  1.L/15.L,  2.L/15.L,  2.L/15.L,
   10.L/15.L,  1.L/15.L,  1.L/15.L,  3.L/15.L,
    9.L/15.L,  4.L/15.L,  1.L/15.L,  1.L/15.L,
    9.L/15.L,  3.L/15.L,  2.L/15.L,  1.L/15.L,
    9.L/15.L,  3.L/15.L,  1.L/15.L,  2.L/15.L,
    9.L/15.L,  2.L/15.L,  3.L/15.L,  1.L/15.L,
    9.L/15.L,  2.L/15.L,  2.L/15.L,  2.L/15.L,
    9.L/15.L,  2.L/15.L,  1.L/15.L,  3.L/15.L,
    9.L/15.L,  1.L/15.L,  4.L/15.L,  1.L/15.L,
    9.L/15.L,  1.L/15.L,  3.L/15.L,  2.L/15.L,
    9.L/15.L,  1.L/15.L,  2.L/15.L,  3.L/15.L,
    9.L/15.L,  1.L/15.L,  1.L/15.L,  4.L/15.L,
    8.L/15.L,  5.L/15.L,  1.L/15.L,  1.L/15.L,
    8.L/15.L,  4.L/15.L,  2.L/15.L,  1.L/15.L,
    8.L/15.L,  4.L/15.L,  1.L/15.L,  2.L/15.L,
    8.L/15.L,  3.L/15.L,  3.L/15.L,  1.L/15.L,
    8.L/15.L,  3.L/15.L,  2.L/15.L,  2.L/15.L,
    8.L/15.L,  3.L/15.L,  1.L/15.L,  3.L/15.L,
    8.L/15.L,  2.L/15.L,  4.L/15.L,  1.L/15.L,
    8.L/15.L,  2.L/15.L,  3.L/15.L,  2.L/15.L,
    8.L/15.L,  2.L/15.L,  2.L/15.L,  3.L/15.L,
    8.L/15.L,  2.L/15.L,  1.L/15.L,  4.L/15.L,
    8.L/15.L,  1.L/15.L,  5.L/15.L,  1.L/15.L,
    8.L/15.L,  1.L/15.L,  4.L/15.L,  2.L/15.L,
    8.L/15.L,  1.L/15.L,  3.L/15.L,  3.L/15.L,
    8.L/15.L,  1.L/15.L,  2.L/15.L,  4.L/15.L,
    8.L/15.L,  1.L/15.L,  1.L/15.L,  5.L/15.L,
    7.L/15.L,  6.L/15.L,  1.L/15.L,  1.L/15.L,
    7.L/15.L,  5.L/15.L,  2.L/15.L,  1.L/15.L,
    7.L/15.L,  5.L/15.L,  1.L/15.L,  2.L/15.L,
    7.L/15.L,  4.L/15.L,  3.L/15.L,  1.L/15.L,
    7.L/15.L,  4.L/15.L,  2.L/15.L,  2.L/15.L,
    7.L/15.L,  4.L/15.L,  1.L/15.L,  3.L/15.L,
    7.L/15.L,  3.L/15.L,  4.L/15.L,  1.L/15.L,
    7.L/15.L,  3.L/15.L,  3.L/15.L,  2.L/15.L,
    7.L/15.L,  3.L/15.L,  2.L/15.L,  3.L/15.L,
    7.L/15.L,  3.L/15.L,  1.L/15.L,  4.L/15.L,
    7.L/15.L,  2.L/15.L,  5.L/15.L,  1.L/15.L,
    7.L/15.L,  2.L/15.L,  4.L/15.L,  2.L/15.L,
    7.L/15.L,  2.L/15.L,  3.L/15.L,  3.L/15.L,
    7.L/15.L,  2.L/15.L,  2.L/15.L,  4.L/15.L,
    7.L/15.L,  2.L/15.L,  1.L/15.L,  5.L/15.L,
    7.L/15.L,  1.L/15.L,  6.L/15.L,  1.L/15.L,
    7.L/15.L,  1.L/15.L,  5.L/15.L,  2.L/15.L,
    7.L/15.L,  1.L/15.L,  4.L/15.L,  3.L/15.L,
    7.L/15.L,  1.L/15.L,  3.L/15.L,  4.L/15.L,
    7.L/15.L,  1.L/15.L,  2.L/15.L,  5.L/15.L,
    7.L/15.L,  1.L/15.L,  1.L/15.L,  6.L/15.L,
    6.L/15.L,  7.L/15.L,  1.L/15.L,  1.L/15.L,
    6.L/15.L,  6.L/15.L,  2.L/15.L,  1.L/15.L,
    6.L/15.L,  6.L/15.L,  1.L/15.L,  2.L/15.L,
    6.L/15.L,  5.L/15.L,  3.L/15.L,  1.L/15.L,
    6.L/15.L,  5.L/15.L,  2.L/15.L,  2.L/15.L,
    6.L/15.L,  5.L/15.L,  1.L/15.L,  3.L/15.L,
    6.L/15.L,  4.L/15.L,  4.L/15.L,  1.L/15.L,
    6.L/15.L,  4.L/15.L,  3.L/15.L,  2.L/15.L,
    6.L/15.L,  4.L/15.L,  2.L/15.L,  3.L/15.L,
    6.L/15.L,  4.L/15.L,  1.L/15.L,  4.L/15.L,
    6.L/15.L,  3.L/15.L,  5.L/15.L,  1.L/15.L,
    6.L/15.L,  3.L/15.L,  4.L/15.L,  2.L/15.L,
    6.L/15.L,  3.L/15.L,  3.L/15.L,  3.L/15.L,
    6.L/15.L,  3.L/15.L,  2.L/15.L,  4.L/15.L,
    6.L/15.L,  3.L/15.L,  1.L/15.L,  5.L/15.L,
    6.L/15.L,  2.L/15.L,  6.L/15.L,  1.L/15.L,
    6.L/15.L,  2.L/15.L,  5.L/15.L,  2.L/15.L,
    6.L/15.L,  2.L/15.L,  4.L/15.L,  3.L/15.L,
    6.L/15.L,  2.L/15.L,  3.L/15.L,  4.L/15.L,
    6.L/15.L,  2.L/15.L,  2.L/15.L,  5.L/15.L,
    6.L/15.L,  2.L/15.L,  1.L/15.L,  6.L/15.L,
    6.L/15.L,  1.L/15.L,  7.L/15.L,  1.L/15.L,
    6.L/15.L,  1.L/15.L,  6.L/15.L,  2.L/15.L,
    6.L/15.L,  1.L/15.L,  5.L/15.L,  3.L/15.L,
    6.L/15.L,  1.L/15.L,  4.L/15.L,  4.L/15.L,
    6.L/15.L,  1.L/15.L,  3.L/15.L,  5.L/15.L,
    6.L/15.L,  1.L/15.L,  2.L/15.L,  6.L/15.L,
    6.L/15.L,  1.L/15.L,  1.L/15.L,  7.L/15.L,
    5.L/15.L,  8.L/15.L,  1.L/15.L,  1.L/15.L,
    5.L/15.L,  7.L/15.L,  2.L/15.L,  1.L/15.L,
    5.L/15.L,  7.L/15.L,  1.L/15.L,  2.L/15.L,
    5.L/15.L,  6.L/15.L,  3.L/15.L,  1.L/15.L,
    5.L/15.L,  6.L/15.L,  2.L/15.L,  2.L/15.L,
    5.L/15.L,  6.L/15.L,  1.L/15.L,  3.L/15.L,
    5.L/15.L,  5.L/15.L,  4.L/15.L,  1.L/15.L,
    5.L/15.L,  5.L/15.L,  3.L/15.L,  2.L/15.L,
    5.L/15.L,  5.L/15.L,  2.L/15.L,  3.L/15.L,
    5.L/15.L,  5.L/15.L,  1.L/15.L,  4.L/15.L,
    5.L/15.L,  4.L/15.L,  5.L/15.L,  1.L/15.L,
    5.L/15.L,  4.L/15.L,  4.L/15.L,  2.L/15.L,
    5.L/15.L,  4.L/15.L,  3.L/15.L,  3.L/15.L,
    5.L/15.L,  4.L/15.L,  2.L/15.L,  4.L/15.L,
    5.L/15.L,  4.L/15.L,  1.L/15.L,  5.L/15.L,
    5.L/15.L,  3.L/15.L,  6.L/15.L,  1.L/15.L,
    5.L/15.L,  3.L/15.L,  5.L/15.L,  2.L/15.L,
    5.L/15.L,  3.L/15.L,  4.L/15.L,  3.L/15.L,
    5.L/15.L,  3.L/15.L,  3.L/15.L,  4.L/15.L,
    5.L/15.L,  3.L/15.L,  2.L/15.L,  5.L/15.L,
    5.L/15.L,  3.L/15.L,  1.L/15.L,  6.L/15.L,
    5.L/15.L,  2.L/15.L,  7.L/15.L,  1.L/15.L,
    5.L/15.L,  2.L/15.L,  6.L/15.L,  2.L/15.L,
    5.L/15.L,  2.L/15.L,  5.L/15.L,  3.L/15.L,
    5.L/15.L,  2.L/15.L,  4.L/15.L,  4.L/15.L,
    5.L/15.L,  2.L/15.L,  3.L/15.L,  5.L/15.L,
    5.L/15.L,  2.L/15.L,  2.L/15.L,  6.L/15.L,
    5.L/15.L,  2.L/15.L,  1.L/15.L,  7.L/15.L,
    5.L/15.L,  1.L/15.L,  8.L/15.L,  1.L/15.L,
    5.L/15.L,  1.L/15.L,  7.L/15.L,  2.L/15.L,
    5.L/15.L,  1.L/15.L,  6.L/15.L,  3.L/15.L,
    5.L/15.L,  1.L/15.L,  5.L/15.L,  4.L/15.L,
    5.L/15.L,  1.L/15.L,  4.L/15.L,  5.L/15.L,
    5.L/15.L,  1.L/15.L,  3.L/15.L,  6.L/15.L,
    5.L/15.L,  1.L/15.L,  2.L/15.L,  7.L/15.L,
    5.L/15.L,  1.L/15.L,  1.L/15.L,  8.L/15.L,
    4.L/15.L,  9.L/15.L,  1.L/15.L,  1.L/15.L,
    4.L/15.L,  8.L/15.L,  2.L/15.L,  1.L/15.L,
    4.L/15.L,  8.L/15.L,  1.L/15.L,  2.L/15.L,
    4.L/15.L,  7.L/15.L,  3.L/15.L,  1.L/15.L,
    4.L/15.L,  7.L/15.L,  2.L/15.L,  2.L/15.L,
    4.L/15.L,  7.L/15.L,  1.L/15.L,  3.L/15.L,
    4.L/15.L,  6.L/15.L,  4.L/15.L,  1.L/15.L,
    4.L/15.L,  6.L/15.L,  3.L/15.L,  2.L/15.L,
    4.L/15.L,  6.L/15.L,  2.L/15.L,  3.L/15.L,
    4.L/15.L,  6.L/15.L,  1.L/15.L,  4.L/15.L,
    4.L/15.L,  5.L/15.L,  5.L/15.L,  1.L/15.L,
    4.L/15.L,  5.L/15.L,  4.L/15.L,  2.L/15.L,
    4.L/15.L,  5.L/15.L,  3.L/15.L,  3.L/15.L,
    4.L/15.L,  5.L/15.L,  2.L/15.L,  4.L/15.L,
    4.L/15.L,  5.L/15.L,  1.L/15.L,  5.L/15.L,
    4.L/15.L,  4.L/15.L,  6.L/15.L,  1.L/15.L,
    4.L/15.L,  4.L/15.L,  5.L/15.L,  2.L/15.L,
    4.L/15.L,  4.L/15.L,  4.L/15.L,  3.L/15.L,
    4.L/15.L,  4.L/15.L,  3.L/15.L,  4.L/15.L,
    4.L/15.L,  4.L/15.L,  2.L/15.L,  5.L/15.L,
    4.L/15.L,  4.L/15.L,  1.L/15.L,  6.L/15.L,
    4.L/15.L,  3.L/15.L,  7.L/15.L,  1.L/15.L,
    4.L/15.L,  3.L/15.L,  6.L/15.L,  2.L/15.L,
    4.L/15.L,  3.L/15.L,  5.L/15.L,  3.L/15.L,
    4.L/15.L,  3.L/15.L,  4.L/15.L,  4.L/15.L,
    4.L/15.L,  3.L/15.L,  3.L/15.L,  5.L/15.L,
    4.L/15.L,  3.L/15.L,  2.L/15.L,  6.L/15.L,
    4.L/15.L,  3.L/15.L,  1.L/15.L,  7.L/15.L,
    4.L/15.L,  2.L/15.L,  8.L/15.L,  1.L/15.L,
    4.L/15.L,  2.L/15.L,  7.L/15.L,  2.L/15.L,
    4.L/15.L,  2.L/15.L,  6.L/15.L,  3.L/15.L,
    4.L/15.L,  2.L/15.L,  5.L/15.L,  4.L/15.L,
    4.L/15.L,  2.L/15.L,  4.L/15.L,  5.L/15.L,
    4.L/15.L,  2.L/15.L,  3.L/15.L,  6.L/15.L,
    4.L/15.L,  2.L/15.L,  2.L/15.L,  7.L/15.L,
    4.L/15.L,  2.L/15.L,  1.L/15.L,  8.L/15.L,
    4.L/15.L,  1.L/15.L,  9.L/15.L,  1.L/15.L,
    4.L/15.L,  1.L/15.L,  8.L/15.L,  2.L/15.L,
    4.L/15.L,  1.L/15.L,  7.L/15.L,  3.L/15.L,
    4.L/15.L,  1.L/15.L,  6.L/15.L,  4.L/15.L,
    4.L/15.L,  1.L/15.L,  5.L/15.L,  5.L/15.L,
    4.L/15.L,  1.L/15.L,  4.L/15.L,  6.L/15.L,
    4.L/15.L,  1.L/15.L,  3.L/15.L,  7.L/15.L,
    4.L/15.L,  1.L/15.L,  2.L/15.L,  8.L/15.L,
    4.L/15.L,  1.L/15.L,  1.L/15.L,  9.L/15.L,
    3.L/15.L, 10.L/15.L,  1.L/15.L,  1.L/15.L,
    3.L/15.L,  9.L/15.L,  2.L/15.L,  1.L/15.L,
    3.L/15.L,  9.L/15.L,  1.L/15.L,  2.L/15.L,
    3.L/15.L,  8.L/15.L,  3.L/15.L,  1.L/15.L,
    3.L/15.L,  8.L/15.L,  2.L/15.L,  2.L/15.L,
    3.L/15.L,  8.L/15.L,  1.L/15.L,  3.L/15.L,
    3.L/15.L,  7.L/15.L,  4.L/15.L,  1.L/15.L,
    3.L/15.L,  7.L/15.L,  3.L/15.L,  2.L/15.L,
    3.L/15.L,  7.L/15.L,  2.L/15.L,  3.L/15.L,
    3.L/15.L,  7.L/15.L,  1.L/15.L,  4.L/15.L,
    3.L/15.L,  6.L/15.L,  5.L/15.L,  1.L/15.L,
    3.L/15.L,  6.L/15.L,  4.L/15.L,  2.L/15.L,
    3.L/15.L,  6.L/15.L,  3.L/15.L,  3.L/15.L,
    3.L/15.L,  6.L/15.L,  2.L/15.L,  4.L/15.L,
    3.L/15.L,  6.L/15.L,  1.L/15.L,  5.L/15.L,
    3.L/15.L,  5.L/15.L,  6.L/15.L,  1.L/15.L,
    3.L/15.L,  5.L/15.L,  5.L/15.L,  2.L/15.L,
    3.L/15.L,  5.L/15.L,  4.L/15.L,  3.L/15.L,
    3.L/15.L,  5.L/15.L,  3.L/15.L,  4.L/15.L,
    3.L/15.L,  5.L/15.L,  2.L/15.L,  5.L/15.L,
    3.L/15.L,  5.L/15.L,  1.L/15.L,  6.L/15.L,
    3.L/15.L,  4.L/15.L,  7.L/15.L,  1.L/15.L,
    3.L/15.L,  4.L/15.L,  6.L/15.L,  2.L/15.L,
    3.L/15.L,  4.L/15.L,  5.L/15.L,  3.L/15.L,
    3.L/15.L,  4.L/15.L,  4.L/15.L,  4.L/15.L,
    3.L/15.L,  4.L/15.L,  3.L/15.L,  5.L/15.L,
    3.L/15.L,  4.L/15.L,  2.L/15.L,  6.L/15.L,
    3.L/15.L,  4.L/15.L,  1.L/15.L,  7.L/15.L,
    3.L/15.L,  3.L/15.L,  8.L/15.L,  1.L/15.L,
    3.L/15.L,  3.L/15.L,  7.L/15.L,  2.L/15.L,
    3.L/15.L,  3.L/15.L,  6.L/15.L,  3.L/15.L,
    3.L/15.L,  3.L/15.L,  5.L/15.L,  4.L/15.L,
    3.L/15.L,  3.L/15.L,  4.L/15.L,  5.L/15.L,
    3.L/15.L,  3.L/15.L,  3.L/15.L,  6.L/15.L,
    3.L/15.L,  3.L/15.L,  2.L/15.L,  7.L/15.L,
    3.L/15.L,  3.L/15.L,  1.L/15.L,  8.L/15.L,
    3.L/15.L,  2.L/15.L,  9.L/15.L,  1.L/15.L,
    3.L/15.L,  2.L/15.L,  8.L/15.L,  2.L/15.L,
    3.L/15.L,  2.L/15.L,  7.L/15.L,  3.L/15.L,
    3.L/15.L,  2.L/15.L,  6.L/15.L,  4.L/15.L,
    3.L/15.L,  2.L/15.L,  5.L/15.L,  5.L/15.L,
    3.L/15.L,  2.L/15.L,  4.L/15.L,  6.L/15.L,
    3.L/15.L,  2.L/15.L,  3.L/15.L,  7.L/15.L,
    3.L/15.L,  2.L/15.L,  2.L/15.L,  8.L/15.L,
    3.L/15.L,  2.L/15.L,  1.L/15.L,  9.L/15.L,
    3.L/15.L,  1.L/15.L, 10.L/15.L,  1.L/15.L,
    3.L/15.L,  1.L/15.L,  9.L/15.L,  2.L/15.L,
    3.L/15.L,  1.L/15.L,  8.L/15.L,  3.L/15.L,
    3.L/15.L,  1.L/15.L,  7.L/15.L,  4.L/15.L,
    3.L/15.L,  1.L/15.L,  6.L/15.L,  5.L/15.L,
    3.L/15.L,  1.L/15.L,  5.L/15.L,  6.L/15.L,
    3.L/15.L,  1.L/15.L,  4.L/15.L,  7.L/15.L,
    3.L/15.L,  1.L/15.L,  3.L/15.L,  8.L/15.L,
    3.L/15.L,  1.L/15.L,  2.L/15.L,  9.L/15.L,
    3.L/15.L,  1.L/15.L,  1.L/15.L, 10.L/15.L,
    2.L/15.L, 11.L/15.L,  1.L/15.L,  1.L/15.L,
    2.L/15.L, 10.L/15.L,  2.L/15.L,  1.L/15.L,
    2.L/15.L, 10.L/15.L,  1.L/15.L,  2.L/15.L,
    2.L/15.L,  9.L/15.L,  3.L/15.L,  1.L/15.L,
    2.L/15.L,  9.L/15.L,  2.L/15.L,  2.L/15.L,
    2.L/15.L,  9.L/15.L,  1.L/15.L,  3.L/15.L,
    2.L/15.L,  8.L/15.L,  4.L/15.L,  1.L/15.L,
    2.L/15.L,  8.L/15.L,  3.L/15.L,  2.L/15.L,
    2.L/15.L,  8.L/15.L,  2.L/15.L,  3.L/15.L,
    2.L/15.L,  8.L/15.L,  1.L/15.L,  4.L/15.L,
    2.L/15.L,  7.L/15.L,  5.L/15.L,  1.L/15.L,
    2.L/15.L,  7.L/15.L,  4.L/15.L,  2.L/15.L,
    2.L/15.L,  7.L/15.L,  3.L/15.L,  3.L/15.L,
    2.L/15.L,  7.L/15.L,  2.L/15.L,  4.L/15.L,
    2.L/15.L,  7.L/15.L,  1.L/15.L,  5.L/15.L,
    2.L/15.L,  6.L/15.L,  6.L/15.L,  1.L/15.L,
    2.L/15.L,  6.L/15.L,  5.L/15.L,  2.L/15.L,
    2.L/15.L,  6.L/15.L,  4.L/15.L,  3.L/15.L,
    2.L/15.L,  6.L/15.L,  3.L/15.L,  4.L/15.L,
    2.L/15.L,  6.L/15.L,  2.L/15.L,  5.L/15.L,
    2.L/15.L,  6.L/15.L,  1.L/15.L,  6.L/15.L,
    2.L/15.L,  5.L/15.L,  7.L/15.L,  1.L/15.L,
    2.L/15.L,  5.L/15.L,  6.L/15.L,  2.L/15.L,
    2.L/15.L,  5.L/15.L,  5.L/15.L,  3.L/15.L,
    2.L/15.L,  5.L/15.L,  4.L/15.L,  4.L/15.L,
    2.L/15.L,  5.L/15.L,  3.L/15.L,  5.L/15.L,
    2.L/15.L,  5.L/15.L,  2.L/15.L,  6.L/15.L,
    2.L/15.L,  5.L/15.L,  1.L/15.L,  7.L/15.L,
    2.L/15.L,  4.L/15.L,  8.L/15.L,  1.L/15.L,
    2.L/15.L,  4.L/15.L,  7.L/15.L,  2.L/15.L,
    2.L/15.L,  4.L/15.L,  6.L/15.L,  3.L/15.L,
    2.L/15.L,  4.L/15.L,  5.L/15.L,  4.L/15.L,
    2.L/15.L,  4.L/15.L,  4.L/15.L,  5.L/15.L,
    2.L/15.L,  4.L/15.L,  3.L/15.L,  6.L/15.L,
    2.L/15.L,  4.L/15.L,  2.L/15.L,  7.L/15.L,
    2.L/15.L,  4.L/15.L,  1.L/15.L,  8.L/15.L,
    2.L/15.L,  3.L/15.L,  9.L/15.L,  1.L/15.L,
    2.L/15.L,  3.L/15.L,  8.L/15.L,  2.L/15.L,
    2.L/15.L,  3.L/15.L,  7.L/15.L,  3.L/15.L,
    2.L/15.L,  3.L/15.L,  6.L/15.L,  4.L/15.L,
    2.L/15.L,  3.L/15.L,  5.L/15.L,  5.L/15.L,
    2.L/15.L,  3.L/15.L,  4.L/15.L,  6.L/15.L,
    2.L/15.L,  3.L/15.L,  3.L/15.L,  7.L/15.L,
    2.L/15.L,  3.L/15.L,  2.L/15.L,  8.L/15.L,
    2.L/15.L,  3.L/15.L,  1.L/15.L,  9.L/15.L,
    2.L/15.L,  2.L/15.L, 10.L/15.L,  1.L/15.L,
    2.L/15.L,  2.L/15.L,  9.L/15.L,  2.L/15.L,
    2.L/15.L,  2.L/15.L,  8.L/15.L,  3.L/15.L,
    2.L/15.L,  2.L/15.L,  7.L/15.L,  4.L/15.L,
    2.L/15.L,  2.L/15.L,  6.L/15.L,  5.L/15.L,
    2.L/15.L,  2.L/15.L,  5.L/15.L,  6.L/15.L,
    2.L/15.L,  2.L/15.L,  4.L/15.L,  7.L/15.L,
    2.L/15.L,  2.L/15.L,  3.L/15.L,  8.L/15.L,
    2.L/15.L,  2.L/15.L,  2.L/15.L,  9.L/15.L,
    2.L/15.L,  2.L/15.L,  1.L/15.L, 10.L/15.L,
    2.L/15.L,  1.L/15.L, 11.L/15.L,  1.L/15.L,
    2.L/15.L,  1.L/15.L, 10.L/15.L,  2.L/15.L,
    2.L/15.L,  1.L/15.L,  9.L/15.L,  3.L/15.L,
    2.L/15.L,  1.L/15.L,  8.L/15.L,  4.L/15.L,
    2.L/15.L,  1.L/15.L,  7.L/15.L,  5.L/15.L,
    2.L/15.L,  1.L/15.L,  6.L/15.L,  6.L/15.L,
    2.L/15.L,  1.L/15.L,  5.L/15.L,  7.L/15.L,
    2.L/15.L,  1.L/15.L,  4.L/15.L,  8.L/15.L,
    2.L/15.L,  1.L/15.L,  3.L/15.L,  9.L/15.L,
    2.L/15.L,  1.L/15.L,  2.L/15.L, 10.L/15.L,
    2.L/15.L,  1.L/15.L,  1.L/15.L, 11.L/15.L,
    1.L/15.L, 12.L/15.L,  1.L/15.L,  1.L/15.L,
    1.L/15.L, 11.L/15.L,  2.L/15.L,  1.L/15.L,
    1.L/15.L, 11.L/15.L,  1.L/15.L,  2.L/15.L,
    1.L/15.L, 10.L/15.L,  3.L/15.L,  1.L/15.L,
    1.L/15.L, 10.L/15.L,  2.L/15.L,  2.L/15.L,
    1.L/15.L, 10.L/15.L,  1.L/15.L,  3.L/15.L,
    1.L/15.L,  9.L/15.L,  4.L/15.L,  1.L/15.L,
    1.L/15.L,  9.L/15.L,  3.L/15.L,  2.L/15.L,
    1.L/15.L,  9.L/15.L,  2.L/15.L,  3.L/15.L,
    1.L/15.L,  9.L/15.L,  1.L/15.L,  4.L/15.L,
    1.L/15.L,  8.L/15.L,  5.L/15.L,  1.L/15.L,
    1.L/15.L,  8.L/15.L,  4.L/15.L,  2.L/15.L,
    1.L/15.L,  8.L/15.L,  3.L/15.L,  3.L/15.L,
    1.L/15.L,  8.L/15.L,  2.L/15.L,  4.L/15.L,
    1.L/15.L,  8.L/15.L,  1.L/15.L,  5.L/15.L,
    1.L/15.L,  7.L/15.L,  6.L/15.L,  1.L/15.L,
    1.L/15.L,  7.L/15.L,  5.L/15.L,  2.L/15.L,
    1.L/15.L,  7.L/15.L,  4.L/15.L,  3.L/15.L,
    1.L/15.L,  7.L/15.L,  3.L/15.L,  4.L/15.L,
    1.L/15.L,  7.L/15.L,  2.L/15.L,  5.L/15.L,
    1.L/15.L,  7.L/15.L,  1.L/15.L,  6.L/15.L,
    1.L/15.L,  6.L/15.L,  7.L/15.L,  1.L/15.L,
    1.L/15.L,  6.L/15.L,  6.L/15.L,  2.L/15.L,
    1.L/15.L,  6.L/15.L,  5.L/15.L,  3.L/15.L,
    1.L/15.L,  6.L/15.L,  4.L/15.L,  4.L/15.L,
    1.L/15.L,  6.L/15.L,  3.L/15.L,  5.L/15.L,
    1.L/15.L,  6.L/15.L,  2.L/15.L,  6.L/15.L,
    1.L/15.L,  6.L/15.L,  1.L/15.L,  7.L/15.L,
    1.L/15.L,  5.L/15.L,  8.L/15.L,  1.L/15.L,
    1.L/15.L,  5.L/15.L,  7.L/15.L,  2.L/15.L,
    1.L/15.L,  5.L/15.L,  6.L/15.L,  3.L/15.L,
    1.L/15.L,  5.L/15.L,  5.L/15.L,  4.L/15.L,
    1.L/15.L,  5.L/15.L,  4.L/15.L,  5.L/15.L,
    1.L/15.L,  5.L/15.L,  3.L/15.L,  6.L/15.L,
    1.L/15.L,  5.L/15.L,  2.L/15.L,  7.L/15.L,
    1.L/15.L,  5.L/15.L,  1.L/15.L,  8.L/15.L,
    1.L/15.L,  4.L/15.L,  9.L/15.L,  1.L/15.L,
    1.L/15.L,  4.L/15.L,  8.L/15.L,  2.L/15.L,
    1.L/15.L,  4.L/15.L,  7.L/15.L,  3.L/15.L,
    1.L/15.L,  4.L/15.L,  6.L/15.L,  4.L/15.L,
    1.L/15.L,  4.L/15.L,  5.L/15.L,  5.L/15.L,
    1.L/15.L,  4.L/15.L,  4.L/15.L,  6.L/15.L,
    1.L/15.L,  4.L/15.L,  3.L/15.L,  7.L/15.L,
    1.L/15.L,  4.L/15.L,  2.L/15.L,  8.L/15.L,
    1.L/15.L,  4.L/15.L,  1.L/15.L,  9.L/15.L,
    1.L/15.L,  3.L/15.L, 10.L/15.L,  1.L/15.L,
    1.L/15.L,  3.L/15.L,  9.L/15.L,  2.L/15.L,
    1.L/15.L,  3.L/15.L,  8.L/15.L,  3.L/15.L,
    1.L/15.L,  3.L/15.L,  7.L/15.L,  4.L/15.L,
    1.L/15.L,  3.L/15.L,  6.L/15.L,  5.L/15.L,
    1.L/15.L,  3.L/15.L,  5.L/15.L,  6.L/15.L,
    1.L/15.L,  3.L/15.L,  4.L/15.L,  7.L/15.L,
    1.L/15.L,  3.L/15.L,  3.L/15.L,  8.L/15.L,
    1.L/15.L,  3.L/15.L,  2.L/15.L,  9.L/15.L,
    1.L/15.L,  3.L/15.L,  1.L/15.L, 10.L/15.L,
    1.L/15.L,  2.L/15.L, 11.L/15.L,  1.L/15.L,
    1.L/15.L,  2.L/15.L, 10.L/15.L,  2.L/15.L,
    1.L/15.L,  2.L/15.L,  9.L/15.L,  3.L/15.L,
    1.L/15.L,  2.L/15.L,  8.L/15.L,  4.L/15.L,
    1.L/15.L,  2.L/15.L,  7.L/15.L,  5.L/15.L,
    1.L/15.L,  2.L/15.L,  6.L/15.L,  6.L/15.L,
    1.L/15.L,  2.L/15.L,  5.L/15.L,  7.L/15.L,
    1.L/15.L,  2.L/15.L,  4.L/15.L,  8.L/15.L,
    1.L/15.L,  2.L/15.L,  3.L/15.L,  9.L/15.L,
    1.L/15.L,  2.L/15.L,  2.L/15.L, 10.L/15.L,
    1.L/15.L,  2.L/15.L,  1.L/15.L, 11.L/15.L,
    1.L/15.L,  1.L/15.L, 12.L/15.L,  1.L/15.L,
    1.L/15.L,  1.L/15.L, 11.L/15.L,  2.L/15.L,
    1.L/15.L,  1.L/15.L, 10.L/15.L,  3.L/15.L,
    1.L/15.L,  1.L/15.L,  9.L/15.L,  4.L/15.L,
    1.L/15.L,  1.L/15.L,  8.L/15.L,  5.L/15.L,
    1.L/15.L,  1.L/15.L,  7.L/15.L,  6.L/15.L,
    1.L/15.L,  1.L/15.L,  6.L/15.L,  7.L/15.L,
    1.L/15.L,  1.L/15.L,  5.L/15.L,  8.L/15.L,
    1.L/15.L,  1.L/15.L,  4.L/15.L,  9.L/15.L,
    1.L/15.L,  1.L/15.L,  3.L/15.L, 10.L/15.L,
    1.L/15.L,  1.L/15.L,  2.L/15.L, 11.L/15.L,
    1.L/15.L,  1.L/15.L,  1.L/15.L, 12.L/15.L
};
DOF_TYPE DOF_DG15_ = {
    DofCache, "DG15", DG15_points, NULL,
    DOF_DG14, DOF_P15, &hp_info_DG,
    phgDofInterpC2FGeneric, phgDofInterpF2CGeneric,
    phgDofInitFuncPoint,
    DG_bas, DG_grad, NULL, FE_L2,
    TRUE, FALSE, -1, 816, 15, -1, 1,
    0, 0, 0, 816
};

/*============ End of structs generated by 'utils/lagrange_gen' ============*/

#endif	/* USE_HB_FOR_DG */
